/*--------------------------------------------------------------------------+
$Id: Opt4JDseBackend.java 3382 2016-08-08 15:26:54Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.contribution;

import java.util.Collection;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.exploration.alg.dse.CompositeExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.Explorer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemParameterContainer;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3.AF3SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.extension.DseInputParametersBase;
import org.fortiss.af3.exploration.extension.IDseBackend;
import org.fortiss.af3.exploration.extension.IDseInputParameters;
import org.fortiss.af3.exploration.model.ExplorationSolution;
import org.fortiss.af3.exploration.model.ExplorationSpecification;
import org.fortiss.af3.exploration.model.IExplorationFeature;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.solutions.ExplorationSolutionSet;
import org.fortiss.af3.platform.model.PlatformArchitecture;

/**
 * Contributor class to allow launching this DSE backend from the DSE perspective.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3382 $
 * @ConQAT.Rating RED Hash: CBF2C6C02ED967765B27C1E6641B9892
 */
public class Opt4JDseBackend implements IDseBackend {

	/** Holds the input parameters required for the DSE in this plugin. */
	private Opt4JDSEInputParameters inputParameterContainer = new Opt4JDSEInputParameters();

	/** {@inheritDoc} */
	@Override
	public IDseInputParameters getRequiredParameterContainer() {
		return inputParameterContainer;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Optional<ExplorationSolution> executeDSE(ExplorationSpecification spec,
			Set<Class<?>> solutionTypes, IDseInputParameters inputParameters,
			Collection<IExplorationFeature> explorationModules, IProgressMonitor monitor)
			throws Exception {
		ExplorationSolutionSet explorationSolutionSet;
		SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

		inputParameters.validateInputParametersPresent();
		if(!(inputParameters instanceof Opt4JDSEInputParameters)) {
			throw new Exception("Wrong input parameter container type.");
		}
		Opt4JDSEInputParameters parameters = (Opt4JDSEInputParameters)inputParameters;

		// Input to DSE: {@link SystemModelAdapter} and {@link DesignSpaceExploration}.
		// SystemModelAdapter systemModelAdapter;
		try {
			systemModelAdapter =
					new AF3SystemModelAdapter(
							parameters.getInputParameter(ComponentArchitecture.class),
							parameters.getInputParameter(PlatformArchitecture.class),
							parameters.getInputParameter(SystemParameterContainer.class));
		} catch(Exception ex) {
			// ExplorationUiUtil
			// .showExceptionErrorAsync(
			// "Error setting up the DSE",
			// "Could not transform the input models to a valid input for the DSE. Please check the Component Architecture, Platform Architecture, and referenced models.",
			// ex);
			ex.printStackTrace();
			return Optional.empty();
		}

		try {
			Explorer explorer = new Explorer();
			Set<Class<? extends IExplorationEncoding>> solPhenotypeTypes = new HashSet<>();
			for(Class<?> solType : solutionTypes) {
				if(!IExplorationEncoding.class.isAssignableFrom(solType)) {
					throw new Exception(
							"The demanded solution is not available in the sleceted DSE backend.");
				}
				solPhenotypeTypes.add((Class<? extends IExplorationEncoding>)solType);
			}
			CompositeExplorationSolution<?> expResult =
					explorer.explore(parameters.getInputParameter(DseSpecification.class),
							systemModelAdapter, solPhenotypeTypes, monitor, true);
			explorationSolutionSet = expResult.convertToExplorationSolution();
			explorationSolutionSet.setInputModelAdapter(systemModelAdapter);
		} catch(Exception ex) {
			// TODO: readd code to present the schedule which failed... Use the LoggingService from
			// the kernel.
			// ExplorationUiUtil.showExceptionErrorAsync(
			// "Design Space Exploration: Error", ex.getMessage(), ex);
			// if(ex instanceof ScheduleRuntimeException) {
			// StrictTTSchedule<?, ?> faultySchedule =
			// ((ScheduleRuntimeException)ex).getSchedule();
			// SchedulePlotter sp = new SchedulePlotter(faultySchedule);
			// sp.createPlot();
			// }
			throw ex;
		}
		return Optional.of(explorationSolutionSet);
	}

	/** Defines the required input parameters of this DSE backend. */
	protected class Opt4JDSEInputParameters extends DseInputParametersBase {
		/** References the input {@link ComponentArchitecture} for the DSE. */
		private ComponentArchitecture componentArchitecture;

		/** References the input {@link PlatformArchitecture} for the DSE. */
		private PlatformArchitecture platformArchitecture;

		/** Holds the extracted parameters needed for the MOEA DSE. */
		SystemParameterContainer systemParameterContainer;

		/** Holds Opt4J DSE parameters. */
		private DseSpecification dseModel;

		/** Constructor. */
		public Opt4JDSEInputParameters() {
			putInputParameterType(ComponentArchitecture.class, componentArchitecture);
			putInputParameterType(PlatformArchitecture.class, platformArchitecture);
			putInputParameterType(SystemParameterContainer.class, systemParameterContainer);
			putInputParameterType(DseSpecification.class, dseModel);
		}
	}
}
