/*--------------------------------------------------------------------------+
$Id: DSEFactory.java 3797 2016-10-28 11:14:06Z diewald $
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.ArrayList;
import java.util.List;

import javax.activation.UnsupportedDataTypeException;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.ArchitectureExplorationCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.safetyfunctionarch.SafetyFunctionArchCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.AbstractTaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.create.taskmapping.TaskMappingCreator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.graph.DecoderDependencyGraph;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.problem.CompositeProblemModuleBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.ArchitectureDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.ArchitectureExplorationProblemModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.StrictTTDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.comm.MessageDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatemapping.AbstractTaskMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph.InstantiatedTaskMappingDecoderAcyclic;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction.SFEncodingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction.SFGraphDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction.SFMappingConstraintDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction.SFMappingDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.taskmapping.AbstractTaskMappingIdentityDecoder;
import org.fortiss.af3.exploration.alg.dse.evaluator.ArchitectureEvaluator;
import org.fortiss.af3.exploration.alg.dse.sysmodel.FailSilentExecModelFactory;
import org.fortiss.af3.exploration.alg.dse.sysmodel.FaultDetectionVotingExecModelFactory;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.feature.SafetyExploration;
import org.fortiss.af3.exploration.moea.model.parameters.ExecutionModel;

/**
 * Factory to create exploration problems solved by the DSE. It is used to create the single parts
 * of the chain that define and solve the associated optimization problem.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3797 $
 * @ConQAT.Rating YELLOW Hash: 17294767F6BCCBD53BF9861D1B22A08E
 */
public class DSEFactory {

	/** The instance (singleton) of the {@link DSEFactory}. */
	private static final class InstanceHolder {
		/** the "real" instance. */
		static final DSEFactory INSTANCE = new DSEFactory();
	}

	/** Private constructor to prevent instantiation from external (singleton). */
	private DSEFactory() {
	}

	/** Returns the instance of this {@link DSEFactory}. */
	public static DSEFactory getInstance() {
		return InstanceHolder.INSTANCE;
	}

	/**
	 * Creates/Registers the sub problem definitions that compose the overall architectural
	 * exploration problem.
	 * 
	 * @param archExpProblemModule
	 *            The {@link CompositeProblemModuleBase} declaring the architectural exploration
	 *            problem.
	 * @param dse
	 *            The design space exploration description.
	 * @param systemModelAdapter
	 *            The adapter to model of the system on which the dse is executed.
	 * @throws Exception
	 *             if (one of) the sub problem(s) cannot be registered.
	 */
	@SuppressWarnings("unchecked")
	public
			<C, S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
			void createSubProblems(ArchitectureExplorationProblemModule archExpProblemModule,
					DseSpecification dse, SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter,
					DecoderDependencyGraph execDepGraph) throws Exception {
		// Java-limitation: it is not possible to pass the generic to the static methods.
		TaskMappingFactory<S, T> tmFactory =
				(TaskMappingFactory<S, T>)TaskMappingFactory.getInstance();
		SafetyExploration sfExplorationModule =
				pickFirstInstanceOf(SafetyExploration.class, dse.getFeatures());
		if(sfExplorationModule != null) {
			// Initialize & register safety function adapters with the SystemModelAdapter.
			sfExplorationModule.getAdapters().parallelStream()
					.forEach(a -> a.initializeAdapter(systemModelAdapter));
			systemModelAdapter.addSafetyFunctionAdapters(sfExplorationModule.getAdapters());
			createSafetyFunctionArchProblem(archExpProblemModule, systemModelAdapter, execDepGraph);
		}
		createAbstractTaskMappingProblem(tmFactory, archExpProblemModule, dse, systemModelAdapter,
				execDepGraph);
		createTaskMappingInstantiationProblem(tmFactory, archExpProblemModule, dse,
				systemModelAdapter, execDepGraph);
	}

	/**
	 * Creates the safety function architecture exploration sub-problem, or, more specifically,
	 * the
	 * creators of this sub-problem.
	 *
	 * @param archExpProblemModule
	 *            The {@link CompositeProblemModuleBase} declaring the architectural exploration
	 *            problem.
	 * @param systemModelAdapter
	 *            The adapter to model of the system on which the dse is executed.
	 * @throws Exception
	 *             if (one of) the sub problem(s) cannot be registered.
	 */
	@SuppressWarnings("unchecked")
	private <C, T extends SafetyFunctionArchEncoding<C>> void createSafetyFunctionArchProblem(
			ArchitectureExplorationProblemModule archExpProblemModule,
			SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter,
			DecoderDependencyGraph execDepGraph) throws Exception {
		// These suppress warnings are required since type erasure in Java does not allow to pass
		// XY<Z>.class.
		@SuppressWarnings("rawtypes") SafetyFunctionArchCreator sfCreator =
				new SafetyFunctionArchCreator<C>(systemModelAdapter, execDepGraph);
		@SuppressWarnings("rawtypes") SFEncodingDecoder sfEncodingDecoder =
				new SFEncodingDecoder<>();
		@SuppressWarnings("rawtypes") SFGraphDecoder sfGraphDecoder =
				new SFGraphDecoder<C>(systemModelAdapter);
		@SuppressWarnings("rawtypes") SFMappingDecoder sfMappingDecoder =
				new SFMappingDecoder<C>(systemModelAdapter);
		@SuppressWarnings("rawtypes") SFMappingConstraintDecoder sfMappingConstraintDecoder =
				new SFMappingConstraintDecoder<C>(systemModelAdapter);

		sfCreator.registerAssocDecoders(sfEncodingDecoder);
		sfCreator.registerAssocDecoders(sfGraphDecoder);
		sfCreator.registerAssocDecoders(sfMappingDecoder);
		sfCreator.registerAssocDecoders(sfMappingConstraintDecoder);

		@SuppressWarnings("rawtypes") List<ComposableDecoder> sfDecoders = new ArrayList<>();
		sfDecoders.add(sfEncodingDecoder);
		sfDecoders.add(sfGraphDecoder);
		sfDecoders.add(sfMappingDecoder);
		sfDecoders.add(sfMappingConstraintDecoder);

		archExpProblemModule.registerSubProblem(SafetyFunctionArchEncoding.class, sfCreator,
				sfDecoders);
	}

	/**
	 * Creates a task mapping sub-problem (see {@link AbstractTaskMappingEncoding}) which describes
	 * the problem of mapping a logical/functional application architecture to a defined platform
	 * architecture. Note that logical components may also be empty containers in this sub problem
	 * description and which must be instantiated first to derive a valid deployment.
	 * 
	 * @param tmFactory
	 *            Factory which manages the creation tasks associated with task mapping problems
	 * @param archExpProblemModule
	 *            The {@link CompositeProblemModuleBase} declaring the architectural exploration
	 *            problem.
	 * @param dse
	 *            The design space exploration description.
	 * @param systemModelAdapter
	 *            The adapter to model of the system on which the dse is executed.
	 * @throws Exception
	 *             if (one of) the sub problem(s) cannot be registered.
	 */
	@SuppressWarnings("unchecked")
	private <S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
			void createAbstractTaskMappingProblem(TaskMappingFactory<S, T> tmFactory,
					ArchitectureExplorationProblemModule archExpProblemModule,
					DseSpecification dse, SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					DecoderDependencyGraph execDepGraph) throws Exception {
		Class<AbstractTaskMappingEncoding> tmEnodingClass = AbstractTaskMappingEncoding.class;
		AbstractTaskMappingCreator tmCreator =
				tmFactory.createAbstractTaskMappingCreator(dse, systemModelAdapter, execDepGraph);
		@SuppressWarnings("rawtypes") List<ComposableDecoder> abstrTaskDecoders = new ArrayList<>();
		AbstractTaskMappingDecoder<?, ?> atmDecoder =
				tmFactory.createAbstractTaskMappingDecoder(systemModelAdapter, dse);
		AbstractTaskMappingIdentityDecoder idtDecoder = new AbstractTaskMappingIdentityDecoder();
		abstrTaskDecoders.add(atmDecoder);
		abstrTaskDecoders.add(idtDecoder);
		// TODO: replace with the registration method that uses lists.
		tmCreator.registerAssocDecoders(atmDecoder);
		tmCreator.registerAssocDecoders(idtDecoder);

		archExpProblemModule.registerSubProblem(tmEnodingClass, tmCreator, abstrTaskDecoders);
	}

	/**
	 * Creates a task mapping sub-problem (see {@link InstantiatedTaskMappingEncoding}) which
	 * describes
	 * the problem of instantiating a mapped logical/functional application architecture to a
	 * defined platform architecture. Here, the focus is on optimizing the instantiation of
	 * "abstract" {@link IDeployableComponentAdapter}s.
	 * 
	 * @param tmFactory
	 *            Factory which manages the creation tasks associated with task mapping problems
	 * @param archExpProblemModule
	 *            The {@link CompositeProblemModuleBase} declaring the architectural exploration
	 *            problem.
	 * @param dse
	 *            The design space exploration description.
	 * @param systemModelAdapter
	 *            The adapter to model of the system on which the dse is executed.
	 * @throws Exception
	 *             if (one of) the sub problem(s) cannot be registered.
	 */
	@SuppressWarnings("unchecked")
	private <S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
			void createTaskMappingInstantiationProblem(TaskMappingFactory<S, T> tmFactory,
					ArchitectureExplorationProblemModule archExpProblemModule,
					DseSpecification dse, SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					DecoderDependencyGraph execDepGraph) throws Exception {
		// These suppress warnings are required since type erasure in Java does not allow to pass
		// XY<Z>.class.
		StrictTTDecoder<S, T> tmDecoder = tmFactory.createStrictTTDecoder(systemModelAdapter, dse);
		InstantiatedTaskMappingDecoder<S, ?> tgDecoder =
				tmFactory.createInstantiatedTaskMappingDecoder(systemModelAdapter);
		InstantiatedTaskMappingDecoderAcyclic<S, ?> acycTgDecoder =
				tmFactory.createAbstractToAcyclicTaskGraphDecoder(systemModelAdapter);
		@SuppressWarnings("rawtypes") MessageDecoder msgDecoder =
				tmFactory.createMessageDecoder(systemModelAdapter);
		@SuppressWarnings("rawtypes") TaskMappingCreator tmCreator =
				tmFactory.createTaskMappingCreator(dse, systemModelAdapter, execDepGraph);
		@SuppressWarnings("rawtypes") List<ComposableDecoder> instTaskDecoders = new ArrayList<>();
		instTaskDecoders.add(tgDecoder);
		instTaskDecoders.add(acycTgDecoder);
		instTaskDecoders.add(msgDecoder);
		instTaskDecoders.add(tmDecoder);
		tmCreator.registerAssocDecoders(instTaskDecoders);

		archExpProblemModule.registerSubProblem(InstantiatedTaskMappingEncoding.class, tmCreator,
				instTaskDecoders);
	}

	/**
	 * Creates a new {@link ArchitectureExplorationCreator} which manages the creation of the
	 * concrete architecture exploration problem, i.e. the {@link ArchitectureExplorationCreator}
	 * calls the creators of the genotypes composing the architecture exploration problem.
	 */
	public ArchitectureExplorationCreator createArchitectureExplorationCreator(
			DseSpecification dse, SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		return new ArchitectureExplorationCreator(dse, systemModelAdapter);
	}

	/**
	 * Returns the {@link TaskMappingFactory} that fits the given {@link ExecutionModel} from the
	 * DSE.
	 * 
	 * @throws UnsupportedDataTypeException
	 *             if the given {@link ExecutionModel} is not "wrapped" to factory in this method.
	 */
	public TaskMappingFactory<? extends TaskMappingEntry, ? extends TaskMappingEncoding<?>>
			createTaskMappingFactory(ExecutionModel execModel) throws UnsupportedDataTypeException {
		switch(execModel) {
			case TT_FAIL_SILENCE:
				return FailSilentExecModelFactory.getInstance();
			case TT_FAULT_DETECTION_VOTING:
				return FaultDetectionVotingExecModelFactory.getInstance();
			default:
				throw new UnsupportedDataTypeException(
						"Cannot create a factory for the execution model " + execModel.getName() +
								". There is no no factory registered for this executio model.");
		}
	}

	/**
	 * Creates a decoder for the architectural exploration problem that shall be solved by the
	 * DSE.
	 * It delegates the subproblems to sub decoders, e.g. a decoder for the task mapping problem.
	 */
	public ArchitectureDecoder createArchitectureDecoder(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			DecoderDependencyGraph execdepGraph) {
		return new ArchitectureDecoder(systemModelAdapter, execdepGraph);
	}

	/**
	 * Creates an evaluator for the architectural exploration problem that shall be solved by the
	 * DSE.
	 * It delegates the evaluation of the overall problem to sub-decoders, e.g. an evaluator of the
	 * power consumption of a TT-schedule.
	 */
	public ArchitectureEvaluator createArchitectureEvaluator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		return new ArchitectureEvaluator(systemModelAdapter);
	}
}
