/*--------------------------------------------------------------------------+
$Id: Opt4JExplorerBackend.java 3730 2016-10-19 15:43:43Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickFirstInstanceOf;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;
import org.fortiss.af3.exploration.alg.dse.CompositeExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.backend.ExplorerBackendBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.mating.MatingCrossoverMutateRepairModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.safetyfunctionarch.CopySafetyFunctionArchModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.copy.taskmapping.CopyTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.CrossoverAbstractTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.NopCrossoverInstantiatedTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.crossover.NopCrossoverSFArchModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.safetyarch.MutateSafetyFunctionArchitectureModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskinstantiation.MutateTaskInstanceModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping.MutateTaskMappingModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.ArchitectureExplorationProblemModule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.ArchitectureSolution;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.fortiss.af3.exploration.moea.model.feature.SafetyExploration;
import org.fortiss.af3.exploration.moea.model.parameters.EAOptimizer;
import org.opt4j.core.common.completer.IndividualCompleterModule;
import org.opt4j.core.config.Task;
import org.opt4j.core.config.TaskStateListener;
import org.opt4j.core.optimizer.Archive;
import org.opt4j.core.optimizer.Control;
import org.opt4j.core.start.Opt4JTask;
import org.opt4j.optimizers.ea.EvolutionaryAlgorithmModule;
import org.opt4j.optimizers.ea.Spea2Module;

import com.google.inject.Module;

/**
 * Opt4J optimization backend that defines the DSE task for this particular optimization framework.
 * Here, the problem definitiona, the variation modules, the optimizer, and the solution format are
 * defined. Additionally, the optimization task is started here and the results are returned.
 * 
 * @author huang
 * @author $Author: diewald $
 * @version $Rev: 3730 $
 * @ConQAT.Rating RED Hash: 2DF68013FA764456520184882EA0822E
 */
public class Opt4JExplorerBackend extends ExplorerBackendBase<EAOptimizer> {

	/** Opt4J optimization process */
	private Opt4JTask opt4JTask;

	/** EA parameters. */
	private EAOptimizer opt4JParameters;

	/**
	 * Contains the problem module that includes the evaluators associated with the goals of the
	 * DSE.
	 */
	private ArchitectureExplorationProblemModule archExplorationProblemModule;

	/**
	 * Constructor that initializes the Opt4J backend, and sets optimizer independent parameters
	 * and the goal specification
	 * 
	 * @throws Exception
	 */
	public Opt4JExplorerBackend(DseSpecification dse,
			Map<ExplorationTarget<?>, ExplorationTarget<?>> targetAssocMap,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends IExplorationEncoding>> requestedSolutions,
			IProgressMonitor progressMonitor) throws Exception {
		// Optimizer independent configuration
		super(dse, targetAssocMap, systemModelAdapter, requestedSolutions, progressMonitor);

		if(dse.getOptimizer() instanceof EAOptimizer) {
			init((EAOptimizer)dse.getOptimizer());
		} else {
			throw new Exception(
					"Opt4JExplorerBackend cannot be configured with optimizer dependent configuration of type " +
							dse.getClass());
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void init(EAOptimizer opt4JParameters) {
		this.opt4JParameters = opt4JParameters;

		System.out.println("Initializing EA optimizer (generations: " + "" +
				opt4JParameters.getIterations() + ", alpha: " + opt4JParameters.getAlpha() + ").");

		// Instantiate modules that describe the optimization problem
		Collection<Module> modules = new ArrayList<Module>();

		// Evolutionary algorithm module
		EvolutionaryAlgorithmModule evolutionaryAlgorithmModule = new EvolutionaryAlgorithmModule();
		evolutionaryAlgorithmModule.setGenerations(opt4JParameters.getIterations());
		evolutionaryAlgorithmModule.setAlpha(opt4JParameters.getAlpha());
		evolutionaryAlgorithmModule.setMu(opt4JParameters.getMu());
		evolutionaryAlgorithmModule.setCrossoverRate(0.3);
		modules.add(evolutionaryAlgorithmModule);

		// SPEA2 specific module
		Spea2Module spea2Module = new Spea2Module();
		spea2Module.setTournament(opt4JParameters.getSpea2Tournament());
		modules.add(spea2Module);
		// Nsga2Module nsga2Module = new Nsga2Module();
		// modules.add(nsga2Module);

		MatingCrossoverMutateRepairModule mating =
				new MatingCrossoverMutateRepairModule(systemModelAdapter);
		modules.add(mating);

		// TODO: Replace {@link MultiEvaluator} with {@link ExplorationGoalPriorityMultiEvaluator}
		// that is
		// not based on a @Priority annotation, but on the priority attribute provided by the goal
		// specifications for the different optimization dimensions.
		// AbstractModule multiEvaluatorModule = new AbstractModule() {
		// @Override
		// protected void configure() {
		// bind(MultiEvaluator.class).to(ExplorationGoalPriorityMultiEvaluator.class);
		//
		// }
		// };
		// modules.add(multiEvaluatorModule);

		archExplorationProblemModule =
				new ArchitectureExplorationProblemModule(dse, systemModelAdapter,
						requestedSolutions);
		modules.add(archExplorationProblemModule);

		// Configure operators
		// 1. Crossover
		CrossoverAbstractTaskMappingModule crossoverModule =
				new CrossoverAbstractTaskMappingModule(systemModelAdapter, 0.3);
		modules.add(crossoverModule);
		NopCrossoverInstantiatedTaskMappingModule nopCrossoverInstMapping =
				new NopCrossoverInstantiatedTaskMappingModule();
		modules.add(nopCrossoverInstMapping);
		NopCrossoverSFArchModule nopCrossoverSFArch = new NopCrossoverSFArchModule();
		modules.add(nopCrossoverSFArch);

		// =====================================
		// 2. Mutation (in order)
		// =====================================

		// Safety functions (if the safety exploration module is enabled)
		if(pickFirstInstanceOf(SafetyExploration.class, dse.getFeatures()) != null) {
			MutateSafetyFunctionArchitectureModule sfArchModule =
					new MutateSafetyFunctionArchitectureModule(systemModelAdapter);
			modules.add(sfArchModule);

			// FIXME: add the safety complianceSpec to the input models if needed.
			// Also add the SafetyCompliance model to the set of input models.
			// SafetyComplianceConstraint scConstr =
			// (SafetyComplianceConstraint)targetAssocMap
			// .entrySet()
			// .stream()
			// .filter(e -> SafetyComplianceConstraint.class.isAssignableFrom(e
			// .getKey().getClass())).findAny().get().getValue();
			//
			// systemModelAdapter.putInputModel(SafetyComplianceSpecification.class,
			// scConstr.getSafetyComplianceSpecification());
		}

		// Task instantiation
		MutateTaskInstanceModule mutateInstanceModule =
				new MutateTaskInstanceModule(systemModelAdapter);
		// mutateInstanceModule.setMutationRateType(MutationRateType.ADAPTIVE);
		modules.add(mutateInstanceModule);

		// Task allocation
		MutateTaskMappingModule mutateModule = new MutateTaskMappingModule(systemModelAdapter);
		// mutateModule.setMutationRateType(MutationRateType.ADAPTIVE);
		// mutateModule.setMutationRate(0.125);
		modules.add(mutateModule);

		// 3. Genotype copy
		modules.add(new CopyTaskMappingModule());
		modules.add(new CopySafetyFunctionArchModule());

		// Define a completer module that "controls" the execution of the decode and the evaluation
		// phases. It also allows to parallelize the execution of these phases.
		IndividualCompleterModule comp = new IndividualCompleterModule();
		if(opt4JParameters.isParallelExecution()) {
			comp.setType(IndividualCompleterModule.Type.PARALLEL);
			comp.setThreads(Runtime.getRuntime().availableProcessors());
		}
		modules.add(comp);

		// Create and initialize optimization task
		opt4JTask = new Opt4JTask(false);
		opt4JTask.init(modules);

		if(progressMonitor != null) {
			opt4JTask.addStateListener(new TaskStateListener() {

				@Override
				public void stateChanged(Task task) {
					// stateChanged() is invoked after every generation
					progressMonitor.worked(1);

					if(progressMonitor.isCanceled()) {
						System.out.println("DSE has been canceled.");
						Control control = opt4JTask.getInstance(Control.class);
						control.doStop();
					}
				}
			});
		}
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws Exception
	 */
	@Override
	public CompositeExplorationSolution<?> explore() throws Exception {

		System.out.println("Running EA optimization...");
		if(progressMonitor != null) {
			progressMonitor.beginTask("EA optimization", opt4JParameters.getIterations());
		}

		CompositeExplorationSolution<?> rval = null;
		Exception ex = null;
		try {
			// Execute optimization
			opt4JTask.execute();

			rval =
					new Opt4JExplorationSolutionSet<ArchitectureSolution>(
							opt4JTask.getInstance(Archive.class), targetAssocMap,
							archExplorationProblemModule);

			rval.setCompletedIterations(opt4JTask.getIteration());

		} catch(Exception e) {
			ex = e;
		} finally {
			if(rval != null) {
				rval = validate(rval, archExplorationProblemModule.getEvaluators());
			}
			opt4JTask.close();
		}

		if(ex != null) {
			throw ex;
		}

		return rval;
	}

}
