/*--------------------------------------------------------------------------+
$Id: SafetyFunctionArchEncoding.java 3797 2016-10-28 11:14:06Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch;

import static org.fortiss.af3.exploration.alg.util.AF3Utils.getChannelNumbersSafetyArch;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getTriggerNumbersSafetyArch;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isDiagConnectedSafetyArch;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isDiagPresentSafetyArch;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISafetyFunctionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.safety.model.SafetyArchitecture;
import org.opt4j.core.Genotype;

/**
 * Encodes the design of safety functions that are contained in a logical architecture. It consists
 * of the {@link SafetyFunctionArchEntry}s of all identified safety function of the input model.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3797 $
 * @ConQAT.Rating YELLOW Hash: 462D04A5BD9D6AC86EEE3505FA48A7A3
 */
public class SafetyFunctionArchEncoding<C> implements IExplorationEncoding, ComposableGenotype,
		ComposablePhenotype<SafetyFunctionArchEncoding<C>>, Cloneable {

	/** Random generator */
	private RandomExploration rand = RandomExploration.getInstance();

	/**
	 * Collection that consists of safety function design entries for each safety function that is
	 * present in the input models.
	 */
	private List<SafetyFunctionArchEntry<C>> safetyFunctionArch = new ArrayList<>();

	/**
	 * Constructor for creating a {@link SafetyFunctionArchEncoding} with entries for each safety
	 * function defined in the input models. The selected entries can be either {@code randomize}d
	 * or default to a 1oo1 design.
	 */
	public SafetyFunctionArchEncoding(Collection<ISafetyFunctionAdapter<C>> safetyFunctions,
			boolean initialize, boolean randomize) {
		if(initialize) {
			for(ISafetyFunctionAdapter<C> currSF : safetyFunctions) {
				if(randomize) {
					safetyFunctionArch.add(createRandomSFDesign(currSF));
				} else {
					safetyFunctionArch.add(createDefaultSFDesign(currSF));
				}
			}
		}
	}

	/**
	 * Constructor for creating a {@link SafetyFunctionArchEncoding} with the given encoding. Each
	 * {@link SafetyFunctionArchEntry} will be cloned.
	 */
	public SafetyFunctionArchEncoding(Collection<SafetyFunctionArchEntry<C>> safetyFunctionDesigns) {
		for(SafetyFunctionArchEntry<C> entry : safetyFunctionDesigns) {
			this.safetyFunctionArch.add(entry.clone());
		}
	}

	/** Creates a 1oo1 {@link SafetyFunctionArchEntry}. */
	public SafetyFunctionArchEntry<C>
			createDefaultSFDesign(ISafetyFunctionAdapter<C> safetyFunction) {
		return new SafetyFunctionArchEntry<C>(safetyFunction, 1, 0, false, false);
	}

	/**
	 * Creates a {@link SafetyFunctionArchEntry} with randomized parameters subject to the
	 * limitations defined in the given {@link ISafetyFunctionAdapter}.
	 */
	public SafetyFunctionArchEntry<C>
			createRandomSFDesign(ISafetyFunctionAdapter<C> safetyFunction) {
		List<SafetyArchitecture> sfArch =
				new ArrayList<>(safetyFunction.getAllowedSafetyArchitectures());
		int randSFArchIdx = rand.nextInt(sfArch.size());
		SafetyArchitecture selectedSFArch = sfArch.get(randSFArchIdx);

		int selChannelNum = getChannelNumbersSafetyArch(selectedSFArch);
		int triggerNum = getTriggerNumbersSafetyArch(selectedSFArch);
		boolean diagPresent = isDiagPresentSafetyArch(selectedSFArch).orElse(false);
		boolean diagUnitConnected = isDiagConnectedSafetyArch(selectedSFArch).orElse(false);

		return new SafetyFunctionArchEntry<C>(safetyFunction, selChannelNum, triggerNum,
				diagPresent, diagUnitConnected);
	}

	/**
	 * Adds a {@link SafetyFunctionArchEntry} to the encoding of all safety functions in the
	 * system.
	 */
	public boolean addEntry(SafetyFunctionArchEntry<C> entry) {
		return safetyFunctionArch.add(entry);
	}

	/**
	 * Returns the {@link SafetyFunctionArchEntry} that is located at the given {@code index} of the
	 * encoding.
	 */
	public SafetyFunctionArchEntry<C> getEntry(int index) {
		if(index < 0 || index >= safetyFunctionArch.size()) {
			return null;
		}
		return safetyFunctionArch.get(index);
	}

	/**
	 * Returns the {@link SafetyFunctionArchEntry}s that compose {@code this}
	 * {@link SafetyFunctionArchEncoding}.
	 */
	public Collection<SafetyFunctionArchEntry<C>> getAllEntries() {
		return safetyFunctionArch;
	}

	/**
	 * Returns the {@link SafetyFunctionArchEntry} that represents the chosen design of the given
	 * {@link ISafetyFunctionAdapter}.
	 */
	public Optional<SafetyFunctionArchEntry<C>>
			getDesignOf(ISafetyFunctionAdapter<?> safetyFunction) {
		return safetyFunctionArch.stream()
				.filter(sfd -> sfd.getSafetyFunctionAdapter() == safetyFunction).findAny();
	}

	/** {@inheritDoc} */
	@Override
	public int size() {
		return safetyFunctionArch.size();
	}

	/** {@inheritDoc} */
	@Override
	public <G extends Genotype> G newInstance() {
		Collection<ISafetyFunctionAdapter<C>> sfAdapters =
				safetyFunctionArch.stream().map(sf -> sf.getSafetyFunctionAdapter())
						.collect(Collectors.toList());
		@SuppressWarnings("unchecked") G sfArchEncoding =
				(G)new SafetyFunctionArchEncoding<C>(sfAdapters, true, true);
		return sfArchEncoding;
	}

	/** {@inheritDoc} */
	@Override
	public Object clone() {
		return new SafetyFunctionArchEncoding<C>(safetyFunctionArch);
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return getAllEntries().toString();
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends ComposableGenotype> getGTypeToRegister() {
		return SafetyFunctionArchEncoding.class;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<? extends ComposablePhenotype<?>> getPTypeToRegister() {
		return (Class<? extends ComposablePhenotype<?>>)SafetyFunctionArchEncoding.class;
	}
}
