/*--------------------------------------------------------------------------+
$Id: TaskGraphEncoding.java 3797 2016-10-28 11:14:06Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph;

import static org.fortiss.af3.exploration.alg.util.GraphUtils.getGraphOf;

import java.util.Collection;
import java.util.HashSet;
import java.util.stream.Collectors;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ChannelAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.opt4j.core.Genotype;

/**
 * Encoding (as a {@link Genotype}) which represents the set of deployable Components that shall be
 * mapped to {@link IExecutionUnitAdapter}s by the DSE.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3797 $
 * @ConQAT.Rating RED Hash: D089A8D66AECE47F6E0A840767FA7891
 */
public class TaskGraphEncoding<G extends Genotype, C, T extends TaskGraph> implements
		IExplorationEncoding, ComposablePhenotype<G>, Cloneable {

	/** References the {@link SystemModelAdapter} which adapts the input models. */
	protected SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter;

	/** {@link TaskGraph} that represents the deployable components and their signals. */
	protected Collection<T> taskGraphs;

	/**
	 * Collection of {@link IDeployableComponentAdapter}s that shall be mapped to deployment targets
	 * ({@link IExecutionUnitAdapter}).
	 */
	protected Collection<IDeployableComponentAdapter<C>> deployableComponents = new HashSet<>();

	/**
	 * Subset of {@link IDeployableComponentAdapter}s from the collection of deployable components
	 * that may be replicated.
	 */
	protected Collection<IDeployableComponentAdapter<C>> replicableComponents = new HashSet<>();

	/**
	 * Constructor that initializes {@code this} encoding with the collection of
	 * {@link IDeployableComponentAdapter}s as they are extracted from the input models.
	 */
	@SuppressWarnings("unchecked")
	public TaskGraphEncoding(SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;

		taskGraphs =
				(Collection<T>)systemModelAdapter.getTaskGraphs().stream().map(tg -> tg.copy())
						.collect(Collectors.toSet());

		replicableComponents.addAll(systemModelAdapter.getReplicableComponents());
		deployableComponents.addAll(systemModelAdapter.getDeployableComponents());
	}

	/**
	 * Constructor that initializes {@code this} encoding with the collection of
	 * {@link IDeployableComponentAdapter}s as they are extracted from the input models.
	 */
	public TaskGraphEncoding(SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter,
			Collection<T> taskGraphs) {
		this(systemModelAdapter);
		this.taskGraphs = taskGraphs;

		this.deployableComponents.clear();

		for(TaskGraph tg : taskGraphs) {
			@SuppressWarnings("unchecked") Collection<? extends IDeployableComponentAdapter<C>> tasks =
					(Collection<? extends IDeployableComponentAdapter<C>>)tg.getTasks();
			deployableComponents.addAll(tasks);
		}
	}

	/**
	 * Copy constructor that takes an existing {@link TaskGraphEncoding} as input and
	 * generates an equivalent copy.
	 */
	public TaskGraphEncoding(TaskGraphEncoding<G, C, T> encoding) {
		taskGraphs.addAll(encoding.getTaskGraphs());
		deployableComponents.addAll(encoding.deployableComponents);
		replicableComponents.addAll(encoding.replicableComponents);
	}

	/** Returns the collection of {@link TaskGraph}s composing {@code this} encoding. */
	public Collection<T> getTaskGraphs() {
		return taskGraphs;
	}

	/**
	 * Adds the given {@link IDeployableComponentAdapter} to its corresponding {@link TaskGraph} in
	 * {@code this} encoding.
	 */
	public void addTask(IDeployableComponentAdapter<C> task) {
		// TODO: Here, we use the assumption that the given task refers to a task that is also
		// present in the system model adapter.

		TaskGraph taskGraph = getGraphOf(task, taskGraphs);
		// Find the right Task Graph, in case the task does not exist in any of the task graphs of
		// this encoding.
		if(taskGraph == null) {
			taskGraph = findTGBySystemModel(systemModelAdapter.getTaskGraph(task));
		}

		// TODO: This should be transformed into an exception.
		assert (taskGraph != null) : "The task to be added could be neither found in this encoding, nor the input models.";

		taskGraph.addTask(task);
		deployableComponents.add(task);
		if(task.getMinReplication() != task.getMaxReplication()) {
			replicableComponents.add(task);
		}
	}

	/**
	 * Detects whether the {@link TaskGraph} of the input system model references any tasks of the
	 * internal {@link TaskGraph}s and returns the first matching {@link TaskGraph}.
	 */
	private TaskGraph findTGBySystemModel(TaskGraph systemTG) {
		for(TaskGraph currTG : taskGraphs) {
			if(currTG.containsAnyTaskOf(systemTG.getTasks())) {
				return currTG;
			}
		}
		return null;
	}

	/**
	 * Adds all signals defined in the input models for the given
	 * {@link IDeployableComponentAdapter} which must be contained in any of the internal
	 * {@link TaskGraph}s.
	 */
	public void addSignalsOf(IDeployableComponentAdapter<C> task) {
		// The given task must be in the any of the task graphs!

		T taskGraph = getGraphOf(task, taskGraphs);
		// TODO: This should be transformed into an exception.
		assert (taskGraph != null);

		TaskGraph systemTG = systemModelAdapter.getTaskGraph(task);
		Collection<ChannelAdapterWeightedEdge> incomingEdges =
				systemTG.getGraph().incomingEdgesOf(task);
		Collection<ChannelAdapterWeightedEdge> outgoingEdges =
				systemTG.getGraph().outgoingEdgesOf(task);

		for(ChannelAdapterWeightedEdge edge : incomingEdges) {
			IDeployableComponentAdapter<?> source = systemTG.getGraph().getEdgeSource(edge);
			taskGraph.getGraph().addEdge(source, task, edge);
		}

		for(ChannelAdapterWeightedEdge edge : outgoingEdges) {
			IDeployableComponentAdapter<?> target = systemTG.getGraph().getEdgeTarget(edge);
			taskGraph.getGraph().addEdge(task, target, edge);
		}
	}

	/**
	 * Adds the signals of a cloned task by querying the existing signals from the clone's original
	 * task.
	 */
	public void addClonedSignalsOf(IDeployableComponentAdapter<C> originalTask,
			IDeployableComponentAdapter<C> clonedTask) {
		// The given task must be in the any of the task graphs!

		T taskGraph = getGraphOf(originalTask, taskGraphs);
		// TODO: This should be transformed into an exception.
		assert (taskGraph != null);

		Collection<ChannelAdapterWeightedEdge> incomingEdges =
				taskGraph.getGraph().incomingEdgesOf(originalTask);
		Collection<ChannelAdapterWeightedEdge> outgoingEdges =
				taskGraph.getGraph().outgoingEdgesOf(originalTask);

		for(ChannelAdapterWeightedEdge edge : incomingEdges) {
			IDeployableComponentAdapter<?> source = taskGraph.getGraph().getEdgeSource(edge);
			taskGraph.getGraph().addEdge(source, clonedTask,
					new ChannelAdapterWeightedEdge(edge.getChannelAdapter()));
		}

		for(ChannelAdapterWeightedEdge edge : outgoingEdges) {
			IDeployableComponentAdapter<?> target = taskGraph.getGraph().getEdgeTarget(edge);
			taskGraph.getGraph().addEdge(clonedTask, target,
					new ChannelAdapterWeightedEdge(edge.getChannelAdapter()));
		}
	}

	/**
	 * Clones the given {@link IDeployableComponentAdapter} and adds it to the corresponding task
	 * graph and further data structures.
	 */
	public IDeployableComponentAdapter<C> cloneTask(IDeployableComponentAdapter<C> task) {
		// Clone the task and add it to the corresponding task graph.
		TaskGraph taskGraph = getGraphOf(task, taskGraphs);
		IDeployableComponentAdapter<C> clonedTask = task.clone();
		taskGraph.addTask(clonedTask);
		addClonedSignalsOf(task, clonedTask);
		// Add the cloned task to the cache sets.
		deployableComponents.add(clonedTask);
		if(replicableComponents.contains(task)) {
			replicableComponents.add(clonedTask);
		}

		return clonedTask;
	}

	/** Returns the {@link TaskGraph} containing the given task or {@code null} if none is found. */
	// TODO: Change the parameter again to IDeployableComponentAdapter<C>: pass the type parameter
	// to the scheduler.
	public T getTaskGraphOf(IDeployableComponentAdapter<?> task) {
		return getGraphOf(task, taskGraphs);
	}

	/** Whether {@code this} encoding contains the given task. */
	public boolean containsTask(IDeployableComponentAdapter<C> task) {
		return getGraphOf(task, taskGraphs) != null;
	}

	/**
	 * Removes the given {@link IDeployableComponentAdapter} from the {@link TaskGraph} containing
	 * the particular task and from the caches.
	 * <p>
	 * <i>Side-note:</i> Each task may be located in only one {@link TaskGraph}.
	 * 
	 * @returns {@code true} iff a containing {@link TaskGraph} has been found.
	 * */
	public boolean removeTask(IDeployableComponentAdapter<C> task) {
		TaskGraph taskGraph = getGraphOf(task, taskGraphs);
		if(taskGraph != null) {
			taskGraph.removeTask(task);
			deployableComponents.remove(task);
			replicableComponents.remove(task);
			return true;
		}
		assert (false); // Never here.
		return false;
	}

	/**
	 * Returns the deployable (software) component that are present in the internal
	 * {@link TaskGraph}s. May contain "abstract" {@link Component}s.
	 */
	public Collection<IDeployableComponentAdapter<?>> getDeployableComponents() {
		// return deployableComponents;
		Collection<IDeployableComponentAdapter<?>> deplComp = new HashSet<>();
		for(T tg : taskGraphs) {
			deplComp.addAll(tg.getTasks());
		}
		return deplComp;
	}

	// TODO: Consider implementing this method as a map, since this operation is run often?
	/**
	 * Returns the {@link IDeployableComponentAdapter} that is constructed form the given object.
	 * 
	 * @param originalEObject
	 *            The object for which the corresponding {@link IDeployableComponentAdapter} shall
	 *            be identified.
	 * @return The asociated {@link IDeployableComponentAdapter}, or {@code null} if none can be
	 *         identified.
	 */
	public synchronized IDeployableComponentAdapter<C> getDeployableComponentOf(
			EObject originalEObject) {
		for(IDeployableComponentAdapter<C> deployableComponentAdapter : deployableComponents) {
			if(deployableComponentAdapter.getObject() == originalEObject) {
				return deployableComponentAdapter;
			}
		}
		return null;
	}

	/**
	 * Returns the deployable (software) components that my be
	 * replicated. May contain "abstract" {@link Component}s.
	 */
	public Collection<IDeployableComponentAdapter<C>> getReplicableComponents() {
		return replicableComponents;
	}

	/**
	 * Returns the {@link IDeployableComponentAdapter} that wraps the given {@link EObject}. Here,
	 * only those {@link IDeployableComponentAdapter} are searched which can be actually executed on
	 * {@link ExecutionUnit}s, i.e. "abstract" components are not considered.
	 */
	public IDeployableComponentAdapter<C>
			getReplicableDeployableComponentOf(EObject originalEObject) {
		for(IDeployableComponentAdapter<C> deployableComponentAdapter : replicableComponents) {
			if(deployableComponentAdapter.getObject() == originalEObject) {
				return deployableComponentAdapter;
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Object clone() {
		return new TaskGraphEncoding<G, C, T>(this);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<? extends ComposablePhenotype<?>> getPTypeToRegister() {
		return (Class<? extends ComposablePhenotype<?>>)TaskGraphEncoding.class;
	}
}
