/*--------------------------------------------------------------------------+
$Id: AbstractTaskMappingEncoding.java 3885 2016-11-11 09:41:01Z diewald $
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping;

import java.util.HashMap;
import java.util.Map;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.SafeTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.opt4j.core.Genotype;

/**
 * {@link Genotype} which encodes an {@link TaskMappingEncoding} that may contain
 * {@link TaskMappingEntry}s which are required to be initialized.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3885 $
 * @ConQAT.Rating YELLOW Hash: 3DF39B0B8106F1038EB2EC0EF0E02D1E
 */
public class AbstractTaskMappingEncoding extends TaskMappingEncoding<AbstractTaskMappingEntry>
		implements ComposablePhenotype<AbstractTaskMappingEncoding> {

	/**
	 * Constructor that is initialized with a reference to the DSE specification and the input
	 * system model.
	 */
	AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			SafeTaskGraphEncoding<?> tgEncoding) {
		super(dse, systemModelAdapter, tgEncoding);
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			SafeTaskGraphEncoding<?> tgEncoding, boolean initialize) throws ExplorationException {
		super(dse, systemModelAdapter, tgEncoding, initialize);
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter, boolean initialize)
			throws ExplorationException {
		super(dse, systemModelAdapter, initialize);
	}

	/**
	 * Constructor that optionally initializes the encoding to a random encoding
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			AbstractTaskGraphEncoding<?, ?> atgEnc, boolean initialize) throws ExplorationException {
		super(dse, systemModelAdapter, atgEnc, initialize);
	}

	/**
	 * Constructor that generated a new {@link AbstractTaskMappingEncoding} preserving the
	 * allocations from the given {@code partialEncoding}.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	public AbstractTaskMappingEncoding(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			AbstractTaskGraphEncoding<?, ?> atgEnc, AbstractTaskMappingEncoding partialEncoding)
			throws ExplorationException {
		super(dse, systemModelAdapter, atgEnc, partialEncoding);
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEntry createEntry(IDeployableComponentAdapter<?> deployableComponent,
			IExecutionUnitAdapter<?> deploymentTarget) {
		// Request not needed to construct a {@link FailSilentTaskMappingEncoding}
		if(deploymentTarget == null) {
			return null;
		}

		return new AbstractTaskMappingEntry(deployableComponent, deploymentTarget);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public AbstractTaskMappingEncoding newInstance() {
		try {
			return new AbstractTaskMappingEncoding(dse, systemModelAdapter, true);
		} catch(ExplorationException e) {
			// Opt4J does not allow to throw Exceptions here.
			throw new RuntimeException(e);
		}
	}

	/** {@inheritDoc} */
	@Override
	public Object clone() {
		AbstractTaskMappingEncoding clonedEncoding;
		try {
			clonedEncoding = new AbstractTaskMappingEncoding(dse, systemModelAdapter, false);
		} catch(ExplorationException e) {
			// Opt4J does not allow to throw Exceptions here.
			throw new RuntimeException(e);
		}

		Map<AbstractTaskMappingEntry, AbstractTaskMappingEntry> copyAssocMap = new HashMap<>();

		clonedEncoding.allocationMap = copyAllocationMap(copyAssocMap);
		clonedEncoding.constraintMap = copyConstraintMap(copyAssocMap);
		clonedEncoding.getDeployedComponents().forEach(c -> c.updateInstatiationData());
		return clonedEncoding;
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends ComposableGenotype> getGTypeToRegister() {
		return AbstractTaskMappingEncoding.class;
	}

	/** {@inheritDoc} */
	@Override
	public Class<? extends ComposablePhenotype<?>> getPTypeToRegister() {
		return AbstractTaskMappingEncoding.class;
	}
}
