/*--------------------------------------------------------------------------+
$Id: InstantiatedTaskMappingEntry.java 3797 2016-10-28 11:14:06Z diewald $
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping;

import java.util.ArrayList;
import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.random.RandomExploration;

/**
 * {@link TaskMappingEntry} whose referenced {@link IDeployableComponentAdapter} only refers to
 * already instantiated {@link IDeployableComponentAdapter}s.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3797 $
 * @ConQAT.Rating YELLOW Hash: C54C99DE857B5AC431ECA573CCE888CF
 */
public abstract class InstantiatedTaskMappingEntry extends TaskMappingEntry implements Cloneable {

	/** Constructs a mapping entry the represents the mapping to a given deployment target. */
	public InstantiatedTaskMappingEntry(IDeployableComponentAdapter<?> deployableComponent,
			IExecutionUnitAdapter<?> deploymentTarget) {
		super(deployableComponent, deploymentTarget);
	}

	/**
	 * Constructor that generates a {@link InstantiatedTaskMappingEntry} of a
	 * {@link InstantiatedTaskMappingEncoding} for each {@link AbstractTaskMappingEntry} of a
	 * {@link AbstractTaskMappingEncoding}. Thereby, {@link AbstractTaskMappingEntry}s requiring
	 * instantiation will be replaced by a matching, randomly picked
	 * {@link IDeployableComponentAdapter}.
	 */
	public InstantiatedTaskMappingEntry(AbstractTaskMappingEntry abstractEntry) {
		this(abstractEntry.getDeployableComponent(), abstractEntry.getTarget());
		addAssociatedElement(abstractEntry);
		abstractEntry.addAssociatedElement(this);
		if(deploymentSource.isInstantiationRequired()) {
			IDeployableComponentAdapter<?> abstractComponent = deploymentSource;
			deploymentSource = pickRandomReplacementComponent(deploymentSource);
			deploymentSource.setReplacedComponent(abstractComponent);
		}
	}

	/**
	 * Iff the given {@link IDeployableComponentAdapter} must be instantiated/replaced by a
	 * "concrete" {@link IDeployableComponentAdapter}, this method returns a randomly picked and
	 * appropriate {@link IDeployableComponentAdapter}. The returned
	 * {@link IDeployableComponentAdapter} is picked from the set of replacement
	 * {@link IDeployableComponentAdapter}s of the given input {@link IDeployableComponentAdapter}.
	 * <p>
	 * {@code null} is returned if the given {@link IDeployableComponentAdapter} does not require
	 * instantiation or no "replacement components were defined.
	 */
	public IDeployableComponentAdapter<?> pickRandomReplacementComponent(
			IDeployableComponentAdapter<?> abstractComponent) {
		if(!deploymentSource.isInstantiationRequired()) {
			return null;
		}

		// TODO: parameterize the collection: cast to raw was used, but is no longer applicable,
		// since we have an EList here now.
		Collection<IDeployableComponentAdapter<?>> replacementComponents =
				new ArrayList<>(abstractComponent.getReplacementComponents());
		RandomExploration randomGenerator = RandomExploration.getInstance();
		int replacementBound = replacementComponents.size() - 1;
		int replacementIdx =
				replacementBound > 0 ? randomGenerator.nextInt(replacementComponents.size()) : 0;
		return ((IDeployableComponentAdapter<?>)replacementComponents.toArray()[replacementIdx])
				.clone();
	}
}
