/*--------------------------------------------------------------------------+
$Id: FlatPhenotypeMap.java 3478 2016-09-09 09:54:32Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype;

import java.util.Collection;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;

/**
 * Allows to capture the already created/decoded {@link Phenotype}s and pass them to their successor
 * creators/decoders as a flat map that associates a {@link Phenotype}s type with its instance.
 * Using a {@link CompositePhenotype} for this job allows to "protect" the internal map from
 * unintended read/write operations.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3478 $
 * @ConQAT.Rating YELLOW Hash: B568C33B362BE17DD68E29B6570BB8FA
 */
public class FlatPhenotypeMap<T extends Phenotype> extends ExplorationEncodingMap<T> {

	/** Constructor. */
	public FlatPhenotypeMap(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		super(systemModelAdapter);
	}

	/**
	 * Returns the {@link Phenotype} of the given type. If no {@link Phenotype} of the given type
	 * can be found, {@code null} is returned.
	 */
	@Override
	public <P extends T> P getEncoding(Class<P> subGenotypeClass) {
		P phenotype = getEncodingInternal(subGenotypeClass);
		if(phenotype == null) {
			if(IExplorationEncoding.class.isAssignableFrom(subGenotypeClass)) {
				@SuppressWarnings("unchecked") Class<IExplorationEncoding> inputEncoding =
						(Class<IExplorationEncoding>)subGenotypeClass;
				assert (systemModelAdapter != null) : "The SystemModelAdapter is required for CompositePhenotypes as an alternate source of input encodings.";
				@SuppressWarnings("unchecked") P encoding =
						(P)systemModelAdapter.getInputEncoding(inputEncoding);
				phenotype = encoding;
			}
		}

		return phenotype;
	}

	/**
	 * Determines whether {@code this} {@link FlatPhenotypeMap} contains {@link ComposablePhenotype}
	 * s of all given types, or it is able to retrieve the demanded types via the
	 * {@link SystemModelAdapter}.
	 */
	@SuppressWarnings("unchecked")
	public <P extends T> boolean containsAll(Collection<Class<?>> phenoCollection) {
		return phenoCollection.stream().allMatch(p -> contains((Class<P>)p));
	}

	/**
	 * Determines whether {@code this} {@link FlatPhenotypeMap} provides access to a
	 * {@link Phenotype} of the given type.
	 */
	public <P extends T> boolean contains(Class<P> phenotypeType) {
		return getEncoding(phenotypeType) != null;
	}

	/**
	 * Returns {@code this} {@link FlatPhenotypeMap} as an {@link ExplorationEncodingMap} such that
	 * its entries are represented by {@link IExplorationEncoding}s.
	 */
	@SuppressWarnings("unchecked")
	public ExplorationEncodingMap<IExplorationEncoding> asExplorationEncoding() {
		return (ExplorationEncodingMap<IExplorationEncoding>)this;
	}
}
