/*--------------------------------------------------------------------------+
$Id: MutateRedundancy.java 3747 2016-10-21 14:10:16Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.mutate.taskmapping;

import java.util.Collection;
import java.util.Optional;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalReplicationConstraint;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;

/**
 * Base class for mutate operator that manipulate the redundancy of allocations.
 * 
 * @author huang
 * @author $Author: diewald $
 * @version $Rev: 3747 $
 * @ConQAT.Rating RED Hash: 1E3C0D12B2649275ABC0CE3B663351A0
 */
public abstract class MutateRedundancy<S extends TaskMappingEntry, T extends TaskMappingEncoding<S>>
		extends MutateTasksBase<S, T> {

	/** Constructs a new redundancy mutation operator for a given exploration run */
	MutateRedundancy(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		super(systemModelAdapter);
	}

	/** {@inheritDoc} */
	@Override
	public void mutate(T encoding, double rate) {
		for(IDeployableComponentAdapter<?> deployableComponent : encoding.getDeployedComponents()) {
			if(rand.nextDouble() < rate) {
				// TODO: remove filter for fixed Deployment: It is not needed.
				// No replication for Components with a fixed Deployment & Components which shall
				// not be replicated (see the SystemModelAdapter).
				if(!encoding.hasFixedDeployment(deployableComponent) &&
						systemModelAdapter.getReplicableComponents().contains(deployableComponent)) {

					// TODO: Model minimum and maximum replication count as constraint?
					// Do the actual mutation of the allocation of the current request
					try {
						int minReplication = deployableComponent.getMinReplication();
						int maxReplication = deployableComponent.getMaxReplication();
						Collection<InternalReplicationConstraint> replConstraints =
								encoding.getConstraintsOf(InternalReplicationConstraint.class);
						Optional<InternalReplicationConstraint> curReplConstraint =
								replConstraints
										.stream()
										.filter(c -> c.getReplicationObject() == deployableComponent)
										.findAny();
						if(curReplConstraint.isPresent()) {
							minReplication = curReplConstraint.get().getMin();
							maxReplication = curReplConstraint.get().getMax();
						}

						mutate(encoding, deployableComponent, minReplication, maxReplication);
					} catch(ExplorationException e) {
						// Opt4J does not allow to throw Exceptions here.
						throw new RuntimeException(e);
					}
				}
				// TODO: What about temporal redundancy at the fixed resource
			}
		}
	}

	/**
	 * Performs the mutation of a single allocation
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	protected abstract void mutate(T encoding, IDeployableComponentAdapter<?> deployableComponent,
			int minReplicatation, int maxReplication) throws ExplorationException;
}
