/*--------------------------------------------------------------------------+
$Id: DseProblemModuleBase.java 3404 2016-08-26 15:01:08Z diewald $
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.commons.lang.IllegalClassException;
import org.fortiss.af3.exploration.alg.dse.TaskMappingFactory;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.CompositeDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.ComposableEvaluator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.CompositeEvaluator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.genotype.ComposableGenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.CompositePhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.problem.CompositeProblemModuleBase;
import org.fortiss.af3.exploration.alg.dse.evaluator.ArchitectureEvaluator;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationObjective;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.MOExplorationObjective;
import org.fortiss.af3.exploration.moea.model.DseSpecification;
import org.opt4j.core.Objective;
import org.opt4j.core.genotype.CompositeGenotype;
import org.opt4j.core.problem.Evaluator;

/**
 * Base class for problems that are solved within this DSE. It allows to create
 * {@link EvaluatorWrapper}s which associate {@link ExplorationTarget}s with {@link Evaluator}s of
 * possible solutions to the problem that shall be solved by the DSE.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3404 $
 * @ConQAT.Rating RED Hash: 40E95589D9A42B5800E037497E51E2F6
 */
public abstract class DseProblemModuleBase<C extends CompositeGenotype<Class<? extends ComposableGenotype>, ComposableGenotype>, P extends CompositePhenotype<?>, CD extends CompositeDecoder<C, P>>
		extends CompositeProblemModuleBase<C, P, CD> {

	/** DSE configuration */
	protected DseSpecification dse;

	/** Representation of input system */
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Set of requested {@link ComposablePhenotype}s that form the solution set. */
	Set<Class<? extends IExplorationEncoding>> requestedSolutions;

	/**
	 * Constructor the DSE specification and an adapter to the input system model has to be
	 * specified.
	 */
	public DseProblemModuleBase(DseSpecification dse,
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			Set<Class<? extends IExplorationEncoding>> requestedSolutions) {
		this.dse = dse;
		this.systemModelAdapter = systemModelAdapter;
		this.requestedSolutions = requestedSolutions;
	}

	/**
	 * Adds the evaluators of the sub problems of the architectural exploration.
	 * 
	 * @throws Exception
	 */
	protected <RT, T extends ExplorationTarget<RT>, S extends Phenotype, E extends Evaluator<?>>
			void addGoalEvaluators(CompositeEvaluator<P> dseEvaluator) throws Exception {
		Collection<ExplorationTarget<?>> targets = dse.getTargetSpecification().getTargets();
		Collection<ExplorationConstraint<?>> constraints =
				targets.stream().filter(t -> t instanceof ExplorationConstraint<?>)
						.map(ExplorationConstraint.class::cast).collect(Collectors.toList());

		int currentConstrPrio = 0;
		Map<Class<? extends ExplorationTarget<?>>, Integer> priorityMap = new LinkedHashMap<>();
		for(ExplorationConstraint<?> currConstraint : constraints) {
			if(!priorityMap.containsKey(currConstraint.getClass())) {
				@SuppressWarnings("unchecked") Class<? extends ExplorationTarget<?>> contraintType =
						(Class<? extends ExplorationTarget<?>>)currConstraint.getClass();
				priorityMap.put(contraintType, currentConstrPrio);
				currentConstrPrio++;
			}
		}

		int objectivePrioOffset = currentConstrPrio;
		Collection<ExplorationObjective<?>> objectives =
				targets.stream().filter(t -> t instanceof ExplorationObjective<?>)
						.map(ExplorationObjective.class::cast).collect(Collectors.toList());
		for(ExplorationObjective<?> currObjective : objectives) {
			if(MOExplorationObjective.class.isAssignableFrom(currObjective.getClass())) {
				if(!priorityMap.containsKey(currObjective.getClass())) {
					int objPrio = ((MOExplorationObjective<?>)currObjective).getPriority();
					@SuppressWarnings("unchecked") Class<? extends ExplorationTarget<?>> objectiveType =
							(Class<? extends ExplorationTarget<?>>)currObjective.getClass();
					priorityMap.put(objectiveType, objectivePrioOffset + objPrio);
				}
			} else {
				throw new IllegalClassException(
						"Objectives that are not MOExplorationObjectives are currently not supported by this backend.");
			}
		}

		for(ExplorationTarget<?> target : targets) {
			@SuppressWarnings({"cast", "unchecked"}) EvaluatorWrapper<?, RT, T> evalWrapper =
					(EvaluatorWrapper<?, RT, T>)addGoalEvaluator(dseEvaluator, (T)target,
							priorityMap.get(target.getClass()));
			Objective evalObjective = evalWrapper.getObjective();
			targetObjectiveAssoc.put(target, evalObjective);
		}
	}

	/**
	 * Adds an goal-specific {@link EvaluatorWrapper} to the {@link ArchitectureEvaluator}.
	 * 
	 * @throws Exception
	 */
	protected
			<RT, T extends ExplorationTarget<RT>, S extends ComposablePhenotype<?>, E extends ComposableEvaluator<S>>
			EvaluatorWrapper<?, RT, T> addGoalEvaluator(CompositeEvaluator<P> dseEvaluator, T goal,
					int priority) throws Exception {
		@SuppressWarnings("unchecked") EvaluatorWrapper<?, RT, T> wrappedEvaluator =
				(EvaluatorWrapper<?, RT, T>)TaskMappingFactory.getInstance()
						.createEvaluatorWrapper(goal, priority, systemModelAdapter);
		if(wrappedEvaluator != null) {
			// Here, we check whether the goal evaluator is defined for composable phenotypes (only
			// composable evaluators are allowed then).
			if(ComposablePhenotype.class
					.isAssignableFrom(wrappedEvaluator.getTargetPhenotypeType())) {

				Class<?> phenotypeClass = wrappedEvaluator.getTargetPhenotypeType();

				if(ComposablePhenotype.class.isAssignableFrom(phenotypeClass)) {
					// if(phenoClass.isAssignableFrom(phenotypeClass) &&
					// evalClass.isAssignableFrom(wrappedEvaluator.getClass())) {
					@SuppressWarnings("unchecked") Class<S> subPhenotypeClass =
							(Class<S>)phenotypeClass;
					@SuppressWarnings("unchecked") E wrappedSubEvaluator = (E)wrappedEvaluator;

					try {
						dseEvaluator.registerSubEvaluator(subPhenotypeClass, wrappedSubEvaluator);
					} catch(Exception e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
					// Register the wrapped Evaluator,
					registerEvaluator(wrappedEvaluator);
					// }
					// else {
					// throw new IllegalArgumentException("The sub-phenotye class " +
					// phenotypeClass + " and the evaluator class " +
					// wrappedEvaluator.getClass() + " does not match the classes " +
					// phenoClass + " and " + evalClass + " used to describe the problem.");
					// }
				} else {
					throw new Exception();
				}
			} else if(!(ComposablePhenotype.class.isAssignableFrom(wrappedEvaluator
					.getTargetPhenotypeType()))) {
				// Otherwise, a "simple" goal evaluator is created
				addEvaluator(wrappedEvaluator);
			} else {
				throw new Exception();
			}
		} else {
			// TODO: Evaluate, if the optimization should stop here, i.e. throw an exception or
			// similar (if possible).
			System.out.println("Could not instantiate an evaluator for goal of dimension \"" +
					goal.getName() + "\". Skipping.");
			return null;
		}

		return wrappedEvaluator;
	}
}
