/*--------------------------------------------------------------------------+
$Id: EvaluatorWrapper.java 3730 2016-10-19 15:43:43Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem;

import org.conqat.lib.commons.reflect.TypeConversionException;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.evaluator.ArchExpSubEvaluator;
import org.fortiss.af3.exploration.alg.dse.evaluator.ExplorationTargetEvaluatorBase;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.dsl.model.MOEAExplorationExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.Expression;
import org.fortiss.af3.exploration.dsl.model.operators.OptimizationDirection;
import org.fortiss.af3.exploration.dsl.model.operators.OptimizationDirectionEnum;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.opt4j.core.Objective;
import org.opt4j.core.Objective.Sign;
import org.opt4j.core.Objectives;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.ClassToInstanceMap;

/**
 * Opt4J specific goal evaluator wrapper that delegates the actual evaluation to a strategy object
 * of type {@link ExplorationTargetEvaluatorBase}
 * 
 * @author huang
 * @author $Author: diewald $
 * @version $Rev: 3730 $
 * @ConQAT.Rating RED Hash: 3C64B8316B0C8468C8FBD3CDF6941F91
 */
public class EvaluatorWrapper<P, RT, T extends ExplorationTarget<RT>> implements
		ArchExpSubEvaluator<P>, Evaluator<P> {

	/** Objective to track evaluation of this {@link ExplorationTarget} */
	private Objective objective;

	/** Actual underlying evaluator */
	private IExplorationTargetEvaluator<P, RT, T> evaluator;

	/** Goal specification */
	private T goal;

	/** Contains the priority of the {@link ExplorationTarget} of the {@link EvaluatorWrapper}. */
	private int targetPriority;

	/** Phenotypes, i.e. (sub-)solutions, to which this operator may be applied to. */
	private Class<P> targetPhenotypeType;

	/**
	 * Constructor that creates an evaluator wrapper for a given {@link ExplorationTarget}
	 * 
	 * @throws TypeConversionException
	 */
	public EvaluatorWrapper(T goal, IExplorationTargetEvaluator<P, RT, T> evaluator, int priority,
			Class<? extends P> primaryInputType) throws TypeConversionException {

		// Construct EA objectives and register it in objective map. Do not give an
		// "optimization direction" (min/max) in case of constraints. They default to minimization.
		OptimizationDirectionEnum direction = OptimizationDirectionEnum.MIN;
		if(goal instanceof MOEAExplorationExpression) {
			Expression moExpr = ((MOEAExplorationExpression)goal).getExpression();
			if(moExpr instanceof OptimizationDirection) {
				direction = ((OptimizationDirection)moExpr).getOperator();
			}
		}
		String objectiveName =
				goal.getName() == null ? goal.getClass().toString() +
						Integer.toString(goal.getId()) : goal.getName();
		objective =
				new Objective(objectiveName, direction == OptimizationDirectionEnum.MIN ? Sign.MIN
						: Sign.MAX);

		// TODO: modify EvaluatorWrapper to indicate the exception below (allow handling of the
		// "validation exception in the UI"
		if(!evaluator.validateGoalSpecification()) {
			throw new RuntimeException("Error. The optimization goal " + goal.getName() +
					" failed the validation. Please re-check this goal.");
		}

		this.goal = goal;
		this.evaluator = evaluator;
		this.targetPriority = priority;
		@SuppressWarnings("unchecked") Class<P> targetPhenotypeType = (Class<P>)primaryInputType;
		this.targetPhenotypeType = targetPhenotypeType;
	}

	/**
	 * Returns the priority of this evaluator (based on the priority specified for the underlying
	 * goal)
	 */
	public int getPriority() {
		return targetPriority;
	}

	/** {@inheritDoc} */
	@Override
	public <CP extends ComposablePhenotype<?>> Objectives evaluateComposable(Class<CP> clazz,
			CP phenotype, FlatPhenotypeMap<Phenotype> phenotypes,
			ClassToInstanceMap<EObject> transformedModels) {
		Objectives objectives = new Objectives();

		@SuppressWarnings("unchecked") P castedPhenotype = (P)phenotype;
		objectives.add(objective,
				evaluator.evaluateGoal(castedPhenotype, phenotypes, transformedModels));

		return objectives;
	}

	/** {@inheritDoc} */
	@Override
	public Objectives evaluate(P phenotype) {
		// TODO: Investigate when this method is called. Should be never the case...
		assert (false);
		Objectives objectives = new Objectives();

		objectives.add(objective, evaluator.evaluateGoal(phenotype, null, null));

		return objectives;
	}

	/** Returns the type of the goal this evaluator is built for. */
	@SuppressWarnings("unchecked")
	public Class<T> getTargetType() {
		return (Class<T>)goal.getClass();
	}

	/** Returns the underlying (actual) evaluator. */
	public IExplorationTargetEvaluator<P, RT, T> getEvaluator() {
		return evaluator;
	}

	/** Returns the objective associated with the wrapped {@link Evaluator}. */
	public Objective getObjective() {
		return objective;
	}

	/** Returns the class of the phenotype to */
	public Class<P> getTargetPhenotypeType() {
		return targetPhenotypeType;
	}
}
