/*--------------------------------------------------------------------------+
$Id: InstantiatedTaskMappingDecoder.java 3841 2016-11-04 14:51:29Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.instantiatetaskgraph;

import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.castToRawColl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.stream.Collectors;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoderBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ChannelAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.DefaultTaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraphInstantiator;

import com.google.inject.Inject;

/**
 * {@link ComposableDecoder} that takes an {@link InstantiatedTaskMappingEncoding} as input and
 * generates an {@link InstantiatedTaskGraphEncoding} from it.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3841 $
 * @ConQAT.Rating RED Hash: 2CABBE8D1AD73618A728B6B80A45D73C
 */
public class InstantiatedTaskMappingDecoder<T extends InstantiatedTaskMappingEntry, C>
		extends
		ComposableDecoderBase<InstantiatedTaskMappingEncoding<T>, InstantiatedTaskGraphEncoding<T, C>> {

	/** Representation of input system model */
	private SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Constructs a new decoder to generate {@link InstantiatedTaskGraphEncoding}. */
	@Inject
	public InstantiatedTaskMappingDecoder(SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;
	}

	/** {@inheritDoc} */
	@Override
	public InstantiatedTaskGraphEncoding<T, C> decodeComposable(
			InstantiatedTaskMappingEncoding<T> genotype,
			FlatPhenotypeMap<Phenotype> phenotypeTypeMap) throws DecodingException {
		@SuppressWarnings("unchecked") AbstractTaskGraphEncoding<AbstractTaskMappingEncoding, C> atgEnc =
				phenotypeTypeMap.getEncoding(AbstractTaskGraphEncoding.class);
		assert (atgEnc != null) : "The required abstract task graph encoding could not be found in the map of already decoded phenotypes.";

		SafetyFunctionArchEncoding<?> sfEnc =
				phenotypeTypeMap.getEncoding(SafetyFunctionArchEncoding.class);

		Collection<DefaultTaskGraph> tgColl = atgEnc.getTaskGraphs();
		TaskGraphInstantiator<DefaultTaskGraph> tgInstatiator =
				new TaskGraphInstantiator<>(tgColl, genotype);
		InstantiatedTaskGraphEncoding<T, C> itgEnc =
				new InstantiatedTaskGraphEncoding<T, C>(systemModelAdapter,
						tgInstatiator.createMultipleInstanceTaskGraphs(tgColl));

		if(sfEnc != null) {
			removeInterSafetyChannelSignals(sfEnc, itgEnc);
		}

		return itgEnc;
	}

	/**
	 * Removes signals from the {@link InstantiatedTaskGraphEncoding} that were introduced by
	 * replications but which connect independent safety channels. This is not desired, so remove
	 * all edges (signals) in the task graph that induce connections between the safety channels.
	 */
	// FIXME: also consider other realizing components.
	// TODO: consider moving this operation
	private void removeInterSafetyChannelSignals(SafetyFunctionArchEncoding<?> sfEnc,
			InstantiatedTaskGraphEncoding<T, C> itgEnc) throws DecodingException {
		for(SafetyFunctionArchEntry<?> sfEntry : sfEnc.getAllEntries()) {
			Collection<IDeployableComponentAdapter<?>> sfCompAdps =
					castToRawColl(sfEntry.getSafetyFunctionAdapter().getRealizingComponents());
			assert (sfCompAdps.size() > 0);
			IDeployableComponentAdapter<?> sfCompAdp = sfCompAdps.stream().findAny().get();

			// Get the "containing" task graph. We must check all replacement components (instances)
			// since we don't know which one was instantiated (or it would be more costly to find
			// out).
			DefaultTaskGraph sfTG = null;
			for(IDeployableComponentAdapter<?> curSFCompAdp : sfCompAdp.getReplacementComponents()) {
				sfTG = itgEnc.getTaskGraphOf(curSFCompAdp);
				if(sfTG != null) {
					break;
				}
			}
			if(sfTG == null) {
				throw new DecodingException(sfEnc, this, "Could not retrieve the TaskGraph" +
						" of the Component Adapter " + sfCompAdp.getName() + ".");
			}

			// TODO: Remove non-allowed communication using the commented out code below to obtain
			// a more consistent and generic solution.
			// for(InternalIsolatedCommunicationSet isoCommSet : genotype
			// .getConstraintsOf(InternalIsolatedCommunicationSet.class)) {
			// Collection<IMappingEntry> entries = isoCommSet.getIsolatedCommunicationSet();
			// Collection<IDeployableComponentAdapter<?>> compAdps =
			// map(entries, e -> e.getDeployableComponent());
			// for(IMappingEntry entry : entries) {
			// IMappingEntry instEntry = entry.getAssociatedElement().stream().findAny().get();
			// IDeployableComponentAdapter<?> currComp = instEntry.getDeployableComponent();
			// Collection<InternalIsolatedCommunicationSet> others =
			// getOtherIsolationCommConstraints(genotype, entry);
			// Collection<IDeployableComponentAdapter<?>> connComps =
			// Graphs.successorListOf(sfTG.getGraph(), currComp);
			// for(IDeployableComponentAdapter<?> connCompAdp : connComps) {
			// if(isPartOfOtherIsolationCommConstraint(others, connCompAdp)) {
			// sfTG.getGraph().removeEdge(currComp, connCompAdp);
			// }
			// }
			//
			// connComps = Graphs.predecessorListOf(sfTG.getGraph(), currComp);
			// for(IDeployableComponentAdapter<?> connCompAdp : connComps) {
			// if(isPartOfOtherIsolationCommConstraint(others, connCompAdp)) {
			// sfTG.getGraph().removeEdge(connCompAdp, currComp);
			// }
			// }
			// }
			// }

			Collection<IDeployableComponentAdapter<?>> connectedSFUnits = new ArrayList<>();
			Collection<IDeployableComponentAdapter<?>> connectedDiagUnits = new ArrayList<>();
			Collection<IDeployableComponentAdapter<?>> connectedOutUnit = new ArrayList<>();
			for(int channelIdx = 0; channelIdx < sfEntry.getNumChannels(); channelIdx++) {
				Collection<IDeployableComponentAdapter<?>> sfUnits =
						castToRawColl(sfEntry.getSafetyFunctionAdapter().getRealizingComponents());
				Collection<IDeployableComponentAdapter<?>> diagUnits =
						castToRawColl(sfEntry.getSafetyFunctionAdapter().getDiagnosticComponents());
				// TODO: knowledge about the modeL: output unit not instantiated:
				IDeployableComponentAdapter<?> outputUnit =
						sfEntry.getSafetyFunctionAdapter().getOutputComponent();
				// TODO: only for a single diag unit.
				Collection<IDeployableComponentAdapter<?>> realizingSFUnits =
						itgEnc.getDeployableComponents().stream()
								.filter(d -> sfUnits.contains(d.getReplacedComponent()))
								.collect(Collectors.toList());
				Collection<IDeployableComponentAdapter<?>> realizingDiagUnits =
						itgEnc.getDeployableComponents().stream()
								.filter(d -> diagUnits.contains(d.getReplacedComponent()))
								.collect(Collectors.toList());
				Collection<IDeployableComponentAdapter<?>> outComps =
						itgEnc.getDeployableComponents().stream()
								.filter(d -> d.getObject() == outputUnit.getObject())
								.collect(Collectors.toList());

				// Pick an sf component instance that is assigned to the current channel.
				IDeployableComponentAdapter<?> currentChannelSFUnit = null;
				for(IDeployableComponentAdapter<?> curSFUnit : realizingSFUnits) {
					if(!connectedSFUnits.contains(curSFUnit)) {
						// Remember that the diagnostic unit is associated with one channel --> no
						// elimination of the signal.
						connectedSFUnits.add(curSFUnit);
						currentChannelSFUnit = curSFUnit;
						break;
					}
				}
				// Pick an diagnostic component instance that is assigned to the current channel.
				IDeployableComponentAdapter<?> currentChannelDiagUnit = null;
				for(IDeployableComponentAdapter<?> curDiagUnit : realizingDiagUnits) {
					if(!connectedDiagUnits.contains(curDiagUnit)) {
						// Remember that the diagnostic unit is associated with one channel --> no
						// elimination of the signal.
						connectedDiagUnits.add(curDiagUnit);
						currentChannelDiagUnit = curDiagUnit;
						break;
					}
				}
				// Pick an io component instance that is assigned to the current channel.
				IDeployableComponentAdapter<?> currentChannelOutputUnit = null;
				for(IDeployableComponentAdapter<?> curOutUnit : outComps) {
					if(!connectedOutUnit.contains(curOutUnit)) {
						// Remember that the diagnostic unit is associated with one channel --> no
						// elimination of the signal.
						connectedDiagUnits.add(curOutUnit);
						currentChannelOutputUnit = curOutUnit;
						break;
					}
				}

				for(IDeployableComponentAdapter<?> curSFUnit : realizingSFUnits) {
					if(currentChannelSFUnit != curSFUnit) {
						// Eliminate all other signals to different channels.
						ChannelAdapterWeightedEdge interChannelSignal =
								sfTG.getGraph().getEdge(currentChannelDiagUnit, curSFUnit);
						sfTG.getGraph().removeEdge(interChannelSignal);
						interChannelSignal =
								sfTG.getGraph().getEdge(currentChannelOutputUnit, curSFUnit);
						sfTG.getGraph().removeEdge(interChannelSignal);
					}
				}
				for(IDeployableComponentAdapter<?> curDiagUnit : realizingDiagUnits) {
					if(currentChannelDiagUnit != curDiagUnit) {
						// Eliminate all other signals to different channels.
						ChannelAdapterWeightedEdge interChannelSignal =
								sfTG.getGraph().getEdge(currentChannelSFUnit, curDiagUnit);
						sfTG.getGraph().removeEdge(interChannelSignal);
						interChannelSignal =
								sfTG.getGraph().getEdge(currentChannelOutputUnit, curDiagUnit);
						sfTG.getGraph().removeEdge(interChannelSignal);
					}
				}
				for(IDeployableComponentAdapter<?> curOutUnit : outComps) {
					if(currentChannelOutputUnit != curOutUnit) {
						// Eliminate all other signals to different channels.
						ChannelAdapterWeightedEdge interChannelSignal =
								sfTG.getGraph().getEdge(currentChannelSFUnit, curOutUnit);
						sfTG.getGraph().removeEdge(interChannelSignal);
						interChannelSignal =
								sfTG.getGraph().getEdge(currentChannelDiagUnit, curOutUnit);
						sfTG.getGraph().removeEdge(interChannelSignal);
					}
				}
			}
		}
	}

	// TODO: Remove non-allowed communication using the commented out code above and below to obtain
	// a more consistent and generic solution.
	// private Collection<InternalIsolatedCommunicationSet> getOtherIsolationCommConstraints(
	// TaskMappingEncoding<?> encoding, IMappingEntry entry) {
	// // TODO: here we use the knowledge that we have instantiated task mapping entries.
	// IMappingEntry actualEntry = entry.getAssociatedElement(AbstractTaskMappingEntry.class);
	//
	// Collection<InternalIsolatedCommunicationSet> otherConstraints = new HashSet<>();
	// Collection<InternalIsolatedCommunicationSet> isoComm =
	// encoding.getConstraintsOf(InternalIsolatedCommunicationSet.class);
	// for(InternalIsolatedCommunicationSet constr : isoComm) {
	// if(!constr.getIsolatedCommunicationSet().contains(actualEntry)) {
	// otherConstraints.add(constr);
	// }
	// }
	//
	// return otherConstraints;
	// }
	//
	// private boolean isPartOfOtherIsolationCommConstraint(
	// Collection<InternalIsolatedCommunicationSet> other,
	// IDeployableComponentAdapter<?> compAdp) {
	// for(InternalIsolatedCommunicationSet constr : other) {
	// for(IMappingEntry entry : constr.getIsolatedCommunicationSet()) {
	// if(entry.getAssociatedElement().stream().findAny().get().getDeployableComponent() == compAdp)
	// {
	// return true;
	// }
	// }
	// }
	//
	// return false;
	// }

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 *             see {@link ComposableDecoder}.
	 */
	@Override
	public InstantiatedTaskGraphEncoding<T, C> updateComposable(
			InstantiatedTaskMappingEncoding<T> inGenotype,
			InstantiatedTaskGraphEncoding<T, C> outGenotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException {
		// Not required: The InstantiatedTaskGraphEncoding is a Phenotype only.
		// TODO: Throw a decoding exception, instead.
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> reqTypes = new ArrayList<>();
		reqTypes.add(InstantiatedTaskMappingEncoding.class);
		reqTypes.add(AbstractTaskGraphEncoding.class);
		return reqTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		Collection<Class<?>> optTypes = new ArrayList<>();
		optTypes.add(SafetyFunctionArchEncoding.class);
		return optTypes;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("rawtypes")
	@Override
	public Class<InstantiatedTaskGraphEncoding> getOutputType() {
		return InstantiatedTaskGraphEncoding.class;
	}
}
