/*--------------------------------------------------------------------------+
$Id: SFGraphDecoder.java 3478 2016-09-09 09:54:32Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.SafeTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.TaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoderBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.CompositeDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.ArchExpSubDecoder;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;

import com.google.inject.Inject;

/**
 * {@link ComposableDecoder} that generates {@link IDeployableComponentAdapter}s from the
 * chosen architecture of a safety function.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3478 $
 * @ConQAT.Rating RED Hash: 45A6D4C10AD00BF2D6ED61F9117FE40C
 */
public class SFGraphDecoder<C> extends
		ComposableDecoderBase<SafetyFunctionArchEncoding<C>, SafeTaskGraphEncoding<C>> implements
		ArchExpSubDecoder {

	/** Reference to the {@link SystemModelAdapter} that handles input models for the DSE. */
	SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Constructs a new decoder that instantiates {@link AbstractTaskMappingEncoding}. */
	@Inject
	public SFGraphDecoder(SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;
	}

	/** {@inheritDoc} */
	@Override
	public SafeTaskGraphEncoding<C> decodeComposable(SafetyFunctionArchEncoding<C> genotype,
			FlatPhenotypeMap<Phenotype> phenotypetypeMap) {
		if(genotype == null) {
			return null;
		}
		// TODO: we should dynamically retrieve the encoding here, since the statement below assumes
		// that this decoding step is performed before others.
		SafeTaskGraphEncoding<C> encoding = new SafeTaskGraphEncoding<>(systemModelAdapter);

		for(SafetyFunctionArchEntry<C> sFAEntry : genotype.getAllEntries()) {
			boolean diagPresent = sFAEntry.isDiagnosicUnitPresent();
			if(!diagPresent && containsDiagUnits(sFAEntry, encoding)) {
				removeDiagUnits(sFAEntry, encoding);
			} else if(diagPresent && !containsDiagUnits(sFAEntry, encoding)) {
				addDiagUnits(sFAEntry, encoding);
			}
		}

		return encoding;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 *             see {@link CompositeDecoder}.
	 */
	@Override
	public SafeTaskGraphEncoding<C> updateComposable(SafetyFunctionArchEncoding<C> inGenotype,
			SafeTaskGraphEncoding<C> outGenotype, FlatPhenotypeMap<Phenotype> decodedPhenotypes)
			throws DecodingException {
		// Not required: The abstract task graph encoding is a Phenotype only.
		// TODO: Throw a decoding exception, instead.
		return null;
	}

	/**
	 * Checks whether the given {@link TaskGraphEncoding} contains diagnostic units from the given
	 * {@link SafetyFunctionArchEntry}.
	 */
	private boolean containsDiagUnits(SafetyFunctionArchEntry<C> sFAEntry,
			SafeTaskGraphEncoding<C> encoding) {
		Collection<IDeployableComponentAdapter<C>> sfComps =
				sFAEntry.getSafetyFunctionAdapter().getDiagnosticComponents();
		if(!sfComps.isEmpty()) {
			return encoding.containsTask(sfComps.stream().findAny().get());
		}
		return false;
	}

	/**
	 * Adds the diagnostic units from the {@link SafetyFunctionArchEntry} to the
	 * {@link TaskGraphEncoding}.
	 */
	private void
			addDiagUnits(SafetyFunctionArchEntry<C> sFAEntry, SafeTaskGraphEncoding<C> encoding) {
		Collection<IDeployableComponentAdapter<C>> diagComps =
				sFAEntry.getSafetyFunctionAdapter().getDiagnosticComponents();
		addTasks(diagComps, encoding);
	}

	/**
	 * Adds the given collection of {@link IDeployableComponentAdapter}s from the
	 * {@link SafetyFunctionArchEntry} to the {@link TaskGraphEncoding}.
	 */
	private void addTasks(Collection<IDeployableComponentAdapter<C>> deplComp,
			SafeTaskGraphEncoding<C> encoding) {
		// Add the Components.
		deplComp.stream().forEach(task -> encoding.addTask(task));

		// Add the corresponding signals.
		deplComp.stream().forEach(task -> encoding.addSignalsOf(task));
	}

	/** Removes the diagnostic units from the {@link TaskGraphEncoding}. */
	private void removeDiagUnits(SafetyFunctionArchEntry<C> sFAEntry,
			SafeTaskGraphEncoding<C> encoding) {
		Collection<IDeployableComponentAdapter<C>> sfComps =
				sFAEntry.getSafetyFunctionAdapter().getDiagnosticComponents();
		sfComps.stream().forEach(t -> encoding.removeTask(t));
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> reqTypes = new ArrayList<>();
		reqTypes.add(SafetyFunctionArchEncoding.class);
		return reqTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@SuppressWarnings("rawtypes")
	@Override
	public Class<AbstractTaskGraphEncoding> getOutputType() {
		return AbstractTaskGraphEncoding.class;
	}
}
