/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalReplicationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.constraint.InternalSeparationConstraint;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoderBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.decode.ConstraintGenerationDecoder;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISafetyFunctionAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;

import com.google.inject.Inject;

/**
 * Generates allocation and replication constraints for the safety architectures of the safety
 * function of a given {@link SafetyFunctionArchEncoding}.
 * 
 * @author diewald
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash: 73B91D74FB8917858F69A78E9B98CE51
 */
public class SFMappingConstraintDecoder<C> extends
		ComposableDecoderBase<SafetyFunctionArchEncoding<C>, AbstractTaskMappingEncoding> implements
		ConstraintGenerationDecoder {

	// TODO: replace with a dynamic mechanism for the retrieval of deployable components.
	/** Reference to the {@link SystemModelAdapter} that handles input models for the DSE. */
	private SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Constructs a new decoder that instantiates {@link AbstractTaskMappingEncoding}. */
	@Inject
	public SFMappingConstraintDecoder(SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding decodeComposable(SafetyFunctionArchEncoding<C> genotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public AbstractTaskMappingEncoding updateComposable(SafetyFunctionArchEncoding<C> inGenotype,
			AbstractTaskMappingEncoding outGenotype, FlatPhenotypeMap<Phenotype> decodedPhenotypes) {
		Collection<IDeployableComponentAdapter<C>> deplComps = new HashSet<>();
		deplComps.addAll(systemModelAdapter.getDeployableComponents());
		for(SafetyFunctionArchEntry<C> sFAEntry : inGenotype.getAllEntries()) {
			int channelNumber = sFAEntry.getNumChannels();

			Collection<InternalReplicationConstraint> replicationConstraints = new HashSet<>();
			for(IDeployableComponentAdapter<?> comp : sFAEntry.getSafetyFunctionAdapter()
					.getAllComponents()) {
				replicationConstraints.add(new InternalReplicationConstraint(comp, channelNumber,
						channelNumber));
			}

			outGenotype.replaceConstraints(InternalReplicationConstraint.class,
					replicationConstraints);
		}

		outGenotype.replaceConstraints(InternalSeparationConstraint.class,
				createSeparationConstraints(inGenotype));

		return outGenotype;
	}

	/**
	 * Creates {@link InternalSeparationConstraint}s for the sets of realizing, diagnostic, and
	 * output units of each safety function.
	 */
	private Collection<InternalSeparationConstraint> createSeparationConstraints(
			SafetyFunctionArchEncoding<C> sfEnc) {
		Collection<InternalSeparationConstraint> sepConstraints = new HashSet<>();

		for(SafetyFunctionArchEntry<C> sfEntry : sfEnc.getAllEntries()) {
			ISafetyFunctionAdapter<C> sfAdp = sfEntry.getSafetyFunctionAdapter();

			// TODO: We should not be aware of the Tile type here.
			sepConstraints.add(new InternalSeparationConstraint(sfAdp.getRealizingComponents(),
					systemModelAdapter.getFaultContainmentRegion()));
			sepConstraints.add(new InternalSeparationConstraint(sfAdp.getDiagnosticComponents(),
					systemModelAdapter.getFaultContainmentRegion()));
			sepConstraints.add(new InternalSeparationConstraint(Arrays.asList(sfAdp
					.getOutputComponent()), systemModelAdapter.getFaultContainmentRegion()));
		}

		return sepConstraints;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> requiredTypes = new HashSet<>();
		requiredTypes.add(SafetyFunctionArchEncoding.class);
		return requiredTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Class<? super AbstractTaskMappingEncoding> getOutputType() {
		return AbstractTaskMappingEncoding.class;
	}

}
