/*--------------------------------------------------------------------------+
$Id: SFMappingDecoder.java 3964 2016-12-12 12:55:09Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.safetyfunction;

import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.comm.constraint.InternalIsolatedCommunicationSet;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.safetyfunctionarch.SafetyFunctionArchEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.AbstractTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.abstractmapping.AbstractTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoder;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.decode.ComposableDecoderBase;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.exception.DecodingException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.decode.ConstraintGenerationDecoder;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;

import com.google.inject.Inject;

/**
 * Decodes a {@link SafetyFunctionArchEncoding} into an {@link AbstractTaskMappingEncoding}.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3964 $
 * @ConQAT.Rating RED Hash: 7126BAD300304543E77EE1F1F430A2C9
 */
public class SFMappingDecoder<C> extends
		ComposableDecoderBase<SafetyFunctionArchEncoding<C>, AbstractTaskMappingEncoding> implements
		ConstraintGenerationDecoder {

	// TODO: replace with a dynamic mechanism for the retrieval of deployable components.
	/** Reference to the {@link SystemModelAdapter} that handles input models for the DSE. */
	private SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Constructs a new decoder that instantiates {@link AbstractTaskMappingEncoding}. */
	@Inject
	public SFMappingDecoder(SystemModelAdapter<C, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 *             see {@link ComposableDecoder}.
	 */
	@Override
	public AbstractTaskMappingEncoding decodeComposable(SafetyFunctionArchEncoding<C> genotype,
			FlatPhenotypeMap<Phenotype> decodedPhenotypes) throws DecodingException {
		AbstractTaskMappingEncoding atmEnc =
				decodedPhenotypes.getEncoding(AbstractTaskMappingEncoding.class);
		assert (atmEnc != null) : "The required abstract task mapping encoding could not be found in the map of already decoded phenotypes.";
		return updateComposable(genotype, atmEnc, decodedPhenotypes);
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws DecodingException
	 *             see {@link ComposableDecoder}.
	 */
	@Override
	public AbstractTaskMappingEncoding updateComposable(SafetyFunctionArchEncoding<C> inGenotype,
			AbstractTaskMappingEncoding outGenotype, FlatPhenotypeMap<Phenotype> decodedPhenotypes)
			throws DecodingException {
		if(inGenotype == null) {
			return null;
		}

		// TODO: replace with a dynamic mechanism for the retrieval of deployable components.
		Collection<IDeployableComponentAdapter<C>> deplComps = new HashSet<>();
		deplComps.addAll(systemModelAdapter.getDeployableComponents());
		for(SafetyFunctionArchEntry<C> sFAEntry : inGenotype.getAllEntries()) {
			boolean diagPresent = sFAEntry.isDiagnosicUnitPresent();
			int channelNumber = sFAEntry.getNumChannels();

			Collection<IDeployableComponentAdapter<C>> sfComps =
					sFAEntry.getSafetyFunctionAdapter().getRealizingComponents();
			Collection<IDeployableComponentAdapter<C>> diagComps =
					sFAEntry.getSafetyFunctionAdapter().getDiagnosticComponents();
			IDeployableComponentAdapter<C> outComponent =
					sFAEntry.getSafetyFunctionAdapter().getOutputComponent();

			// We assume that the entry number must be equal for all realizing components of a
			// safety function.
			Collection<AbstractTaskMappingEntry> sfEntries =
					outGenotype.getMappingEntriesOf(sfComps.stream().findAny().get());
			int entryNumber = sfEntries.size();

			// TODO: use a more generic solution to instantiate these tyoes of constraints during
			// the creation phase.
			Set<IMappingEntry> existingEntries = new HashSet<>();
			if(outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class).isEmpty()) {
				for(IDeployableComponentAdapter<C> comp : sFAEntry.getSafetyFunctionAdapter()
						.getAllComponents()) {
					Collection<AbstractTaskMappingEntry> curEntry =
							outGenotype.getMappingEntriesOf(comp);
					assert (curEntry.size() <= 1);
					existingEntries.addAll(curEntry);
				}
				outGenotype.putConstraint(InternalIsolatedCommunicationSet.class,
						new InternalIsolatedCommunicationSet(existingEntries));
			}

			// In case that diagnostic units are not present in the *previous* encoding, we need to
			// add some more.
			Collection<AbstractTaskMappingEntry> diagEntries =
					outGenotype.getMappingEntriesOf(diagComps.stream().findAny().get());
			if(diagPresent && diagEntries != null && diagEntries.isEmpty()) {
				try {
					addMissingDiagUnits(outGenotype, diagComps, entryNumber);
				} catch(ExplorationException e) {
					throw new DecodingException(
							inGenotype,
							this,
							"Error adding diagnostic components to match the number of safety channels.",
							e);
				}
			}

			// Match the encoding with the selected channel number.
			for(int i = entryNumber; i < channelNumber; i++) {
				try {
					Set<IMappingEntry> createdEntries = new HashSet<>();
					createdEntries.addAll(addSFComponents(outGenotype, sfComps));
					if(diagPresent) {
						createdEntries.addAll(addSFComponents(outGenotype, diagComps));
					}
					// TODO: for each channel, the output component is "replicated" here.
					createdEntries
							.addAll(addSFComponents(outGenotype, Arrays.asList(outComponent)));

					outGenotype.putConstraint(InternalIsolatedCommunicationSet.class,
							new InternalIsolatedCommunicationSet(createdEntries));
				} catch(ExplorationException e) {
					throw new DecodingException(
							inGenotype,
							this,
							"Error adding safety components to match the number of safety channels.",
							null);
				}
			}
			List<InternalIsolatedCommunicationSet> isoCommSet =
					new ArrayList<>(
							outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class));
			for(int i = entryNumber - 1; i > channelNumber; i--) {
				Collection<IMappingEntry> rmEntries =
						isoCommSet.get(i).getIsolatedCommunicationSet();

				removeSFEntries(outGenotype, rmEntries);
				removeMatchingConstraint(outGenotype, getFirst(rmEntries, (s) -> true).orElse(null));
			}

			if(!diagPresent) {
				// If no diagnostic unit is present in the encoding, remove its entire task.
				removeDiagUnits(outGenotype, diagComps);
			}
		}

		return outGenotype;
	}

	/**
	 * Adds entries for diagnostic units in the task mapping up to the given number. To be used when
	 * the encoding previously did not contain diagnostic units.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	private void addMissingDiagUnits(AbstractTaskMappingEncoding outGenotype,
			Collection<IDeployableComponentAdapter<C>> diagComps, int entryNumber)
			throws ExplorationException {
		List<InternalIsolatedCommunicationSet> isoComm =
				new ArrayList<>(
						outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class));
		for(int i = 0; i < entryNumber; i++) {
			Set<IMappingEntry> createdEntries = addSFComponents(outGenotype, diagComps);
			InternalIsolatedCommunicationSet curConstr = isoComm.get(i);
			curConstr.addToIsolatedCommunicationSet(createdEntries);
		}
	}

	/** Entirely removes the diagnostic units of a safety function from a given mapping. */
	private void removeDiagUnits(AbstractTaskMappingEncoding outGenotype,
			Collection<IDeployableComponentAdapter<C>> diagComps) {
		for(IDeployableComponentAdapter<?> diagComp : diagComps) {
			Collection<AbstractTaskMappingEntry> removedEntries = outGenotype.removeTask(diagComp);
			Collection<InternalIsolatedCommunicationSet> isoCommConstrs =
					outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class);
			if(isoCommConstrs != null) {
				// During the first iteration, Diagnostic units may be removed but no constraints
				// were generated for them yet, hence check for null here.
				isoCommConstrs.forEach(c -> c.removeFromIsolatedCommunicationSet(removedEntries));
			}
		}

	}

	/**
	 * Adds one random entry to the task mapping for each of the given
	 * {@link IDeployableComponentAdapter}s.
	 * 
	 * @throws ExplorationException
	 *             if no valid allocation targets can be identified for one
	 *             {@link IDeployableComponentAdapter}.
	 */
	private Set<IMappingEntry> addSFComponents(AbstractTaskMappingEncoding outGenotype,
			Collection<IDeployableComponentAdapter<C>> sfComps) throws ExplorationException {
		Set<IMappingEntry> createdEntries = new HashSet<>();
		if(sfComps != null) {
			for(IDeployableComponentAdapter<C> curSFComp : sfComps) {
				createdEntries.add(outGenotype.createAndAddRandomEntry(curSFComp));
			}
		}

		return createdEntries;
	}

	/** Removes the given {@link IMappingEntry}s from the {@link AbstractTaskMappingEncoding}. */
	private void removeSFEntries(AbstractTaskMappingEncoding outGenotype,
			Collection<IMappingEntry> sfEntries) {
		for(IMappingEntry sfEntry : sfEntries) {
			outGenotype.removeEntry((AbstractTaskMappingEntry)sfEntry);
		}
	}

	/**
	 * Returns the first found {@link InternalIsolatedCommunicationSet} constraint that includes
	 * the given {@link IMappingEntry}. If none is found {@code null} is returned.
	 */
	private InternalIsolatedCommunicationSet getMatchingIsolationCommConstraint(
			AbstractTaskMappingEncoding outGenotype, IMappingEntry entry) {
		InternalIsolatedCommunicationSet matchingConstraint = null;
		Collection<InternalIsolatedCommunicationSet> isoComm =
				outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class);
		for(InternalIsolatedCommunicationSet constr : isoComm) {
			if(constr.getIsolatedCommunicationSet().contains(entry)) {
				matchingConstraint = constr;
				break;
			}
		}

		return matchingConstraint;
	}

	/**
	 * Removes the first found {@link InternalIsolatedCommunicationSet} constraint that includes the
	 * given {@link IMappingEntry} from the given {@link AbstractTaskMappingEncoding}.
	 */
	private void removeMatchingConstraint(AbstractTaskMappingEncoding outGenotype,
			IMappingEntry entry) {
		InternalIsolatedCommunicationSet rmConstr =
				getMatchingIsolationCommConstraint(outGenotype, entry);
		assert rmConstr != null;
		outGenotype.getConstraintsOf(InternalIsolatedCommunicationSet.class).remove(rmConstr);
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> reqTypes = new ArrayList<>();
		reqTypes.add(SafetyFunctionArchEncoding.class);
		reqTypes.add(AbstractTaskGraphEncoding.class);
		reqTypes.add(AbstractTaskMappingEncoding.class);
		return reqTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Class<AbstractTaskMappingEncoding> getOutputType() {
		return AbstractTaskMappingEncoding.class;
	}

}
