/*--------------------------------------------------------------------------+
$Id: TimeSlot.java 2419 2016-03-04 18:05:52Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution;

import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;

/**
 * Slot of a {@link StrictTTSchedule} reserving a particular time frame at a a
 * {@link IResourceAdapter}.
 * 
 * @author huang
 * @author $Author: diewald $
 * @version $Rev: 2419 $
 * @ConQAT.Rating RED Hash: 31A6A942AA3B8E0C8D441A6F5D2871D1
 */

public class TimeSlot implements java.lang.Comparable<TimeSlot> {

	/**
	 * The {@link IResourceAdapter} for which this {@link TimeSlot} is allocated. For instance, a
	 * {@link StrictTTSchedule} can contain time slots for {@link IExecutionUnitAdapter} and I
	 */
	private IResourceAdapter<?> resource;

	/** The {@link IMappingEntry} that is represented by this {@link TimeSlot}. */
	private IMappingEntry deployedRequester;

	/** The {@link Message} reserving this {@link TimeSlot} of a communication resource. */
	private Message deployedMessage;

	/** References the {@link TimeSlot}s that are causal predecessors of this {@link TimeSlot}. */
	private Collection<TimeSlot> predecessorSlots;

	/** References the {@link TimeSlot}s that are causal successors of this {@link TimeSlot}. */
	private Collection<TimeSlot> successorSlots;

	/** Start time of {@link TimeSlot}. */
	private double startTime;

	/** End time of {@link TimeSlot}. */
	private double endTime;

	/** Hyper period iteration number of this {@link TimeSlot}. Used for latency checks. */
	private long iteration;

	// TODO: Replace isVoterSlot flag with a subclass of TimeSlot?
	/**
	 * Flag if the time slot is a voter that has been added automatically during the reliability
	 * analysis.
	 */
	private boolean isVoterSlot;

	/**
	 * Constructs a time slot for a given resource, and enables to specify if the slot has been
	 * allocated for a voter.
	 */
	public TimeSlot(IMappingEntry deployedRequest, IResourceAdapter<?> resource, double startTime,
			double endTime, int iteration, boolean isVoterSlot) {
		this.deployedRequester = deployedRequest;
		this.startTime = startTime;
		this.endTime = endTime;
		this.iteration = iteration;
		this.resource = resource;
		this.isVoterSlot = isVoterSlot;

		this.predecessorSlots = new HashSet<TimeSlot>();
		this.successorSlots = new HashSet<TimeSlot>();
	}

	/**
	 * Constructs a time slot for a given resource, and enables to specify if the slot has been
	 * allocated for a voter. Furthermore, the collection of predecessors is defined.
	 */
	public TimeSlot(IMappingEntry deployedRequest, IResourceAdapter<?> resource,
			Collection<TimeSlot> predecessorSlots, double startTime, double endTime, int iteration,
			boolean isVoterSlot) {
		this(deployedRequest, resource, startTime, endTime, iteration, isVoterSlot);

		this.predecessorSlots.addAll(predecessorSlots);
	}

	/** Constructs a time slot for a given mappingEntry. */
	public TimeSlot(IMappingEntry deployedRequest, double startTime, double endTime) {
		this(deployedRequest, deployedRequest.getTarget(), startTime, endTime, 0, false);
	}

	/**
	 * Constructs a time slot for a given {@link Message}. Specialized constructor for
	 * {@link ICommunicationResourceAdapter}s.
	 */
	public TimeSlot(Message deployedMessage, ICommunicationResourceAdapter<?> resource,
			double startTime, double endTime) {
		this(null, resource, startTime, endTime, 0, false);
		this.deployedMessage = deployedMessage;
	}

	/** Returns deployedRequester. */
	public IMappingEntry getDeployedRequester() {
		return deployedRequester;
	}

	/** Returns the {@link Message} requesting this {@link TimeSlot}. */
	public Message getDeployedMessage() {
		return deployedMessage;
	}

	/** Returns the slots start time. */
	public double getStartTime() {
		return startTime;
	}

	/** Returns the slots end time. */
	public double getEndTime() {
		return endTime;
	}

	/**
	 * Returns the duration of the slot, i.e. {@link #getEndTime()} - {@link #getStartTime()}.
	 */
	public double getDuration() {
		return endTime - startTime;
	}

	/**
	 * Shifts the time slot in the schedule.
	 */
	public void shift(double d) {
		startTime += d;
		endTime += d;
	}

	/** Hyper period iteration number of this {@link TimeSlot}. */
	public long getIteration() {
		return iteration;
	}

	/**
	 * Returns the resource for which this slot has been allocated.
	 */
	public IResourceAdapter<?> getResource() {
		return resource;
	}

	/** Returns true iff the slot has been allocated for a voter */
	public boolean isVoter() {
		return isVoterSlot;
	}

	/**
	 * Sets a flag if the time slot is a voter that has been added automatically during the
	 * reliability analysis.
	 */
	public void setVoter(boolean isVoterSlot) {
		this.isVoterSlot = isVoterSlot;
	}

	/**
	 * Add a {@link TimeSlot} to the collection of causal predecessors {@link TimeSlot}s of this
	 * {@link TimeSlot}.
	 * 
	 * @returns {@code true} if the given slot could be added and is not yet present in the list of
	 *          predecessors, {@code false} otherwise.
	 */
	public boolean addPredecessor(TimeSlot slot) {
		return predecessorSlots.add(slot);
	}

	/**
	 * Replaces the predecessor {@link TimeSlot} {#link slotOrig} with {#link replaceSlot}. If
	 * {#link slotOrig} is not contained in the list of predecessor {@link TimeSlot}s, no
	 * modification will be performed.
	 */
	public void replacePredecessor(TimeSlot slotOrig, TimeSlot replaceSlot) {
		if(predecessorSlots.remove(slotOrig)) {
			predecessorSlots.add(replaceSlot);
		}
	}

	/** Returns the collection of causal predecessors {@link TimeSlot}s of this {@link TimeSlot}. */
	public Collection<TimeSlot> getPredecessors() {
		return predecessorSlots;
	}

	/** Clears the collection of causal successors {@link TimeSlot}s of this {@link TimeSlot}. */
	public void clearPredecessors() {
		predecessorSlots.clear();
	}

	/**
	 * Add a {@link TimeSlot} to the collection of causal successors {@link TimeSlot}s of this
	 * {@link TimeSlot}.
	 * 
	 * @returns {@code true} if the given slot could be added and is not yet present in the list of
	 *          successors, {@code false} otherwise.
	 */
	public boolean addSuccessor(TimeSlot slot) {
		return successorSlots.add(slot);
	}

	/** Returns the collection of causal successors {@link TimeSlot}s of this {@link TimeSlot}. */
	public Collection<TimeSlot> getSuccessors() {
		return successorSlots;
	}

	/**
	 * Replaces the successor {@link TimeSlot} {#link slotOrig} with {#link replaceSlot}. If {#link
	 * slotOrig} is not contained in the list of successor {@link TimeSlot}s, no modification will
	 * be performed.
	 */
	public void replaceSuccessor(TimeSlot slotOrig, TimeSlot replaceSlot) {
		if(successorSlots.remove(slotOrig)) {
			successorSlots.add(replaceSlot);
		}
	}

	/** Clears the collection of causal successors {@link TimeSlot}s of this {@link TimeSlot}. */
	public void clearSuccessors() {
		successorSlots.clear();
	}

	/**
	 * Comparison of two time slots:
	 * NO LONGER VALID!:1) if the two slot belong to different resources, compare the resources
	 * names
	 * if the two slot belong to the same resources, compare the start time
	 * */
	// TODO: re-implement this compareTo method, it is a complete mess (refs #6262)
	@Override
	public int compareTo(TimeSlot slot) {
		// TODO: if they are both null, consider them as equal
		if(this.resource == null && slot.resource == null) {
			return Double.compare(this.startTime, slot.startTime);
		} else if(this.resource == null || slot.resource == null) {
			return -1;
		}

		if(this.resource != slot.resource) {
			String n1 = this.resource.getName();
			String n2 = slot.resource.getName();
			if(n1 == null || n2 == null) {
				// TODO: if they are both null, consider them as equal
				if(n1 == null && n2 == null) {
					return Double.compare(this.startTime, slot.startTime);
				}
				return -1;
			}
			return n1.compareTo(n2);
		}

		// on the same provider
		int result = Double.compare(this.startTime, slot.startTime);
		return result;
	}

	/** {@inheritDoc} */
	@Override
	public boolean equals(Object object) {
		// Content-based comparison of two {@link TimeSlot}s.
		if(object instanceof TimeSlot) {
			return(this.compareTo((TimeSlot)object) == 0);
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public int hashCode() {
		// TODO Meaningful implementation due to custom equals() required?
		return super.hashCode();
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return "(" + startTime + ", " + endTime + ")";
	}

	/**
	 * Comparator which determines the (ascending) order of two given {@link TimeSlot}s by their
	 * start times.
	 */
	public static class StartTimeSlotComparator implements Comparator<TimeSlot> {
		/** {@inheritDoc} */
		@Override
		public int compare(TimeSlot slot1, TimeSlot slot2) {
			if(slot1 == null && slot2 == null) {
				return 0;
			}
			if(slot1 == null) {
				return 1;
			}
			if(slot2 == null) {
				return -1;
			}
			return Double.compare(slot1.getStartTime(), slot2.getStartTime());
		}
	}

	/**
	 * Comparator which determines the (ascending) order of two given {@link TimeSlot}s by their
	 * finish(end) times.
	 */
	public static class ReverseStartTimeSlotComparator extends StartTimeSlotComparator {
		/** {@inheritDoc} */
		@Override
		public int compare(TimeSlot slot1, TimeSlot slot2) {
			return -super.compare(slot1, slot2);
		}
	}
}
