/*--------------------------------------------------------------------------+
$Id: SafetyIntegrityLevelConstraintEvaluator.java 3730 2016-10-19 15:43:43Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.constraint;

import static java.lang.Math.pow;
import static org.fortiss.af3.exploration.alg.util.ExplorationAlgUtils.castToRawColl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.eclipse.emf.common.util.Enumerator;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.SolutionQuantification;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.moea.model.predefined.SafetyIntegrityLevelConstraint;
import org.fortiss.af3.safety.model.SIL;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.ClassToInstanceMap;

/**
 * Preliminary {@link Evaluator} to check whether the allocation of specific
 * {@link IDeployableComponentAdapter}s to {@link IExecutionUnitAdapter}s satisfy the Safety
 * Integrity Level required by the {@link IDeployableComponentAdapter}.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3730 $
 * @ConQAT.Rating RED Hash: EFAD8CF49AEA47FF284FAA505216E3FA
 */
public class SafetyIntegrityLevelConstraintEvaluator<S extends InstantiatedTaskMappingEntry, T extends InstantiatedTaskMappingEncoding<S>>
		extends ConstraintEvaluatorBase<T, SIL, SafetyIntegrityLevelConstraint> {

	/** Constructor. */
	public SafetyIntegrityLevelConstraintEvaluator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			SafetyIntegrityLevelConstraint optGoal) {
		super(systemModelAdapter, optGoal);
	}

	/** {@inheritDoc} */
	@Override
	public DoubleValue evaluateGoal(T phenotype, FlatPhenotypeMap<Phenotype> phenotypeTypeMap,
			ClassToInstanceMap<EObject> transformedModels) {
		List<SIL> silViolations = getSILViolations(phenotype);
		double aggregatedViolations =
				silViolations.stream().mapToDouble(sil -> pow(sil.getValue(), 2.0)).sum();

		return new DoubleValue(aggregatedViolations);
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateInternal(ExplorationTarget<?> expTarget, T mapping,
			SolutionQuantification valueSet) {
		Map<IDeployableComponentAdapter<?>, Collection<S>> allocationMap =
				mapping.getAllocationMap();
		List<IDeployableComponentAdapter<?>> silComponents = getSILComponentAdapters();
		List<SIL> silBounds = optTarget.getLimitation();
		for(int idx = 0; idx < silComponents.size(); idx++) {
			SIL actualSIL = detectSIL(silComponents.get(idx), allocationMap);
			if(actualSIL.getValue() < silBounds.get(idx).getValue()) {
				return false;
			}
		}
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateGoalSpecification() {
		// Currently, there is no known error case needing validation
		return true;
	}

	/**
	 * Returns the set of {@link IDeployableComponentAdapter}s that is equivalent to the set of
	 * {@link Component}s given in the {@link SafetyIntegrityLevelConstraint}.
	 */
	private List<IDeployableComponentAdapter<?>> getSILComponentAdapters() {
		Collection<IDeployableComponentAdapter<?>> deplComp =
				castToRawColl(systemModelAdapter.getDeployableComponents());
		return deplComp.stream().filter(c -> optTarget.getComponents().contains(c))
				.collect(Collectors.toList());
	}

	/**
	 * Returns the list of of violated {@link SIL} requirements as the difference between the
	 * required {@link SIL} and the actually achieved {@link SIL}.
	 * The list has the same order as the list of referenced {@link Component}s in the
	 * {@link SafetyIntegrityLevelConstraint}.
	 */
	private List<SIL> getSILViolations(T mapping) {
		List<SIL> silViolations = new ArrayList<>();
		Map<IDeployableComponentAdapter<?>, Collection<S>> allocationMap =
				mapping.getAllocationMap();
		List<IDeployableComponentAdapter<?>> silComponents = getSILComponentAdapters();
		List<SIL> silBounds = optTarget.getLimitation();
		for(int idx = 0; idx < silComponents.size(); idx++) {
			SIL actualSIL = detectSIL(silComponents.get(idx), allocationMap);
			silViolations.add(calculateSILViolation(silBounds.get(idx), actualSIL));
		}
		return silViolations;
	}

	// TODO: Port to org.fortiss.af3.safety
	/** Helper function for the calculation of the actual SIL of a deployed component. */
	private SIL detectSIL(IDeployableComponentAdapter<?> deployableComponent,
			Map<IDeployableComponentAdapter<?>, Collection<S>> allocationMap) {
		// Note: real hw replication is not meant here, just the allocation to different cores/ecus
		Collection<S> replica = allocationMap.get(deployableComponent);

		// Determine the number of replications onto different execution units. This separation is
		// yet very primitive.
		Collection<IExecutionUnitAdapter<?>> hwReplica = new HashSet<IExecutionUnitAdapter<?>>();
		for(S currentTaskReplic : replica) {
			hwReplica.add(currentTaskReplic.getTarget());
		}
		int silTypeToInt = 0;

		// Determine the SIL of a "safety function", i.e. a safety-annotated component mapped onto
		// an execution unit, according to the tables given in IEC 61508-2.
		List<SIL> parallelSILs = new ArrayList<SIL>();
		for(IExecutionUnitAdapter<?> ecu : hwReplica) {
			double sff = ecu.getSafeFailureFraction();
			if(sff >= 0.0 && sff < 0.6) {
				parallelSILs.add(SIL.SIL0);
			} else if(sff < 0.9) {
				parallelSILs.add(SIL.SIL1);
			} else if(sff < 0.99) {
				parallelSILs.add(SIL.SIL2);
			} else if(sff <= 1.0) {
				parallelSILs.add(SIL.SIL3);
			} else {
				throw new RuntimeException(
						"An invalid value for the safe failure fraction has been given to the optimization. The allowed range is 0 to 1.");
			}
		}

		// Determine the overall SIL of the safety function considering the hardware replication.
		// The upper limit is set to 4, in accordance with IEC 61508.
		for(SIL sil : parallelSILs) {
			silTypeToInt += sil.getValue();
		}
		if(silTypeToInt > 4)
			silTypeToInt = 4;

		return SIL.get(silTypeToInt);
	}

	/**
	 * Helper function for the calculation of the punishment value for a single SIL violation.
	 * Currently, this value is determined by the iteration number (weak punishment in the
	 * beginning) and the quadratical deviation from the desired SIL.
	 */
	// TODO: Port to org.fortiss.af3.safety
	private SIL calculateSILViolation(Enumerator requiredSIL, SIL actualSIL) {
		int differenceSIL = requiredSIL.getValue() - actualSIL.getValue();
		return SIL.get(differenceSIL);

		// TODO divide by number of iterations, currently not accessible
		// if(differenceSIL > 0) {
		// return getIterationCountAndIncrement() * Math.pow(differenceSIL, 2.0);
		// }
		// return 0.0;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(InstantiatedTaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return InstantiatedTaskMappingEncoding.class;
	}

	/** {@inheritDoc} */
	@Override
	public IExplorationTargetEvaluator<T, SIL, SafetyIntegrityLevelConstraint> createEvaluator(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			SafetyIntegrityLevelConstraint explorationTarget) {
		return new SafetyIntegrityLevelConstraintEvaluator<S, T>(systemModelAdapter,
				explorationTarget);
	}
}
