/*--------------------------------------------------------------------------+
$Id: EnergyObjectiveEvaluator.java 3730 2016-10-19 15:43:43Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator.objective;

import java.util.Collection;
import java.util.HashSet;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule.ResourceSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.evaluator.ExplorationTargetEvaluatorBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.moea.model.predefined.EnergyMinObjective;
import org.opt4j.core.DoubleValue;

import com.google.common.collect.ClassToInstanceMap;

/**
 * Simple energy evaluator that just sums the execution times.
 * 
 * @author huang
 * @author $Author: diewald $
 * @version $Rev: 3730 $
 * @ConQAT.Rating YELLOW Hash: 22A0CB4CE45AAD148AB6234873859C4A
 */
public class EnergyObjectiveEvaluator<S extends TaskMappingEntry, T extends TaskMappingEncoding<S>>
		extends ExplorationTargetEvaluatorBase<StrictTTSchedule<S, T>, Double, EnergyMinObjective> {

	/** Creates a template energy objective evaluator */
	public EnergyObjectiveEvaluator() {
		super(null, null);
	}

	/** Creates an energy objective evaluator */
	public EnergyObjectiveEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			EnergyMinObjective goal) {
		super(systemModelAdapter, goal);
	}

	/**
	 * This energy evaluator sums the slot lengths of the {@link StrictTTSchedule}. Thus, it
	 * considers all {@link IResourceAdapter}s that are under the regime of that schedule, e.g.,
	 * {@link IExecutionUnitAdapter}s and {@link ITransmissionUnitAdapter}s.
	 */
	@Override
	public DoubleValue evaluateGoal(StrictTTSchedule<S, T> schedule,
			FlatPhenotypeMap<Phenotype> phenotypeTypeMap,
			ClassToInstanceMap<EObject> transformedModels) {
		InstantiatedTaskMappingEncoding<?> itmEnc =
				phenotypeTypeMap.getEncoding(InstantiatedTaskMappingEncoding.class);
		assert (itmEnc != null) : "The required instantiated task mapping encoding could not be found in the map of already decoded phenotypes.";

		if(schedule.getResources().isEmpty()) {
			return new DoubleValue(Double.MAX_VALUE);
		}

		double energy = 0;

		for(IResourceAdapter<?> deploymentTarget : schedule.getResources()) {
			ResourceSchedule localSchedule = schedule.getResourceSchedule(deploymentTarget);
			// Catch the case where no components are deployed on the given resource.
			if(localSchedule != null) {
				for(TimeSlot t : localSchedule.keySet()) {
					assert (t.getDuration() >= 0);
					if(deploymentTarget instanceof IExecutionUnitAdapter) {
						IExecutionUnitAdapter<?> executionUnit =
								(IExecutionUnitAdapter<?>)deploymentTarget;
						IDeployableComponentAdapter<?> deployedComponent =
								t.getDeployedRequester().getDeployableComponent();

						energy +=
								getConsumedEnergy(t.getDuration(), executionUnit, deployedComponent);
					} else if(deploymentTarget instanceof ICommunicationResourceAdapter) {
						ICommunicationResourceAdapter<?> commResource =
								(ICommunicationResourceAdapter<?>)deploymentTarget;
						energy += commResource.getPower(t.getDeployedMessage().getMessageSize());
					}
				}
			}
		}

		return new DoubleValue(energy);
	}

	/**
	 * Returns the consumed energy of the given allocation given by the deployment specific
	 * parameter map or by calculation via the WCET of the deployed
	 * {@link IDeployableComponentAdapter} and the speed of the {@link IExecutionUnitAdapter}.
	 */
	private double getConsumedEnergy(double duration, IExecutionUnitAdapter<?> executionUnit,
			IDeployableComponentAdapter<?> deployedComponent) {
		double deployedEnergyDemand = deployedComponent.getWcet(executionUnit);
		if(deployedEnergyDemand <= 0.0) {
			deployedEnergyDemand = executionUnit.getPower(duration);
		}
		return deployedEnergyDemand;
	}

	/** {@inheritDoc} */
	@Override
	public boolean validateGoalSpecification() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredInputTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(TaskMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Class<?> getPrimaryInputType() {
		return StrictTTSchedule.class;
	}

	/** {@inheritDoc} */
	@Override
	public IExplorationTargetEvaluator<StrictTTSchedule<S, T>, Double, EnergyMinObjective>
			createEvaluator(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					EnergyMinObjective explorationTarget) {
		return new EnergyObjectiveEvaluator<>(systemModelAdapter, explorationTarget);
	}
}
