/*--------------------------------------------------------------------------+
$Id: ComponentArchitectureInserter.java 3561 2016-09-23 14:22:56Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.comparch;

import static org.fortiss.af3.component.utils.ComponentModelElementFactory.createChannelAndAttach;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.arrangePorts;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.calulateOffset;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.getTargetPorts;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.replicateInputPortAndAttach;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.replicateOutputPortAndAttach;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.utils.ComponentModelElementFactory;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.tooling.base.model.layout.Point;

/**
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3561 $
 * @ConQAT.Rating RED Hash:
 */
public class ComponentArchitectureInserter extends TransformationModuleBase<ComponentArchitecture> {

	/** {@inheritDoc} */
	@Override
	public Class<ComponentArchitecture> getOutputType() {
		return ComponentArchitecture.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(ComponentArchitecture.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ComponentArchitectureTransformer.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.INSERTING;
	}

	/**
	 * {@inheritDoc}
	 * 
	 * @throws TransformationModuleException
	 *             see {@link ITransformationModule}.
	 */
	@Override
	public <N extends EObject, S extends EObject> ComponentArchitecture transform(
			TransformationState state, TransformationContext context,
			SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		ComponentArchitecture genCA = (ComponentArchitecture)outputModel;
		if(genCA == null) {
			throw new TransformationModuleException(this,
					"Could not retrieve the generated Component Architecture.");
		}

		insertVoters(genCA);

		return (ComponentArchitecture)outputModel;
	}

	/**
	 * Inserts voter {@link Component}s into the given {@link ComponentArchitecture}. Voter
	 * {@link Component}s are added directly in front of the receiver of a signal/message. It takes
	 * the signals of the original {@link Component} and its replica as an input and forwards the
	 * result of the voting to the actual receiver {@link Component}.
	 */
	private Map<Component, IPlatformResource> insertVoters(
			ComponentArchitecture replicaAwareComponentArchitecture) {
		Map<Component, IPlatformResource> voterAllocations =
				new HashMap<Component, IPlatformResource>();

		Collection<InputPort> allTargetPortsOfReplicatedArch =
				getTargetPorts(replicaAwareComponentArchitecture);

		for(InputPort currentTargetPort : allTargetPortsOfReplicatedArch) {
			// Evaluate the need to add voters by examining the incoming channels for each target
			// port. If these channels were not present in the original architecture, voting is
			// needed.
			Map<Integer, List<Port>> sourcePortMap = new HashMap<Integer, List<Port>>();
			Map<Port, Channel> sourcePortToTargetChannelMap = new HashMap<Port, Channel>();
			for(Channel currentIncomingChannel : currentTargetPort.getIncomingChannels()) {
				int currentId = currentIncomingChannel.getSource().getId();
				List<Port> currentPortList = sourcePortMap.get(currentId);
				if(currentPortList == null) {
					currentPortList = new ArrayList<Port>();
				}
				currentPortList.add(currentIncomingChannel.getSource());
				sourcePortMap.put(currentId, currentPortList);
				sourcePortToTargetChannelMap.put(currentIncomingChannel.getSource(),
						currentIncomingChannel);
			}

			Set<Integer> remainingSourcePortIds = sourcePortMap.keySet();
			while(!remainingSourcePortIds.isEmpty()) {
				Integer currentSourcePortId = remainingSourcePortIds.iterator().next();
				Collection<Port> sourcePorts = sourcePortMap.get(currentSourcePortId);
				if(sourcePorts.size() > 1) {
					// Create voter.
					// TODO: assumption that the container is a Component.
					Component replicatedComponent = currentTargetPort.getComponent();

					Component voter =
							createVoter(currentTargetPort, sourcePortToTargetChannelMap,
									sourcePorts, replicatedComponent);

					// A voter has exactly one output port
					assert (voter.getOutputPorts().size() == 1);
					OutputPort voterOutput = voter.getOutputPorts().get(0);

					// Connect voter to "original" Component
					createChannelAndAttach((Component)voter.eContainer(), "Voting_result",
							voterOutput, currentTargetPort);

					// voterAllocations.put(voter, deploymentComponents.get(replicatedComponent));
				}
				remainingSourcePortIds.remove(currentSourcePortId);
			}
		}

		return voterAllocations;
	}

	/** Creates a voter {@link Component}. */
	private Component createVoter(InputPort currentTargetPort,
			Map<Port, Channel> sourcePortToTargetChannelMap, Collection<Port> sourcePorts,
			Component replicatedComponent) {
		Component voter =
				ComponentModelElementFactory.createComponentAndAttach(
						(Component)replicatedComponent.getContainer(), "Voter_" +
								replicatedComponent.getName(), false);

		if(replicatedComponent.getLayoutData().get(0) instanceof Point) {
			Point p = (Point)EcoreUtil.copy(replicatedComponent.getLayoutData().get(0));
			p = calulateOffset(p, replicatedComponent, -10, true, false);
			voter.getLayoutData().set(0, p);
		}

		// Create an InputPort at the voter for each incoming Channel of replicated
		// Components.
		for(Port currentSourcePort : sourcePorts) {
			String voterName = "Vin_";
			if(currentSourcePort.getIncoming().isEmpty()) {
				voterName += currentSourcePort.getComponent().getName();
			} else {
				voterName += currentSourcePort.getName();
			}
			InputPort voterInput = replicateInputPortAndAttach(currentSourcePort, voter, voterName);
			sourcePortToTargetChannelMap.get(currentSourcePort).setTarget(voterInput);
		}

		// Connect voter output port
		replicateOutputPortAndAttach(currentTargetPort, voter, "Result_" + voter.getName());

		arrangePorts(voter.getInputPorts(), voter);
		arrangePorts(voter.getOutputPorts(), voter);

		return voter;
	}

	// /** Specifies the given voter {@link Component} as a majority voter. */
	// private void addMajorityVoterSpec(Component voterComponent) {
	// VoterType vType = MoeaFactory.eINSTANCE.createVoterType();
	// vType.setType(VoterType_t.MAJORITY_VOTER);
	// List<IModelElementSpecification> specs = voterComponent.getSpecifications();
	// specs.add(vType);
	// }
}
