/*--------------------------------------------------------------------------+
$Id: DeploymentFinalizer.java 3561 2016-09-23 14:22:56Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.deployment;

import static org.fortiss.af3.deployment.utils.DeploymentModelElementFactory.createComponentAllocation;
import static org.fortiss.af3.deployment.utils.DeploymentParameterUtils.getDeploymentsWithParametersOf;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.getFirstComponentPool;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.resetIds;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.deployment.model.AF3DeploymentFactory;
import org.fortiss.af3.deployment.model.ComponentAllocation;
import org.fortiss.af3.deployment.model.ComponentArchitectureReference;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.deployment.model.DeploymentParameterKey;
import org.fortiss.af3.deployment.model.DeploymentParameterValue;
import org.fortiss.af3.deployment.model.impl.DeploymentParameterKeyToDeploymentParameterValueMapImpl;
import org.fortiss.af3.deployment.utils.DeploymentModelElementFactory;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.EMFModelConcurrencyHandler;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.exception.TransformationModuleException;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationEncodingMap;
import org.fortiss.af3.exploration.alg.service.IExplorationEncoding;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.tooling.kernel.utils.EcoreUtils;

/**
 * Creates the {@link ComponentAllocation}s and the
 * {@link DeploymentParameterKeyToDeploymentParameterValueMapImpl}.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3561 $
 * @ConQAT.Rating RED Hash: DD31B30D577F35527815FDB295256E1C
 */
public class DeploymentFinalizer extends TransformationModuleBase<Deployment> {

	/** {@inheritDoc} */
	@Override
	public Class<Deployment> getOutputType() {
		return Deployment.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		inputTypes.add(ComponentArchitecture.class);
		inputTypes.add(PlatformArchitecture.class);
		inputTypes.add(IMappingEncoding.class);
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(ExplorationSolutionToDeployment.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return TransformationState.REPLICATED;
	}

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> Deployment transform(TransformationState state,
			TransformationContext context, SubClassToInstanceMap<S> baseModels,
			ExplorationEncodingMap<IExplorationEncoding> explorationEncodings,
			SubClassToInstanceMap<N> transformedModels, EObject outputModel)
			throws TransformationModuleException {
		IMappingEncoding mapping = explorationEncodings.getEncoding(IMappingEncoding.class);
		if(mapping == null) {
			throw new TransformationModuleException(this,
					"An IMappingEncoding is required for this module to operate properly.");
		}

		ComponentArchitecture transformedCA =
				(ComponentArchitecture)transformedModels.get(ComponentArchitecture.class);
		if(transformedCA == null) {
			transformedCA = (ComponentArchitecture)baseModels.get(ComponentArchitecture.class);
			if(transformedCA == null) {
				throw new TransformationModuleException(
						this,
						"Neither a transformed, nor an input component architecture has been found that shall be referenced by the transformed deployment.");
			}
		}

		ComponentArchitecture originalCA =
				(ComponentArchitecture)baseModels.get(ComponentArchitecture.class);
		if(originalCA == null) {
			throw new TransformationModuleException(
					this,
					"Could not retrieve the original component architecture from the set of base models. Be sure to add it to the required input types.");
		}

		PlatformArchitecture originalPA =
				(PlatformArchitecture)baseModels.get(PlatformArchitecture.class);
		if(originalPA == null) {
			throw new TransformationModuleException(
					this,
					"Could not retrieve the original platform architecture from the set of base models. Be sure to add it to the required input types.");
		}

		ComponentArchitectureReference caRef =
				AF3DeploymentFactory.eINSTANCE.createComponentArchitectureReference();
		((Deployment)outputModel).setComponentArchitectureReference(caRef);
		EMFModelConcurrencyHandler.getInstance().setReference(caRef, transformedCA);

		// TODO: Maybe distribute the generation of component allocation over the various modules.
		((Deployment)outputModel).getComponentAllocations().addAll(
				allocateComponentsToExecutionUnits(mapping));

		createDeploymentParameterMap(originalCA, originalPA, (Deployment)outputModel, mapping);

		// Re-enable Notifications & the Annotation view update.
		EMFModelConcurrencyHandler.getInstance().tryEnableNotificationsSyncOf(outputModel);

		return (Deployment)outputModel;
	}

	/**
	 * Allocates {@link Component}s to {@link ExecutionUnit}s using the given {@link Map}. This
	 * methods uses the DeploymentUtils and returns a {@link ComponentAllocation} that can be
	 * processed in an AF3 compatible {@link Deployment}.
	 */
	private List<ComponentAllocation> allocateComponentsToExecutionUnits(
			IMappingEncoding mappingEncoding) {
		List<ComponentAllocation> componentToExecutionUnitAllocations =
				new ArrayList<ComponentAllocation>();
		for(IMappingEntry curEntry : mappingEncoding.getIMappingEntries()) {
			Component curComponent =
					(Component)encEntryModelElemAssoc.get(curEntry, Component.class);
			// If the given entry is unknown, we examine the associated entries since the
			// association may be based on some other relevant entry. This is the case for
			// replication that is performed for abstract entries while we have instantiated entries
			// here.
			if(curComponent == null) {
				for(IMappingEntry assocEntry : curEntry.getAssociatedElement()) {
					curComponent =
							(Component)encEntryModelElemAssoc.get(assocEntry, Component.class);
					if(curComponent != null) {
						curEntry = assocEntry;
						break;
					}
				}
			}

			ExecutionUnit curExecUnit = (ExecutionUnit)curEntry.getTarget().getObject();
			assert (curComponent != null && curExecUnit != null);
			componentToExecutionUnitAllocations.add(createComponentAllocation(curComponent,
					curExecUnit));
		}
		return componentToExecutionUnitAllocations;
	}

	/**
	 * Creates the deployment parameter map for the given {@link Deployment} by copying this map
	 * from an existing deployment in the {@link FileProject}. The map is extended to consider
	 * replica of {@link Component}s if required.
	 */
	// TODO: how to handle instantiated components?
	private void createDeploymentParameterMap(ComponentArchitecture originalCA,
			PlatformArchitecture originalPA, Deployment deployment, IMappingEncoding mapping) {
		List<Deployment> allDeploymentsWithParameters =
				getDeploymentsWithParametersOf(originalCA, originalPA);

		// Get the referenced Component Pool of the original Component Architecture & the deployment
		// specific parameters.
		ComponentArchitecture compPool = getFirstComponentPool(mapping);
		List<Deployment> poolParameterDeployments =
				getDeploymentsWithParametersOf(compPool, originalPA);

		if(allDeploymentsWithParameters.size() > 0) {
			// Pick the first deployments with a matching parameter map.
			Deployment deploymentWithParameters = allDeploymentsWithParameters.get(0);
			Deployment poolParameterDeployment = null;
			if(poolParameterDeployments.size() > 0) {
				poolParameterDeployment = poolParameterDeployments.get(0);
			}

			// For transformed components that were part of the "actual" CA.
			fillDeploymentParameterMap(deployment, deploymentWithParameters);
			// For transformed components that are instantiated from a "Component Pool".
			fillDeploymentParameterMap(deployment, poolParameterDeployment);
		}
	}

	/** Adds the map of deployment specific parameters to the transformed {@link Deployment} model. */
	private void fillDeploymentParameterMap(Deployment deployment, Deployment paramDeployment) {
		if(paramDeployment == null) {
			return;
		}

		for(Entry<DeploymentParameterKey, DeploymentParameterValue> deploymentParameter : paramDeployment
				.getDeploymentParameterMap().entrySet()) {
			DeploymentParameterValue value = deploymentParameter.getValue();
			Component parameterComponent = value.getComponent();

			for(EObject genComp : origToGenAssocMap.get(parameterComponent)) {
				replicateDeploymentParameterEntry(deployment, value, (Component)genComp);
			}
		}
	}

	/**
	 * Creates a new key, copies and adjust the parameter value, and inserts it into the parameter
	 * map.
	 */
	private void replicateDeploymentParameterEntry(Deployment deployment,
			DeploymentParameterValue parameterValue, Component component) {
		DeploymentParameterKey replicaKey =
				DeploymentModelElementFactory.createDeploymentParameterKey(component,
						parameterValue.getExecutionUnit());
		DeploymentParameterValue copiedValue = EcoreUtils.copy(parameterValue);
		copiedValue.setComponent(component);
		resetIds(copiedValue);
		deployment.getDeploymentParameterMap().put(replicaKey, copiedValue);
	}
}
