/*--------------------------------------------------------------------------+
$Id: PlatformCommunicationGraph.java 3449 2016-09-02 11:06:00Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import static org.fortiss.af3.exploration.util.ExplorationUtils.isDebugVerboseEnabled;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.alg.graph.display.JGraphTVisualizer;
import org.jgrapht.DirectedGraph;
import org.jgrapht.Graphs;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

/**
 * Abstracts the communication network present in a platform system meta-model to a flat graph. This
 * graph is needed to manage the communication of logical {@link IDeployableComponentAdapter}s after
 * they have been deployed onto a concrete platform. More precisely, the graph is used for routing
 * of (physical) messages and their scheduling on the communication resources, like busses.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3449 $
 * @ConQAT.Rating YELLOW Hash: 9ABA58EF4FFD588CDA28D13701904A06
 */
public class PlatformCommunicationGraph {

	/** Representation of input system model */
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/** The graph representing the communication structure of a platform. */
	protected DirectedGraph<IResourceAdapter<?>, DefaultEdge> platformGraph;

	/**
	 * /**
	 * Constructor.
	 * 
	 * @param systemModelAdapter
	 *            The adapter for the input model which contains the platform
	 */
	public PlatformCommunicationGraph(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;

		platformGraph =
				new DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge>(DefaultEdge.class);

		createGraph();
	}

	/**
	 * Creates the communication (or connection) graph which is implicitly given by the platform
	 * referenced in the {@link SystemModelAdapter}. This graph is used for routing messages.
	 */
	private void createGraph() {
		DirectedGraph<IResourceAdapter<?>, DefaultEdge> tempPlatformCommGraph =
				new DefaultDirectedGraph<IResourceAdapter<?>, DefaultEdge>(DefaultEdge.class);

		// Add all resources present in the platform as vertices to the "connection graph"
		for(IResourceAdapter<?> resource : systemModelAdapter.getDeploymentTargets()) {
			tempPlatformCommGraph.addVertex(resource);
		}
		for(IResourceAdapter<?> resource : systemModelAdapter.getMemoryUnits()) {
			tempPlatformCommGraph.addVertex(resource);
		}
		for(IResourceAdapter<?> resource : systemModelAdapter.getGatewayUnits()) {
			tempPlatformCommGraph.addVertex(resource);
		}
		for(IResourceAdapter<?> resource : systemModelAdapter.getTransmissionUnits()) {
			tempPlatformCommGraph.addVertex(resource);
		}

		// Successively connect the resources in the graph. The edges are given within the
		// collection resource connectors are assumed to be fully connected.
		for(IResourceConnectionAdapter connection : systemModelAdapter.deploymentTargetConnectors) {
			// We add edges for each receiver in order to support broadcast connections
			for(IResourceAdapter<?> currentReceiverResource : connection.getTargetResources()) {
				// If the current resource is a child of a {@link TransmissionUnit} (like a NoC
				// router), respect the direction of connections such that complex networks are
				// correctly reflected.
				if(connection.getConnectionType(currentReceiverResource) == ConnectionType.OUTGOING) {
					Graphs.addEdgeWithVertices(platformGraph, connection.getSourceResource(),
							currentReceiverResource);
				} else if(connection.getConnectionType(currentReceiverResource) == ConnectionType.INCOMING) {
					Graphs.addEdgeWithVertices(platformGraph, currentReceiverResource,
							connection.getSourceResource());
				} else {
					// Add the edges of a single connection. We add to edges for each connection
					// such that an undirected graph is generated effectively. It is required to use
					// this "trick" in order to be able to use the standard shortest path
					// algorithms.
					Graphs.addEdgeWithVertices(platformGraph, connection.getSourceResource(),
							currentReceiverResource);
					Graphs.addEdgeWithVertices(platformGraph, currentReceiverResource,
							connection.getSourceResource());
				}
			}
		}

		// Debug code
		if(isDebugVerboseEnabled()) {
			display();
		}
	}

	/** Creates a Dialog that illustrates this {@link DirectedGraph}. */
	public void display() {
		JGraphTVisualizer<IResourceAdapter<?>, DefaultEdge> taskViz =
				new JGraphTVisualizer<IResourceAdapter<?>, DefaultEdge>(platformGraph);
		taskViz.displayGraph("Platform communication graph");
	}

	/**
	 * Returns the actual graph that contains information about the {@link IResourceAdapter}s and
	 * their connections.
	 */
	public DirectedGraph<IResourceAdapter<?>, DefaultEdge> getActualGraph() {
		return platformGraph;
	}
}
