/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch;

import java.util.Collection;
import java.util.Map;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceConnectionAdapter.ConnectionType;
import org.fortiss.af3.exploration.extension.IDseInputParameters.SignalType;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.IPlatformArchitectureElement;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.platform.model.annotation.ResourceLink;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.element.IModelElement;

import com.google.common.collect.Multimap;

/**
 * 
 * @author diewald
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public class SystemParameterContainer {

	/** Collection of deployable {@link Component}s for the target platform. */
	private Collection<Component> deployableComponents;

	/**
	 * Maps the {@link Port}s of the deployable {@link Component}s to their temporal
	 * {@link SignalType} (aperiodic, sporadic, etc.).
	 */
	private Map<Port, SignalType> portSignalTypes;

	/** {@link ExecutionUnit}s to which {@link Component}s are mapped. */
	private Collection<ExecutionUnit> deploymentTargets;

	/** Map relating deployable {@link Component}s with their respective periods. */
	private Map<Component, Double> componentPeriodMap;

	/**
	 * Relates {@link Component}s from the input {@link ComponentArchitecture} to their diverse
	 * implementations, i.e. concrete instances (if any).
	 */
	private Multimap<Component, Component> replacementComponents;

	/**
	 * Associates deployable {@link Component}s with the number of minimally and maximally allowed
	 * replication (1 == no replication).
	 */
	private Map<Component, Pair<Integer, Integer>> componentReplicationBounds;

	/**
	 * Associates the OutputPorts of the deployable {@link Component}s with the size of the messages
	 * they emit.
	 */
	private Map<OutputPort, Long> portMessageSizes;

	/**
	 * Assotiates elements of the {@link PlatformArchitecture} with their respective mean failure
	 * rate.
	 */
	private Map<IPlatformArchitectureElement, Double> resourceFailureProbabilities;

	/**
	 * Associates {@link IPlatformResource}s with their respective active power consumption (in
	 * watts).
	 */
	private Map<IPlatformResource, Double> resourcePowerConsumption;

	/**
	 * Look up table to determine the input/output directions of {@link IConnector} types of the
	 * target {@link PlatformArchitecture}.
	 */
	private Map<Class<? extends IConnector>, ConnectionType> connectionTypeLUT;

	/**
	 * Identifies the {@link IModelElement} types which establish a connection between two
	 * {@link PlatformArchitecture}s via an {@link ResourceLink}.
	 */
	private Pair<Class<? extends IModelElement>, Class<? extends IModelElement>> platfromInterconnectTypes;

	/** Defines the fault containment region defined for the target {@link PlatformArchitecture}. */
	private Class<? extends IPlatformResource> faultContainmentRegion;

	/** Constructor. */
	public SystemParameterContainer(
			Collection<Component> deployableComponents,
			Map<Port, SignalType> portSignalTypes,
			Collection<ExecutionUnit> deploymentTargets,
			Map<Component, Double> componentPeriodMap,
			Multimap<Component, Component> replacementComponents,
			Map<Component, Pair<Integer, Integer>> componentReplicationBounds,
			Map<OutputPort, Long> portMessageSizes,
			Map<IPlatformArchitectureElement, Double> resourceFailureMap,
			Map<IPlatformResource, Double> resourcePowerConsumption,
			Map<Class<? extends IConnector>, ConnectionType> connectionTypeLUT,
			Pair<Class<? extends IModelElement>, Class<? extends IModelElement>> platfromInterconnectTypes,
			Class<? extends IPlatformResource> faultContainmentRegion) {
		this.deployableComponents = deployableComponents;
		this.portSignalTypes = portSignalTypes;
		this.deploymentTargets = deploymentTargets;
		this.componentPeriodMap = componentPeriodMap;
		this.replacementComponents = replacementComponents;
		this.componentReplicationBounds = componentReplicationBounds;
		this.portMessageSizes = portMessageSizes;
		this.resourceFailureProbabilities = resourceFailureMap;
		this.resourcePowerConsumption = resourcePowerConsumption;
		this.connectionTypeLUT = connectionTypeLUT;
		this.platfromInterconnectTypes = platfromInterconnectTypes;
		this.faultContainmentRegion = faultContainmentRegion;

	}

	/** Returns the deployable {@link Component}s. */
	public Collection<Component> getDeployableComponents() {
		return deployableComponents;
	}

	/** Returns the {@link Port} to signal type associations. */
	public Map<Port, SignalType> getPortSignalTypes() {
		return portSignalTypes;
	}

	/** Returns the set of deployment targets (to where {@link Component}s may be mapped). */
	public Collection<ExecutionUnit> getDeploymentTargets() {
		return deploymentTargets;
	}

	/** Returns the associations of deployable {@link Component}s and their defined periods. */
	public Map<Component, Double> getComponentPeriodMap() {
		return componentPeriodMap;
	}

	/**
	 * Returns the associations between deployable {@link Component}s and their diverse
	 * implementations (if any).
	 */
	public Multimap<Component, Component> getReplacementComponents() {
		return replacementComponents;
	}

	/**
	 * Returns the association between deployable {@link Component}s and their minimum and maximum
	 * replication bound.
	 */
	public Map<Component, Pair<Integer, Integer>> getComponentReplicationBounds() {
		return componentReplicationBounds;
	}

	/**
	 * Returns the associations between the {@link OutputPort}s of deployable {@link Component}s and
	 * the size of the data they are sending.
	 */
	public Map<OutputPort, Long> getPortMessageSizes() {
		return portMessageSizes;
	}

	/** Returns the associations between platform elements and their mean failur probability. */
	public Map<IPlatformArchitectureElement, Double> getResourceFailureProbabilities() {
		return resourceFailureProbabilities;
	}

	/** Returns the associations between platform elements and their mean active power consumption. */
	public Map<IPlatformResource, Double> getResourcePowerConsumption() {
		return resourcePowerConsumption;
	}

	/**
	 * Returns the lookup table mapping platform connectors their respective input output direction
	 * type.
	 */
	public Map<Class<? extends IConnector>, ConnectionType> getConnectionTypeLUT() {
		return connectionTypeLUT;
	}

	/**
	 * Returns the {@link IModelElement} types connecting entities from different
	 * {@link PlatformArchitecture}s, e.g. a System Software and a Hardware Architecture.
	 */
	public Pair<Class<? extends IModelElement>, Class<? extends IModelElement>>
			getPlatfromInterconnectTypes() {
		return platfromInterconnectTypes;
	}

	/**
	 * Returns the {@link IPlatformResource} that is assumed to be the fault containment region of
	 * the target platform.
	 */
	public Class<? extends IPlatformResource> getFaultContainmentRegion() {
		return faultContainmentRegion;
	}
}
