/*--------------------------------------------------------------------------+
$Id: AF3DeployableComponentAdapter.java 3096 2016-06-27 08:39:10Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.af3;

import static org.fortiss.tooling.base.utils.AnnotationUtils.getAnnotation;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;

import org.eclipse.emf.common.util.Enumerator;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.InternalComponentParameters;
import org.fortiss.af3.safety.model.annotation.SafetyIntegrityLevel;

/**
 * Adapter between the AF3 deployable {@link Component} (=tasks) and the DSE.
 * 
 * @author barner
 * @author $Author: diewald $
 * @version $Rev: 3096 $
 * @ConQAT.Rating RED Hash: 0076A472847522AEEEE16A347661D7C1
 */
public class AF3DeployableComponentAdapter extends AF3RequestAdapter<Component> implements
		IDeployableComponentAdapter<Component> {
	/** Map of possible Wcets defined by the deployment target. */
	private Map<IExecutionUnitAdapter<?>, Double> wcetMap;

	/** Map of possible energy consumption defined by the deployment target. */
	private Map<IExecutionUnitAdapter<?>, Double> energyMap;

	/** Minimum number of replica (1 == no replication, just the component itself). */
	private int minReplication;

	/** Maximum number of replica (1 == no replication, just the component itself). */
	private int maxReplication;

	/** Period annotation */
	private Double period;

	/** SIL annotation */
	private SafetyIntegrityLevel sil;

	/** Reference to the corresponding elements of the diverse implementations. */
	private Collection<IDeployableComponentAdapter<Component>> replacementComponents;

	/**
	 * "Trace" to the {@link IDeployableComponentAdapter} which {@code this}
	 * {@link IDeployableComponentAdapter} replaced (if applicable).
	 */
	private IDeployableComponentAdapter<Component> replacedComponent;

	/**
	 * Creates an {@link IDeployableComponentAdapter} for an {@link Component}.
	 * 
	 * @throws Exception
	 */
	public AF3DeployableComponentAdapter(Component deployableComponent,
			Map<Component, InternalComponentParameters> internalParameters) throws Exception {
		super(deployableComponent);

		// TODO: possible NPE
		this.wcetMap = internalParameters.get(deployableComponent).getWcetMap();
		this.energyMap = internalParameters.get(deployableComponent).getEnergyConsumptionMap();

		// Default value.
		period = 0.0;

		sil = getAnnotation(deployableComponent, SafetyIntegrityLevel.class);

		replacementComponents = new HashSet<IDeployableComponentAdapter<Component>>();
	}

	/**
	 * Creates an {@link IDeployableComponentAdapter} for an {@link Component} with the given
	 * period and replication bounds.
	 * 
	 * @throws Exception
	 */
	public AF3DeployableComponentAdapter(Component deployableComponent,
			Map<Component, InternalComponentParameters> internalParameters, Integer minReplication,
			Integer maxReplication, Double period) throws Exception {
		this(deployableComponent, internalParameters);

		this.minReplication = minReplication;
		this.maxReplication = maxReplication;
		this.period = period;
	}

	/**
	 * Creates an {@link IDeployableComponentAdapter} for an {@link Component} with the given
	 * period.
	 * 
	 * @throws Exception
	 */
	public AF3DeployableComponentAdapter(Component deployableComponent,
			Map<Component, InternalComponentParameters> internalParameters, Integer minReplication,
			Integer maxReplication, Double period,
			Collection<IDeployableComponentAdapter<Component>> componentReplacementAdapters)
			throws Exception {
		this(deployableComponent, internalParameters, minReplication, maxReplication, period);
		this.period = period;

		replacementComponents = componentReplacementAdapters;
	}

	/**
	 * Creates an "concrete" {@link IDeployableComponentAdapter} for an {@link Component} with the
	 * given period.
	 * 
	 * @param replacedComponent
	 *            References the {@link IDeployableComponentAdapter} which the to be constructed
	 *            {@link IDeployableComponentAdapter} instantiates.
	 * 
	 * @throws Exception
	 */
	public AF3DeployableComponentAdapter(Component deployableComponent,
			Map<Component, InternalComponentParameters> internalParameters,
			IDeployableComponentAdapter<Component> replacedComponent) throws Exception {
		this(deployableComponent, internalParameters, replacedComponent.getMinReplication(),
				replacedComponent.getMaxReplication(), replacedComponent
						.getTimingRequirementPeriod());

		this.replacedComponent = replacedComponent;
	}

	/** {@inheritDoc} */
	@Override
	public double getWcet(IExecutionUnitAdapter<?> executionUnit) {
		return wcetMap.get(executionUnit);
	}

	/** {@inheritDoc} */
	@Override
	public double getEnergyConsumption(IExecutionUnitAdapter<?> executionUnit) {
		return energyMap.get(executionUnit);
	}

	/** {@inheritDoc} */
	@Override
	public Enumerator getRequiredSafetyIntegrityLevel() {
		return sil.getValue();
	}

	/** {@inheritDoc} */
	@Override
	public Double getTimingRequirementPeriod() {
		return period;
	}

	/** {@inheritDoc} */
	@Override
	public void setTimingRequirementPeriod(double period) {
		this.period = period;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isInstantiationRequired() {
		return !replacementComponents.isEmpty();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isInstantiation() {
		return replacedComponent != null;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IDeployableComponentAdapter<Component>> getReplacementComponents() {
		return replacementComponents;
	}

	/** {@inheritDoc} */
	@Override
	public IDeployableComponentAdapter<Component> getReplacedComponent() {
		return replacedComponent;
	}

	/** {@inheritDoc} */
	// TODO: If possible, find a way to avoid these suppresswarnings.
	@SuppressWarnings("unchecked")
	@Override
	public void setReplacedComponent(
			@SuppressWarnings("rawtypes") IDeployableComponentAdapter abstractComponent) {
		replacedComponent = abstractComponent;
	}

	/** {@inheritDoc} */
	@Override
	public void updateInstatiationData() {
		replacedComponent = null;
	}

	/** {@inheritDoc} */
	@Override
	public int getMinReplication() {
		return minReplication;
	}

	/** {@inheritDoc} */
	@Override
	public int getMaxReplication() {
		return maxReplication;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isStronglyCausal() {
		if(modelElement.isStronglyCausal()) {
			return true;
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public AF3DeployableComponentAdapter clone() {
		try {
			return (AF3DeployableComponentAdapter)super.clone();
		} catch(CloneNotSupportedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}
}
