/*--------------------------------------------------------------------------+
$Id: TaskGraph.java 3841 2016-11-04 14:51:29Z diewald $
|                                                                          |
| Copyright 2015 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph;

import java.util.Collection;
import java.util.Set;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ISignalAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.jgrapht.DirectedGraph;
import org.jgrapht.graph.ListenableDirectedWeightedGraph;

/**
 * Interface for {@link TaskGraph}s which are the (flattened) DSE-internal representations of a
 * (hierarchical) logical architecture.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3841 $
 * @ConQAT.Rating RED Hash: 1A77DB8F1A9DB7D52AA66EEA283A0CB6
 */
public interface TaskGraph {

	/** Registers a data dependencies between a source and its destination components. */
	public boolean addDataDependency(IDeployableComponentAdapter<?> sourceComponent,
			Collection<IDeployableComponentAdapter<?>> destinationComponents,
			ISignalAdapter<?> channelAdapter);

	// TODO: Get rid of this method!
	/** Returns the underlying {@link ListenableDirectedWeightedGraph}. */
	public
			ListenableDirectedWeightedGraph<IDeployableComponentAdapter<?>, ChannelAdapterWeightedEdge>
			getGraph();

	/**
	 * Updates the period of this {@link TaskGraph}: Determine the common period of the tasks
	 * within
	 * an independent task graph if it exists. Otherwise, set it to 0.
	 * 
	 * @throws InvalidTimingModelException
	 *             if the periods of the {@link TaskGraph}'s task are inconsistent.
	 */
	public void updatePeriod() throws InvalidTimingModelException;

	/** Returns the {@link TaskGraph}s period. */
	public Double getPeriod();

	/** Returns the set of deployable components (=tasks) of this {@link TaskGraph}. */
	public Set<IDeployableComponentAdapter<?>> getTasks();

	/**
	 * Checks whether {@code this} {@link TaskGraph} contains the given
	 * {@link IDeployableComponentAdapter} or an instance of it.
	 */
	public boolean containsTask(IDeployableComponentAdapter<?> task);

	/**
	 * Checks whether {@code this} {@link TaskGraph} contains any of the given
	 * {@link IDeployableComponentAdapter}s.
	 */
	public boolean containsAnyTaskOf(Collection<IDeployableComponentAdapter<?>> taskSet);

	/**
	 * Adds the given {@link IDeployableComponentAdapter} to the {@link TaskGraph}, i.e. as a vertex
	 * to the corresponding {@link DirectedGraph}.
	 */
	public boolean addTask(IDeployableComponentAdapter<?> task);

	/**
	 * Adds the given {@code task} to the {@link TaskGraph}, i.e. as a vertex
	 * to the corresponding {@link DirectedGraph}, and further makes it accessible if the given
	 * {@code refTask} is queried.
	 */
	public boolean addTask(IDeployableComponentAdapter<?> task,
			IDeployableComponentAdapter<?> refTask);

	/** Removes the given {@code task} and its data dependencies from the {@link TaskGraph}. */
	public void removeTask(IDeployableComponentAdapter<?> task);

	/** Creates a Dialog that illustrates this {@link TaskGraph}. */
	public void display();

	/** Creates and returns a copy of {@code this} {@link TaskGraph}. */
	public <T extends TaskGraph> T copy();

	/** Registers a fault detector for a given task. */
	public void addFaultDetector(IDeployableComponentAdapter<?> deployableComponent,
			FaultDetector faultDetector);

	/** Returns a random fault detector index / ID for the given task. */
	public int getRandomFaultDetectorId(IDeployableComponentAdapter<?> deployableComponent);
}
