/*--------------------------------------------------------------------------+
$Id: MessageRouterBase.java 3062 2016-06-20 17:59:20Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.comm;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.jgrapht.DirectedGraph;
import org.jgrapht.Graph;
import org.jgrapht.graph.DefaultEdge;

/**
 * Base class which caches previously calculated routes of {@link Message}s. Varying methods to
 * calculate the routes can be implemented in the derived classes.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3062 $
 * @ConQAT.Rating YELLOW Hash: 2945C8FBB00A6399D9C77DB177644AEA
 */
public abstract class MessageRouterBase implements IMessageRouter {

	/** Representation of input system model */
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Graph containing the communication graph of the platform onto which the deployment is done. */
	protected Graph<IResourceAdapter<?>, DefaultEdge> platformCommunicationGraph;

	/** Contains the calculated routes of the generated {@link Message}s. */
	private Map<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>> messagesRoutes;

	/** Constructs a scheduler for a given input model. */
	protected MessageRouterBase(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter) {
		this.systemModelAdapter = systemModelAdapter;
		messagesRoutes =
				new ConcurrentHashMap<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>>();
		platformCommunicationGraph =
				systemModelAdapter.getPlatformCommunicationGraph().getActualGraph();
	}

	/** {@inheritDoc} */
	@Override
	public synchronized Map<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>>
			calculateMessageRoutes(Collection<Message> messagesToRoute) {
		Map<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>> messagesRoutes =
				new HashMap<Message, DirectedGraph<IResourceAdapter<?>, DefaultEdge>>();

		// Calculate the route of each message and associate the route with the message.
		for(Message currentMessage : messagesToRoute) {
			DirectedGraph<IResourceAdapter<?>, DefaultEdge> calculatedMessageRoute =
					calculateMessageRoute(currentMessage.getSender().getTarget(),
							currentMessage.getReceiverResources());
			// Check if any *invalid* communication were required by the mapping.
			if(calculatedMessageRoute == null) {
				return null;
			}
			messagesRoutes.put(
					currentMessage,
					calculateMessageRoute(currentMessage.getSender().getTarget(),
							currentMessage.getReceiverResources()));
		}

		return messagesRoutes;
	}

	/** {@inheritDoc} */
	@Override
	public DirectedGraph<IResourceAdapter<?>, DefaultEdge> getRouteOf(Message msg) {
		return messagesRoutes.get(msg);
	}
}
