/*--------------------------------------------------------------------------+
$Id: FaultDetectionVotingEDFScheduler.java 3032 2016-06-16 18:13:12Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.sysmodel.sched.proc;

import java.util.Map;

import org.apfloat.Apfloat;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskgraph.InstantiatedAcyclicTaskGraphEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FaultDetectionVotingTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.FaultDetectionVotingTaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IDeployableComponentAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.ChannelAdapterWeightedEdge;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.taskgraph.TaskGraph;
import org.fortiss.af3.exploration.alg.dse.sysmodel.faultmodel.FaultDetector;
import org.fortiss.af3.exploration.model.ExplorationSpecification;

/**
 * Base class for EDF task schedulers
 * 
 * @author huang, barner
 * @author $Author: diewald $
 * @version $Rev: 3032 $
 * @ConQAT.Rating RED Hash: DE3830D2F2585C406513F8918AE57716
 */
public class FaultDetectionVotingEDFScheduler
		extends
		EDFSchedulerBase<FaultDetectionVotingTaskMappingEntry, FaultDetectionVotingTaskMappingEncoding> {

	/** Constructs a simple EDF scheduler for a given exploration run */
	public FaultDetectionVotingEDFScheduler(
			SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
			ExplorationSpecification expSpec) {
		super(systemModelAdapter, expSpec);
	}

	/**
	 * {@inheritDoc}.
	 * <p>
	 * For {@link FaultDetectionVotingEDFScheduler}, this method injects voters (deployed with the
	 * same deployment target as the sender) that are required to reduce the output of replicated
	 * sender components.
	 * </p>
	 */
	@Override
	protected
			Apfloat
			registerIncomingMessagesWithSchedule(
					FaultDetectionVotingTaskMappingEncoding encoding,
					InstantiatedAcyclicTaskGraphEncoding<FaultDetectionVotingTaskMappingEntry, ?> acycItgEnc,
					StrictTTSchedule<FaultDetectionVotingTaskMappingEntry, FaultDetectionVotingTaskMappingEncoding> schedule,
					Map<FaultDetectionVotingTaskMappingEntry, TimeSlot> mappingEntryToTimeSlot,
					IDeployableComponentAdapter<?> receiver, Apfloat deployableComponentStartTime,
					IExecutionUnitAdapter<?> deploymentTarget, long iteration, TimeSlot slot) {

		Apfloat voterDelay = new Apfloat(0, DOUBLE_PRECISION);

		TaskGraph taskGraph = acycItgEnc.getTaskGraphOf(receiver);
		for(ChannelAdapterWeightedEdge edge : taskGraph.getGraph().incomingEdgesOf(receiver)) {
			IDeployableComponentAdapter<?> sender = taskGraph.getGraph().getEdgeSource(edge);

			// See if the sender has been replicated
			// TODO: Remove the need to use the encoding here: can the information be obtained in
			// another fashion?
			if(encoding.getMappingEntriesOf(sender) != null &&
					encoding.getMappingEntriesOf(sender).size() > 1) {

				// Add voter to reduce result produced by replicated senders.
				// TODO currently, the voter execution time is assumed to be 10 us, and it needs to
				// be specified in nano seconds
				Apfloat voterExecutionTime = new Apfloat(10.0 * 1000);

				// Compute the starting time of the voter. We add 'voterDelay' that accumulates
				// the total delay caused by voters added in previous iterations of this loops
				// (i.e.,
				// for incoming messages from other replicated senders)
				Apfloat tempTimp =
						deployableComponentStartTime.add(voterExecutionTime).add(voterDelay);

				// Accumulate delay caused by all voters
				voterDelay = voterDelay.add(voterExecutionTime);

				// TODO: Create a class that represents voters similar to
				// IDeployableComponentAdapter.
				@SuppressWarnings("unused") TimeSlot voterSlot =
						schedule.createVoterSlot(null, deploymentTarget,
								deployableComponentStartTime.add(voterDelay).floatValue(),
								tempTimp.floatValue(), iteration);

			}
		}

		return voterDelay;
	}

	/** {@inheritDoc} */
	@Override
	protected
			Apfloat
			getExecutionTime(
					TaskMappingEncoding<? extends TaskMappingEntry> encoding,
					InstantiatedAcyclicTaskGraphEncoding<FaultDetectionVotingTaskMappingEntry, ?> acycItgEnc,
					IDeployableComponentAdapter<?> deployableComponent,
					FaultDetectionVotingTaskMappingEntry entry) {

		// Base execution time
		Apfloat executionTime =
				new Apfloat(deployableComponent.getWcet(entry.getTarget()), DOUBLE_PRECISION);

		// Scale with fault-detector overhead
		FaultDetector detector =
				acycItgEnc.getTaskGraphOf(deployableComponent).getFaultDetector(
						deployableComponent, entry.getFaultDetectorID());
		Apfloat overhead = new Apfloat(1.0 + detector.getOverhead(), DOUBLE_PRECISION);

		executionTime = executionTime.multiply(overhead);

		return executionTime;
	}
}
