/*--------------------------------------------------------------------------+
$Id: ObjectJGraphTVisualizer.java 3449 2016-09-02 11:06:00Z diewald $
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.graph.display;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.swing.JFrame;

import org.jgrapht.DirectedGraph;
import org.jgrapht.Graph;
import org.jgrapht.ext.JGraphXAdapter;
import org.jgrapht.graph.DefaultDirectedGraph;
import org.jgrapht.graph.DefaultEdge;

import com.mxgraph.layout.mxFastOrganicLayout;
import com.mxgraph.swing.mxGraphComponent;
import com.mxgraph.util.mxConstants;

/**
 * Used for visualizing graphs (from JGraphT). To display a graph, this class must be instantiated
 * first, and then the {@code displayGraph} method must be called.
 * Used to display generic Java Objects.
 * 
 * Usage Example:
 * {@code ObjectJGraphTVisualizer<IDeployableComponentAdapter, ChannelAdapterWeightedEdge> taskViz =
					new ObjectJGraphTVisualizer<IDeployableComponentAdapter, ChannelAdapterWeightedEdge>(
							taskGraph.getGraph());
			taskViz.displayGraph(); }
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3449 $
 * @ConQAT.Rating YELLOW Hash: AF7DE016F64BD089300DB13C81607FE5
 */
public class ObjectJGraphTVisualizer<V, E extends DefaultEdge> {

	/**
	 * Contains the graph which was given to the Constructor, except that the vertices have been
	 * converted to Named vertices to support comprehensive visualization.
	 */
	private DirectedGraph<ObjectVertex<V>, DefaultEdge> namedGraph;

	/** Constructor. */
	public ObjectJGraphTVisualizer(Graph<V, E> g) {
		namedGraph = constructNamedGraph(g);
	}

	/** Displays the graph given in the constructor. */
	public void displayGraph(String graphWindowTitle) {
		// Create an adapter for JGraphX, via an adapter.
		JGraphXAdapter<ObjectVertex<V>, DefaultEdge> jgxAdapter = new JGraphXAdapter<>(namedGraph);

		// Disable labels of edges.
		jgxAdapter.getStylesheet().getDefaultEdgeStyle().put(mxConstants.STYLE_NOLABEL, "1");

		// positioning via jgraphx layouts
		mxFastOrganicLayout layout = new mxFastOrganicLayout(jgxAdapter);
		// layout.setMinDistanceLimit(layout.getMinDistanceLimit() * 2);
		// The higher, the more separated layout.
		layout.setForceConstant(layout.getForceConstant() * 1.5);
		layout.execute(jgxAdapter.getDefaultParent());

		mxGraphComponent graphComponent = new mxGraphComponent(jgxAdapter);

		// Draws the window containing the graph.
		JFrame frame = new JFrame();
		frame.setTitle(graphWindowTitle);
		frame.getContentPane().add(graphComponent);
		frame.pack();
		frame.setVisible(true);

		// mxGraphView view = graphComponent.getGraph().getView();
		// int compLen = graphComponent.getWidth();
		// int viewLen = (int)view.getGraphBounds().getWidth();
		// view.setScale((double)compLen / viewLen * view.getScale());
	}

	/**
	 * Converts the given graph to graph that has {@link NamedVertex}-Nodes instead of the original
	 * vertices. Displaying those is more convenient.
	 */
	private DirectedGraph<ObjectVertex<V>, DefaultEdge> constructNamedGraph(
			Graph<V, E> originalGraph) {
		DirectedGraph<ObjectVertex<V>, DefaultEdge> namedGraph =
				new DefaultDirectedGraph<>(DefaultEdge.class);
		Set<E> originalEdgeSet = originalGraph.edgeSet();
		Map<V, ObjectVertex<V>> relateVertexToNamedElement = new HashMap<>();

		for(V currentVertex : originalGraph.vertexSet()) {
			ObjectVertex<V> currentNamedVertex = new ObjectVertex<V>(currentVertex);
			namedGraph.addVertex(currentNamedVertex);
			relateVertexToNamedElement.put(currentVertex, currentNamedVertex);
		}

		for(E edge : originalEdgeSet) {
			namedGraph.addEdge(relateVertexToNamedElement.get(originalGraph.getEdgeSource(edge)),
					relateVertexToNamedElement.get(originalGraph.getEdgeTarget(edge)));
		}

		return namedGraph;
	}

}
