/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.guava;

import java.util.Map;
import java.util.Set;

import com.google.common.collect.BiMap;
import com.google.common.collect.ForwardingMap;
import com.google.common.collect.HashBiMap;

/**
 * {@link BiMap} implementation whose keys and values are pairs of an object and an associated type.
 * This {@link BiMap} is symmetric: It allows references of multiple keys to the same object iff
 * the keys have different types. Hence, the 1:1 association between keys and values of
 * {@link BiMap}s is preserved.
 * 
 * For an extended documentation please see {@link BiMap}.
 * 
 * @author diewald
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public class HashTypedBiMap<K, V> extends ForwardingMap<TypedPair<K, V>, TypedPair<V, K>> implements
		TypedBiMap<K, V>/* , BiMap<TypedPair<K, V>, TypedPair<V, K>> */{

	/**
	 * Returns a new, empty {@code HashBiMap} with the default initial capacity (16).
	 */
	public static <K, V> HashTypedBiMap<K, V> create() {
		return create(16);
	}

	/**
	 * Constructs a new, empty bimap with the specified expected size.
	 *
	 * @param expectedSize
	 *            the expected number of entries
	 * @throws IllegalArgumentException
	 *             if the specified expected size is negative
	 */
	public static <K, V> HashTypedBiMap<K, V> create(int expectedSize) {
		return new HashTypedBiMap<K, V>(expectedSize);
	}

	/**
	 * Constructs a new bimap containing initial values from {@code map}. The bimap is created with
	 * an
	 * initial capacity sufficient to hold the mappings in the specified map.
	 */
	public static <K, V> HashTypedBiMap<K, V> create(
			Map<? extends TypedPair<K, V>, ? extends TypedPair<V, K>> map) {
		HashTypedBiMap<K, V> bimap = create(map.size());
		bimap.putAll(map);
		return bimap;
	}

	/** {@link BiMap} holding the actual entries of {@code this} map. */
	protected BiMap<TypedPair<K, V>, TypedPair<V, K>> delegateMap;

	/**
	 * Provides the inverse view on {@code this} {@link TypedBiMap} that is backed by the same
	 * delegate {@link BiMap}.
	 */
	protected TypedBiMap<V, K> inverseView;

	/** Hidden constructor to be called from the static create methods. */
	private HashTypedBiMap(int expectedSize) {
		this(expectedSize, null, null);
	}

	/** Hidden constructor to be called from the static create methods. */
	private HashTypedBiMap(int expectedSize, TypedBiMap<V, K> inverseView,
			BiMap<TypedPair<K, V>, TypedPair<V, K>> delegateMap) {
		if(inverseView == null) {
			this.delegateMap = HashBiMap.create(expectedSize);
			this.inverseView =
					new HashTypedBiMap<V, K>(expectedSize, this, this.delegateMap.inverse());
		} else {
			this.inverseView = inverseView;
			this.delegateMap = delegateMap;
		}
	}

	/** {@inheritDoc} */
	@Override
	protected BiMap<TypedPair<K, V>, TypedPair<V, K>> delegate() {
		return delegateMap;
	}

	/** {@inheritDoc} */
	@Override
	public V get(K key, Class<? extends V> valType) {
		@SuppressWarnings("unchecked") TypedPair<V, K> retVal =
				delegate().get(new TypedPair<K, V>(key, (Class<V>)valType));
		return retVal != null ? retVal.getFirst() : null;
	}

	/** {@inheritDoc} */
	@Override
	public V put(K key, Class<? extends V> valType, V value, Class<? extends K> keyType) {
		@SuppressWarnings("unchecked") TypedPair<V, K> oldPair =
				delegate().forcePut(new TypedPair<K, V>(key, (Class<V>)valType),
						new TypedPair<V, K>(value, (Class<K>)keyType));
		return oldPair != null ? oldPair.getFirst() : null;
	}

	/** {@inheritDoc} */
	@Override
	public TypedPair<V, K> forcePut(TypedPair<K, V> key, TypedPair<V, K> value) {
		return delegate().forcePut(key, value);
	}

	/** {@inheritDoc} */
	@Override
	public Set<TypedPair<V, K>> values() {
		return delegate().values();
	}

	/** {@inheritDoc} */
	@Override
	public TypedBiMap<V, K> inverse() {
		return inverseView;
	}
}
