/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.guava;

import com.google.common.collect.BiMap;

/**
 * {@link BiMap} implementation whose keys and values are pairs of an object and an associated type.
 * This {@link BiMap} is symmetric: It allows references of multiple keys to the same object iff
 * the keys have different types. Hence, the 1:1 association between keys and values of
 * {@link BiMap}s is preserved.
 * 
 * @author diewald
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash: BBD46247817C0DEFFDA8BB8F76B2215B
 */
public interface TypedBiMap<K, V> extends BiMap<TypedPair<K, V>, TypedPair<V, K>> {

	/**
	 * Returns the value from the map which is referenced by the given key and which has the
	 * given type.
	 * In a {@link TypedBiMap} a key may reference multiple values that have a different type.
	 * Please see {@link BiMap#get(Object)} for extended documentation on the put method.
	 * 
	 * @param key
	 *            identifies a set of values with their types.
	 * @param valType
	 *            Type of the value to return.
	 * @return value uniquely identified by the key and the value type.
	 */
	V get(K key, Class<? extends V> valType);

	/**
	 * Adds the given {@code value} to {@code this} {@link TypedBiMap}. It is uniquely identified by
	 * the {@code key} and the type of the {@code value} ({@code valueType}).
	 * In order to allow an inverse lookup, the keys must be also associated with a type (
	 * {@code keyType}).
	 * For more details about {@link BiMap}s, please consult {@link BiMap#put(Object, Object)}
	 * 
	 * @param key
	 *            The referencing key object.
	 * @param valType
	 *            Type of the value.
	 * @param value
	 *            The value object placed in the map.
	 * @param keyType
	 *            Type of the key object.
	 * @return The old value uniquely identified by the {@code key} and the {@code valueType}.
	 */
	V put(K key, Class<? extends V> valType, V value, Class<? extends K> keyType);

	/** {@inheritDoc} */
	@Override
	TypedBiMap<V, K> inverse();
}
