/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.service.internal;

import static org.fortiss.af3.exploration.alg.service.ExplorationService.getService;
import static org.fortiss.tooling.common.util.LambdaUtils.filterSet;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.conqat.lib.commons.reflect.TypeConversionException;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.evaluate.ComposableEvaluator;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.problem.EvaluatorWrapper;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationServiceException;
import org.fortiss.af3.exploration.alg.service.IExplorationEvaluatationService;
import org.fortiss.af3.exploration.alg.service.IExplorationService;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.MOExplorationObjective;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;
import org.opt4j.core.Genotype;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

/**
 * Implementation of the {@link IExplorationEvaluatationService} which handles the instantiation of
 * the Evaluators based on the Exploration target. The evaluators are registered as extensions.
 * 
 * @author diewald
 * @author $Author: hoelzl $
 * @version $Rev: 18709 $
 * @ConQAT.Rating YELLOW Hash: 6DDFF3B8BB2691518FF2117434B051E1
 */
public class ExplorationEvaluationService<T extends IExplorationTargetEvaluator<?, ?, ?>> extends
		EObjectAwareServiceBase<T> implements IExplorationEvaluatationService<T> {

	/** The evaluator provider extension point ID. */
	private static final String EXTENSION_POINT_NAME = "org.fortiss.af3.exploration.alg.evaluator";

	/** The evaluator provider configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "evaluator";

	/** The evaluator class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "module";

	/** Holds the instance of the evaluation service. */
	private volatile static ExplorationEvaluationService<?> INSTANCE;

	/**
	 * Associates the registered {@link ComposableEvaluator}s with their required types from the
	 * {@link IExplorationTransformationService}
	 */
	private Multimap<Class<IExplorationTargetEvaluator<?, ?, ?>>, Class<?>> requiredTransformedModels =
			HashMultimap.create();

	/** Constructor. */
	private ExplorationEvaluationService() {
		// Prevent direct instantiation.
	}

	/** The instance (singleton) of the ExplorationTransformationService. */
	@SuppressWarnings("rawtypes")
	public static synchronized ExplorationEvaluationService<?> getInstance() {
		if(INSTANCE == null) {
			INSTANCE = new ExplorationEvaluationService();
		}
		return INSTANCE;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return "This services enables the registration of evaluators (quantifiers) to the DSE.";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Evaluator Service";
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}

	/** {@inheritDoc} */
	// NB: Each {@link ExplorationTarget} may have only _ONE_ {@link Evaluator}.
	@Override
	protected boolean requiresUniqueHandler() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Collection<Class<?>> getRequiredTransformedModels(
			Set<IExplorationTargetEvaluator<?, ?, ?>> evaluators) {
		Set<Class<?>> retVal = new HashSet<>();
		evaluators.forEach(e -> retVal.addAll(requiredTransformedModels
				.get((Class<IExplorationTargetEvaluator<?, ?, ?>>)e.getClass())));
		return retVal;
	}

	/** {@inheritDoc} */
	@Override
	public void startService() {
		super.startService();

		Collection<Class<?>> allRequiredInputs = new HashSet<>();

		for(IExplorationTargetEvaluator<?, ?, ?> evaluator : getInstance().getAllHandlers()) {
			allRequiredInputs.addAll(evaluator.getRequiredInputTypes());
			// @CodeFormatterOff
			Collection<Class<?>> curNonPhenotypeInputTypes =
					filterSet(allRequiredInputs,
							t -> !(ComposablePhenotype.class.isAssignableFrom(t)
								   || Genotype.class.isAssignableFrom(t)));
			// @CodeFormatterOn

			// TODO: check also if the phenotypes exist.
			try {
				checkInputConsistency(curNonPhenotypeInputTypes);
			} catch(ExplorationServiceException e) {
				throw new RuntimeException("Error starting the ExplorationEvaluationService.", e);
			}

			// Here, we assume that all *checked* (see above) remaining models must be provided by
			// the transformation service.
			@SuppressWarnings("unchecked") Class<IExplorationTargetEvaluator<?, ?, ?>> evaluatorType =
					(Class<IExplorationTargetEvaluator<?, ?, ?>>)evaluator.getClass();
			requiredTransformedModels.putAll(evaluatorType, curNonPhenotypeInputTypes);
		}
	}

	/**
	 * Returns whether the given collection of non-Phenotype types can be obtained by the
	 * {@link IExplorationTransformationService}.
	 */
	@SuppressWarnings("unchecked")
	private void checkInputConsistency(Collection<Class<?>> requiredNonPhenoInputs)
			throws ExplorationServiceException {
		for(Class<?> curInputType : requiredNonPhenoInputs) {
			if(!getService(IExplorationTransformationService.class).canTransformOutputType(
					curInputType)) {
				throw new ExplorationServiceException(
						(Class<? extends IExplorationService<?>>)IExplorationEvaluatationService.class,
						"The required input type " + curInputType.getSimpleName() +
								" is not a Phenotype, nor can it be produced by the" +
								" IExplorationTransformationService. Please check the required" +
								" types and/or the registration of the needed ITransformationStrategies.");
			}
		}
	}

	/**
	 * Creates an {@link EvaluatorWrapper} for the given {@link ExplorationTarget} if a handler has
	 * been registered for it.
	 * 
	 * @param systemModelAdapter
	 *            {@link SystemModelAdapter} abstracting the base models (e.g., technical
	 *            architecture).
	 * @param explorationTarget
	 *            {@link ExplorationTarget} to create an {@link EvaluatorWrapper} for.
	 * @return The instantiated {@link EvaluatorWrapper}.
	 * @throws TypeConversionException
	 *             If some model element cannot be translated to the internal representation of the
	 *             DSE.
	 * @throws InstantiationException
	 *             If the instantiation of the {@link EvaluatorWrapper} fails.
	 */
	public <P, RT, ET extends ExplorationTarget<RT>> EvaluatorWrapper<P, RT, ET>
			createEvalWrapperFor(SystemModelAdapter<?, ?, ?, ?, ?, ?> systemModelAdapter,
					ET explorationTarget) throws TypeConversionException, InstantiationException {
		@SuppressWarnings("unchecked") Collection<IExplorationTargetEvaluator<P, RT, ET>> evaluatorHandlers =
				(Collection<IExplorationTargetEvaluator<P, RT, ET>>)getInstance()
						.getRegisteredHandlers(explorationTarget.getClass());
		if(evaluatorHandlers != null && evaluatorHandlers.size() > 0) {
			// assert (evaluatorHandlers.size() > 0) : "No evaluator registered for " +
			// explorationTarget.getClass().getSimpleName() + ".";
			assert (evaluatorHandlers.size() <= 1) : "Multiple evaluators registered for " +
					explorationTarget.getClass().getSimpleName() + ".";

			// The "template" evaluator.
			IExplorationTargetEvaluator<P, RT, ET> evaluatorHandler =
					evaluatorHandlers.stream().findAny().get();
			// The actual instance used for evaluation.
			IExplorationTargetEvaluator<P, RT, ET> evaluator =
					evaluatorHandler.createEvaluator(systemModelAdapter, explorationTarget);

			@SuppressWarnings("unchecked") Class<P> primaryInputType =
					(Class<P>)evaluatorHandler.getPrimaryInputType();
			assert (Phenotype.class.isAssignableFrom(primaryInputType)) : "The primary input type is not a genotype. Only Genotypes are currently supported.";

			int priority = 9999;
			if(explorationTarget instanceof MOExplorationObjective<?>) {
				priority = ((MOExplorationObjective<?>)explorationTarget).getPriority();
			}

			return new EvaluatorWrapper<P, RT, ET>(explorationTarget, evaluator, priority,
					primaryInputType);
		}

		return null;
	}
}
