/*--------------------------------------------------------------------------+
$Id: ExplorationAlgUtils.java 3788 2016-10-27 08:23:59Z barner $
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.util;

import static org.eclipse.emf.ecore.util.EcoreUtil.getAllContents;
import static org.fortiss.af3.exploration.alg.util.AF3Utils.isAnyParentDeploymentTarget;
import static org.fortiss.af3.platform.utils.PlatformArchitectureUtils.getReferencedElementsWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getParentsWithType;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;
import java.util.function.Predicate;
import java.util.stream.Stream;

import org.conqat.lib.commons.reflect.ReflectionUtils;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.component.model.Channel;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.TimeSlot;
import org.fortiss.af3.exploration.alg.exception.InvalidTimingModelException;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.IProblemDimension;
import org.fortiss.af3.exploration.util.ExplorationUtils;
import org.fortiss.af3.platform.model.IPlatformResource;
import org.fortiss.af3.platform.model.IVirtualizationPlatformArchitectureElement;
import org.fortiss.af3.platform.model.annotation.ResourceLink;
import org.fortiss.af3.platform.utils.PlatformArchitectureUtils;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.base.model.element.IHierarchicElementContainer;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.model.element.IModelElementSpecification;

import com.google.common.collect.Multimap;

/**
 * Static utility methods mainly used in the Design Space Exploration.
 * 
 * @author diewald
 * @author $Author: barner $
 * @version $Rev: 3788 $
 * @ConQAT.Rating RED Hash: 39BD65A7E68F663DC99B74FF4827C558
 */
public class ExplorationAlgUtils {

	/** Constructor. Prevent initialization. */
	private ExplorationAlgUtils() {
	}

	/**
	 * Returns the given Collection as a Collection of the raw types of its elements.
	 * 
	 * @param origCollection
	 *            Collection to be casted.
	 * @return Collection of the raw types of the input collection, or {@code null} if the
	 *         Collection cannot be created.
	 */
	@SuppressWarnings("unchecked")
	public static <C extends Collection<T>, T, CR extends Collection<? super T>> CR castToRawColl(
			C origCollection) {
		CR castedCollection = (CR)Collections.emptySet();
		if(origCollection != null && !origCollection.isEmpty()) {
			try {
				castedCollection = (CR)origCollection.getClass().newInstance();
				((Collection<T>)castedCollection).addAll(origCollection);
			} catch(InstantiationException | IllegalAccessException e) {
				// Never here.
				e.printStackTrace();
			}
		}
		return castedCollection;
	}

	/**
	 * Returns the given Collection as a Collection of the raw types of its elements.
	 * 
	 * @param origCollection
	 *            Collection to be casted.
	 * @return Collection of the raw types of the input collection, or {@code null} if the
	 *         Collection cannot be created.
	 */
	@SuppressWarnings("unchecked")
	public static <C extends Collection<T>, T, CR extends Collection<? super T>> CR castToRawColl(
			C origCollection, Class<CR> castedCollType) {
		CR castedCollection = (CR)Collections.emptySet();
		if(origCollection != null && !origCollection.isEmpty()) {
			try {
				castedCollection = castedCollType.newInstance();
				((Collection<T>)castedCollection).addAll(origCollection);
			} catch(InstantiationException | IllegalAccessException e) {
				// Never here.
				e.printStackTrace();
			}
		}
		return castedCollection;
	}

	/**
	 * For a given Object, determine whether it is a sub-class of any of the classes contained in
	 * the given collection.
	 */
	public static boolean isAssignableFromAny(Object obj, Collection<Class<?>> classes) {
		return classes.stream().anyMatch(cls -> cls.isAssignableFrom(obj.getClass()));
	}

	// // TODO: move to dedicated AF3 adapter class, or get the set dynamically from the adapters
	// /** Contains a collection of platform elements which are connected via directed connections.
	// */
	// public static Collection<Class<?>> directedConnectionElements = Collections
	// .unmodifiableSet(new HashSet<Class<?>>() {
	// {
	// add(NocRouter.class);
	// }
	// });

	/**
	 * Finds the (target) {@link InputPort}s to which the given {@link OutputPort} is connected.
	 * 'Target Ports' are connected {@link InputPort}s of {@link Component}s from the set of
	 * {@code deployableComponents}.
	 * 
	 * @param sourcePort
	 *            The port to start the search from.
	 * @param deployableComponents
	 *            Collection of 'deployable' {@link Component}s present in the input logical
	 *            architecture.
	 * @return The list of target ports of atomic {@link Component}s.
	 */
	public static List<Port> findTargetPorts(OutputPort sourcePort,
			Collection<Component> deployableComponents) {
		if(!deployableComponents.contains(sourcePort.getComponent())) {
			return Collections.emptyList();
		}
		List<Port> targets = new ArrayList<>();
		LinkedList<Channel> exploreChannels = new LinkedList<Channel>();
		exploreChannels.addAll(sourcePort.getOutgoingChannels());
		while(!exploreChannels.isEmpty()) {
			Channel currentExploreChannel = exploreChannels.removeFirst();
			Component currentTargetComponent = currentExploreChannel.getTarget().getComponent();
			if(deployableComponents.contains(currentTargetComponent)) {
				targets.add(currentExploreChannel.getTarget());
			} else {
				exploreChannels.addAll(currentExploreChannel.getTarget().getOutgoingChannels());
			}
		}
		return targets;
	}

	/**
	 * Finds the (source) {@link OutputPort}s to which the given {@link InputPort} is connected. The
	 * search follows all channels to input ports of deployable/atomic {@link Component}s. The
	 * source port must be attached to an deployable/atomic {@link Component}, otherwise this method
	 * returns an empty list.
	 * 
	 * @param sourcePort
	 *            The port to start the search from.
	 * @return The list of target ports of atomic {@link Component}s.
	 * @throws Exception
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	// FIXME: HERE
	// public static List<OutputPort> findSourcePorts(InputPort sourcePort) throws Exception {
	// if(getEventAnnotationOf(sourcePort) == null &&
	// !isDefinedAsDeployableElement(sourcePort.getComponent())) {
	// return Collections.emptyList();
	// }
	// List<OutputPort> sources = new ArrayList<OutputPort>();
	// LinkedList<Channel> exploreChannels = new LinkedList<Channel>();
	// exploreChannels.addAll(sourcePort.getIncomingChannels());
	// while(!exploreChannels.isEmpty()) {
	// Channel currentExploreChannel = exploreChannels.removeFirst();
	// Component currentSourceComponent = currentExploreChannel.getSource().getComponent();
	// if(/* getEventAnnotationOf(currentExploreChannel.getTarget()) != null || */
	// isDefinedAsDeployableElement(currentSourceComponent)) {
	// sources.add((OutputPort)currentExploreChannel.getSource());
	// } else {
	// exploreChannels.addAll(currentExploreChannel.getSource().getIncomingChannels());
	// }
	// }
	// return sources;
	// }

	/**
	 * Returns a list of source ports within the given {@link ComponentArchitecture}. Source ports
	 * are defined as ports that do not have direct inputs.
	 * 
	 * @param componentArchitecture
	 *            The {@link ComponentArchitecture} for whom source ports shall be identified.
	 * @return {@link EList} of source {@link OutputPort}s.
	 * @throws InvalidTimingModelException
	 *             if the timing specification of any traversed element is inconsistent.
	 */
	public static EList<OutputPort> getSourcePorts(ComponentArchitecture componentArchitecture)
			throws InvalidTimingModelException {
		if(componentArchitecture == null) {
			return null;
		}

		EList<OutputPort> allOutputPortsInArch =
				getChildrenWithType(componentArchitecture, OutputPort.class);
		EList<OutputPort> sourcePortList = new BasicEList<OutputPort>();
		for(OutputPort currentOutputPort : allOutputPortsInArch) {
			// Filter out ports that represent the transition through a hierarchic hull.
			if(currentOutputPort.getIncoming().isEmpty() /*
														 * &&
														 * !isAnyParentComponentDeployable(
														 * currentOutputPort.getComponent())
														 */) {
				// FIXME: HERE
				sourcePortList.add(currentOutputPort);
			}
		}

		return sourcePortList;
	}

	/**
	 * Returns the set of {@link OutputPort}s of all given 'deployable' {@link Component}s.
	 * 
	 * @param deployableComponents
	 *            Collection of all {@link OutputPort}s of 'deployable' {@link Component}s.
	 * @return List of 'deployable' {@link OutputPort}s or an empty list if none can be identified.
	 */
	public static EList<OutputPort> getDeployableOutputPorts(
			Collection<Component> deployableComponents) {
		EList<OutputPort> sourcePortList = new BasicEList<OutputPort>();

		for(Component component : deployableComponents) {
			sourcePortList.addAll(component.getOutputPorts());
		}

		return sourcePortList;
	}

	/**
	 * Returns all Objects from the given Collection which inherit from the given class.
	 * 
	 * @param targetClass
	 *            Class providing the type filer.
	 * @param sourceColl
	 *            Collection of Objects to be filtered.
	 * @return Collection of objects inheriting from the given class.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Collection<T> pickInstanceOf(Class<T> targetClass, Collection<?> sourceColl) {
		Collection<T> filteredColl = new ArrayList<>();
		for(Object srcObj : sourceColl) {
			if(targetClass.isAssignableFrom(srcObj.getClass())) {
				filteredColl.add((T)srcObj);
			}
		}
		return filteredColl;
	}

	/**
	 * Returns the first found object from the given collection that inherits from the given class.
	 * 
	 * @param targetClass
	 *            Type to check the inheritance of the collection's objects.
	 * @param sourceColl
	 *            Objects to examine.
	 * @return The first found object inheriting from the given class, or {@code null} if none is
	 *         found.
	 */
	@SuppressWarnings("unchecked")
	public static <T> T pickFirstInstanceOf(Class<T> targetClass, Collection<?> sourceColl) {
		for(Object srcObj : sourceColl) {
			if(targetClass.isAssignableFrom(srcObj.getClass())) {
				return (T)srcObj;
			}
		}
		return null;
	}

	/**
	 * Returns all Classes from the given Collection which inherit from the given class.
	 * 
	 * @param targetClass
	 *            Class providing the type filer.
	 * @param sourceTypeColl
	 *            Collection of Classes to be filtered.
	 * @return Collection of Classes inheriting from the given class.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Collection<Class<T>> pickSubTypesOf(Class<T> targetClass,
			Collection<Class<?>> sourceTypeColl) {
		Collection<Class<T>> filteredColl = new ArrayList<>();
		for(Class<?> srcType : sourceTypeColl) {
			if(targetClass.isAssignableFrom(srcType)) {
				filteredColl.add((Class<T>)srcType);
			}
		}
		return filteredColl;
	}

	/**
	 * Returns all Objects from the given Collection which inherit from the given collection of
	 * classes.
	 * 
	 * @param filterClasses
	 *            Class providing the type filer.
	 * @param unfilteredCollection
	 *            Collection of Objects to be filtered.
	 * @return Collection of Classes that do not inherit from the given class.
	 */
	public static <T, S extends T> Collection<T> filterByTypeUnsafe(
			Collection<Class<?>> filterClasses, Collection<T> unfilteredCollection) {
		if(unfilteredCollection == null) {
			return null;
		}

		Collection<T> filteredCollection = new ArrayList<>();
		for(T element : unfilteredCollection) {
			if(filterClasses.stream().anyMatch(c -> c.isAssignableFrom(element.getClass()))) {
				filteredCollection.add(element);
			}
		}
		return filteredCollection;
	}

	/**
	 * Filters all Classes from the given Collection which inherit from the given class.
	 * 
	 * @param targetClass
	 *            Class providing the type filer.
	 * @param sourceTypeColl
	 *            Collection of Classes to be filtered.
	 * @return Collection of Classes that do not inherit from the given class.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Collection<Class<T>> filterOutSubTypesOf(Class<T> targetClass,
			Collection<Class<?>> sourceTypeColl) {
		Collection<Class<T>> filteredColl = new ArrayList<>();
		for(Class<?> srcType : sourceTypeColl) {
			if(!targetClass.isAssignableFrom(srcType)) {
				filteredColl.add((Class<T>)srcType);
			}
		}
		return filteredColl;
	}

	/**
	 * Determines whether the given {@link IHierarchicElement} is relevant for the
	 * deployment exploration (atomic communication resources, deployment targets etc).
	 * 
	 * @param element
	 *            The element to be examined.
	 * @return Boolean indicating the relevance of of the given {@link IHierarchicElement} for the
	 *         DSE.
	 */
	public static boolean isPlatformTargetResource(IPlatformResource element,
			Collection<IPlatformResource> targetResources) {
		if(!isAnyParentDeploymentTarget(element, targetResources) &&
				(element instanceof IHierarchicElementContainer && ((IHierarchicElementContainer)element)
						.getContainedElements().isEmpty()) || targetResources.contains(element)) {
			return true;
		}
		return false;
	}

	/**
	 * Determines whether an {@link IPlatformResource} is a virtual platform element (like a
	 * partition).
	 * 
	 * @param element
	 *            The element to be examined.
	 * @return see above
	 */
	public static boolean isVirtualResource(IHierarchicElement element) {
		if(element instanceof IVirtualizationPlatformArchitectureElement) {
			return true;
		}

		return false;
	}

	// /**
	// * Determines whether the given element is relevant for a deployment. It examines whether the
	// * given element is a sub-element of a deployment target or a sub-element of a virtualized
	// * hardware element. If neither applies, the element is not part of a 'black box' and, hence,
	// it
	// * is 'relevant'.
	// *
	// * @param element
	// * The platform element to be examined.
	// * @param virtualizationProviders
	// * The list of all present virtualization providers (hypervisors).
	// * @return Boolean indicating the relevance for of the element for a deployment.
	// */
	// public static boolean isPlatformTargetResource(IPlatformResource element,
	// Collection<IPlatformResource> deploymentTargets,
	// Collection<IVirtualizationPlatformArchitectureElement> virtualizationProviders) {
	// if(isPlatformTargetResource(element, deploymentTargets) &&
	// (!hasVirtualizationLayer(element, virtualizationProviders)) ||
	// isReferencedByNetworkDriver(element, virtualizationProviders)) {
	// return true;
	// }
	// return false;
	// }
	//
	// /**
	// * Determines whether the given element is referenced by network driver. Useful to determine
	// * {@link GatewayUnit}s that would not be considered in the communication structure, if it
	// were
	// * "hidden" by a parent element.
	// *
	// * @param element
	// * The {@link IHierarchicElement} to be examined.
	// * @param virtualizationProviders
	// * List of {@link IVirtualizationPlatformArchitectureElement}s that contain the
	// * network drivers.
	// * @return Predicate if the given element is referenced by a network driver located contained
	// in
	// * the given virtualization providers.
	// */
	// private static boolean isReferencedByNetworkDriver(IPlatformResource element,
	// Collection<IVirtualizationPlatformArchitectureElement> virtualizationProviders) {
	// EList<GatewayUnit> allPresentNetworkDrivers =
	// getAllChildrenWithType(virtualizationProviders, OnChipNetworkDriver.class);
	// for(GatewayUnit currentNetworkDriver : allPresentNetworkDrivers) {
	// EList<IHierarchicElement> referencedGateway =
	// getReferencedElementsWithType(currentNetworkDriver, IHierarchicElement.class);
	// assert (referencedGateway.size() == 1) : "The OnChipNetworkDriver " +
	// currentNetworkDriver.getName() +
	// " references none or more than one network interface. This is not supported.";
	// if(referencedGateway.get(0) == element) {
	// return true;
	// }
	// }
	// return false;
	// }

	/**
	 * Returns a list of child elements of the given type for each parent in the given collection.
	 * NOTE: This operation can be very costly, if elements are contained in the list which are far
	 * up in the hierarchy.
	 * 
	 * @param parents
	 *            Collection of parents to be examined.
	 * @param type
	 *            The class to be filter for.
	 * @return List of child elements of the given type.
	 */
	public static <S> EList<S> getAllChildrenWithType(Collection<? extends EObject> parents,
			Class<S> type) {
		EList<S> childrenWithType = new BasicEList<S>();
		for(EObject currentParent : parents) {
			childrenWithType.addAll(getChildrenWithType(currentParent, type));
		}
		return childrenWithType;
	}

	// /**
	// * Predicate, if the given {@link IHierarchicElement} has direct child elements are connected
	// * via directed connections.
	// */
	// public static boolean hasCildrenWithDirectedConnections(IHierarchicElement element) {
	// for(IHierarchicElement currentElement : element.getContainedElements()) {
	// if(directedConnectionElements.contains(currentElement.getClass())) {
	// return true;
	// }
	// }
	// return false;
	// }

	/**
	 * Determines whether the given hardware element is virtualized by a
	 * {@link IVirtualizationPlatformArchitectureElement}.
	 * 
	 * @param element
	 *            The hardware platform element to be examined.
	 * @param virtualizationProviders
	 *            The (complete) collection of {@link IVirtualizationPlatformArchitectureElement}s
	 *            present in the system.
	 * @return Boolean indicating the presence of a
	 *         {@link IVirtualizationPlatformArchitectureElement} connected to the given
	 *         hardware element.
	 */
	public static boolean hasVirtualizationLayer(IPlatformResource element,
			Collection<IVirtualizationPlatformArchitectureElement> virtualizationProviders) {
		Collection<IPlatformResource> selfAndParentElements = new HashSet<>();
		selfAndParentElements.add(element);
		selfAndParentElements.addAll(getParentsWithType(element, IPlatformResource.class));

		Collection<IPlatformResource> virtualizedResources = new HashSet<IPlatformResource>();
		for(IVirtualizationPlatformArchitectureElement currentVirtualizationProvider : virtualizationProviders) {
			virtualizedResources.addAll(PlatformArchitectureUtils
					.getResourceReferences((IModelElement)currentVirtualizationProvider));
		}

		return (Collections.disjoint(selfAndParentElements, virtualizedResources)) ? false : true;
	}

	/**
	 * Collects and returns all {@link IModelElementSpecification}s of the given type which are
	 * contained in the specification list of the elements and their children which are referenced
	 * by a {@link ResourceLink} {@link IModelElementSpecification} of the given
	 * {@link IModelElement}. If no {@link ResourceLink} is specified for the given
	 * {@link IModelElement} or if no referenced {@link IModelElement} or one of its children
	 * contains a {@link IModelElementSpecification} of the given type, an empty list is returned.
	 * 
	 * @param specType
	 *            The {@link IModelElementSpecification} type to search for.
	 * @param modelElement
	 *            {@link IModelElement} containing a {@link ResourceLink} to other
	 *            {@link IModelElement}s.
	 * @return List of {@link IModelElementSpecification}s contained in the referenced
	 *         {@link IModelElement}s or one of its children.
	 */
	public static <T extends IModelElementSpecification> List<T>
			getAllReferencedSpecificationsWithType(Class<T> specType, IModelElement modelElement) {
		List<T> specifications = new ArrayList<T>();
		if(modelElement == null) {
			return specifications;
		}

		Collection<IHierarchicElement> referencedElements =
				getReferencedElementsWithType(modelElement, IHierarchicElement.class);
		for(IHierarchicElement refElement : referencedElements) {
			specifications.addAll(getChildrenWithType(refElement, specType));
		}

		return specifications;
	}

	/**
	 * Returns the {@link ResourceLink} {@link IModelElementSpecification} of the given
	 * {@link IModelElement} or the {@link ResourceLink} or the {@link ResourceLink} of the parent
	 * {@link IModelElement} that is "closest" to the given {@link IModelElement} regarding the
	 * hierarchy.
	 * 
	 * @param element
	 *            The {@link IModelElement} which a matching (see above) {@link ResourceLink} shall
	 *            be found.
	 * @return {@link ResourceLink} {@link IModelElementSpecification}, or <code>null</code> if
	 *         neither the given {@link IModelElement}, nor any of its containing elements contains
	 *         a {@link ResourceLink} {@link IModelElementSpecification}.
	 */
	public static ResourceLink getSelfOrAnyParentResourceLink(IModelElement element) {
		ResourceLink resourceLink =
				pickFirstInstanceOf(ResourceLink.class, element.getSpecifications());
		if(resourceLink == null) {
			List<ResourceLink> parentResourceLinks =
					getParentsWithType(element, ResourceLink.class);
			if(!parentResourceLinks.isEmpty()) {
				// List is ordered from the element itself to the top element -> find "closest"
				// resource link of the given element.
				resourceLink = parentResourceLinks.get(0);
			}
		}
		return resourceLink;
	}

	/**
	 * Determines whether the given {@link IHierarchicElement}s are contained within the
	 * same (parent) {@link IHierarchicElement}.
	 * 
	 * @param elementA
	 *            The first {@link IHierarchicElement} to be checked.
	 * @param elementB
	 *            The {@link IHierarchicElement} to be checked against.
	 * @return see above.
	 */
	public static boolean haveSameParentElement(IHierarchicElement elementA,
			IHierarchicElement elementB) {
		return (elementA.getContainer() == elementB.getContainer()) ? true : false;
	}

	/**
	 * Determines whether the given {@link IHierarchicElement}s are contained within the
	 * same (parent) {@link IHierarchicElement}s. This method is an extension to {#link
	 * haveSameParentElement(IHierarchicElement, IHierarchicElement)} which also considers the case
	 * where the given elements can be part of a virtual platform.
	 * 
	 * @param elementA
	 *            The first {@link IHierarchicElement} to be checked.
	 * @param elementB
	 *            The {@link IHierarchicElement} to be checked against.
	 * @param virtualizationProviders
	 *            The (complete) collection of {@link IVirtualizationPlatformArchitectureElement}s
	 *            present in the
	 *            system.
	 * @return see above.
	 */
	public static boolean haveSameParentHardwareElement(IHierarchicElement elementA,
			IHierarchicElement elementB,
			Collection<IVirtualizationPlatformArchitectureElement> virtualizationProviders) {
		elementA = getReferencedElementsWithType(elementA, IHierarchicElement.class).get(0);
		elementB = getReferencedElementsWithType(elementB, IHierarchicElement.class).get(0);
		return (haveSameParentElement(elementA, elementB)) ? true : false;
	}

	/**
	 * Creates a filtered list from a given list of superclasses of the filter class.
	 * 
	 * @param generalList
	 *            The list of superclasses of {#link filterClass} which will be filtered.
	 * @param filterClass
	 *            The class by which the given list shall be filtered
	 * @return The filter list with elements of the type T
	 */
	@SuppressWarnings("unchecked")
	public static <S, T extends S> EList<T> getElementsWithType(EList<S> generalList,
			Class<T> filterClass) {
		EList<T> filteredList = new BasicEList<T>();
		for(S listElement : generalList) {
			if(filterClass.isAssignableFrom(listElement.getClass())) {
				filteredList.add((T)listElement);
			}
		}

		return filteredList;
	}

	/**
	 * Returns the List of <i>all</i> {@link TimeSlot}s that are predecessors of the given
	 * {@link TimeSlot}. Therefore, this method evaluates the predecessor field of {@link TimeSlot}
	 * s.
	 * 
	 * @param timeSlot
	 *            The {@link TimeSlot} for which the predecessors shall be identified.
	 * @return List of all predecessor {@link TimeSlot}s.
	 */
	public static List<TimeSlot> getAllPredecessorsOfTimeSlot(TimeSlot timeSlot) {
		List<TimeSlot> predescessorList = new ArrayList<TimeSlot>();
		Queue<TimeSlot> toBeVisited = new LinkedList<TimeSlot>();
		toBeVisited.addAll(timeSlot.getPredecessors());
		while(!toBeVisited.isEmpty()) {
			TimeSlot currentPredecessor = toBeVisited.poll();
			if(!predescessorList.contains(currentPredecessor)) {
				predescessorList.add(currentPredecessor);
				toBeVisited.addAll(currentPredecessor.getPredecessors());
			}
		}

		return predescessorList;
	}

	/**
	 * Returns the List of <i>all</i> {@link TimeSlot}s that are successors of the given
	 * {@link TimeSlot}. Therefore, this method evaluates the successor field of {@link TimeSlot}s.
	 * 
	 * @param timeSlot
	 *            The {@link TimeSlot} for which the successors shall be identified.
	 * @return List of all successor {@link TimeSlot}s.
	 */
	public static List<TimeSlot> getAllSuccessorsOfTimeSlot(TimeSlot timeSlot) {
		List<TimeSlot> successorList = new ArrayList<TimeSlot>();
		Queue<TimeSlot> toBeVisited = new LinkedList<TimeSlot>();
		toBeVisited.addAll(timeSlot.getSuccessors());
		while(!toBeVisited.isEmpty()) {
			TimeSlot currentPredecessor = toBeVisited.poll();
			if(!successorList.contains(currentPredecessor)) {
				successorList.add(currentPredecessor);
				toBeVisited.addAll(currentPredecessor.getSuccessors());
			}
		}

		return successorList;
	}

	/**
	 * Returns a collection of all child {@link EObject}s that have either of the given types.
	 * 
	 * @param parent
	 *            {@link EObject} to start the search from.
	 * @param typeList
	 *            filter types for the child elements.
	 * @return Collection of child {@link EObject}s conforming to either of the given types.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends EObject> Collection<T> getChildrenWithTypes(EObject parent,
			Collection<Class<?>> typeList, boolean isReflexive) {
		Collection<T> children = new HashSet<T>();
		if(isReflexive && isAssignableFromAny(parent, typeList)) {
			children.add((T)parent);
		}

		TreeIterator<EObject> content = getAllContents(parent, true);
		while(content.hasNext()) {
			EObject child = content.next();
			if(isAssignableFromAny(child, typeList)) {
				children.add((T)child);
			}
		}
		return children;
	}

	/**
	 * Debug method to search collections for strings.
	 * 
	 * @param searchString
	 *            The string for which the elements in the collection shall be evaluated against.
	 * @param objectCollection
	 *            The collection of objects to be examined
	 * @return boolean, if any object's toString method contains the search string.
	 */
	public static boolean containsElementByString(String searchString,
			Collection<?> objectCollection) {
		for(Object obj : objectCollection) {
			if(obj.toString().contains(searchString)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Returns the superclass of the given element which implements the given interface.
	 * 
	 * @param element
	 *            The element whose implementing interface class shall be returned.
	 * @param superClass
	 *            The interface class for which the superclasses of {#link element} shall be
	 *            examined
	 * @return The first identified instance of a superclass implementing {#link superClass}, or
	 *         {#code null} if none was found.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Class<T> getSuperClassOfImplementingInterface(EObject element,
			Class<T> superClass) {
		EList<EClass> superClasses = element.eClass().getESuperTypes();
		for(EClass currentSuperClass : superClasses) {
			if(superClass.isAssignableFrom(currentSuperClass.getInstanceClass())) {
				return (Class<T>)currentSuperClass.getInstanceClass();
			}
		}
		return null;
	}

	/**
	 * Determines the {@link IProblemDimension} of the given {@link ExplorationTarget}.
	 * 
	 * @param goal
	 *            The {@link ExplorationTarget} whose {@link IProblemDimension} shall be identified.
	 * @return The {@link IProblemDimension} of the given {@link ExplorationTarget} or {#code null}
	 *         if none is identified.
	 */
	public static <T extends IProblemDimension> Class<IProblemDimension>
			getAssociatedDimensionFromGoal(ExplorationTarget<?> goal) {
		return getSuperClassOfImplementingInterface(goal, IProblemDimension.class);
	}

	/**
	 * Given an object that directly or indirectly inherits from the given interface, this method
	 * returns the substitution class type of the generic parameter of the given interface.
	 * <p>
	 * <b>NOTE:</b> This method works only for interfaces that have a single generic type parameter.
	 * <p>
	 * This method makes no assumptions about the location of the substitution in the class
	 * hierarchy.
	 * <p>
	 * <p>
	 * This method first identifies the path from the given object's class to the given interface in
	 * the class hierarchy. This path is then traversed and each found generic substitution is
	 * examined whether it substitutes the generic of the given interface.
	 * 
	 * @param object
	 *            Object for which the substitution type defined in the given class shall be found.
	 * @param interfaceClass
	 *            Interface defining or redefining the generic type of interest.
	 * @return Class type substituting the generic parameter of the given interface.
	 */
	public static Class<?>
			getSubstitutionTypeOfGenericIface(Object object, Class<?> interfaceClass) {
		// First, identify the "hierarchy path" from the given object to the target interface.
		LinkedList<Class<?>> inheritancePath = getClassPathToIface(object, interfaceClass);

		Stack<Type> paramTypeStack = new Stack<>();
		LinkedList<Class<?>> hierTraversalList = new LinkedList<>();
		hierTraversalList.addAll(inheritancePath);

		Type currParamRawType = null;
		Predicate<Class<?>> targetIfaceFound = t -> t != null && t.isAssignableFrom(interfaceClass);
		// Outer loop: Inspect the "hierarchy path".
		do {
			Class<?> hierStackElem = hierTraversalList.poll();
			Type[] genTypes = hierStackElem.getGenericInterfaces();
			LinkedList<Type> typesToTraverse = new LinkedList<>();
			typesToTraverse.addAll(Arrays.asList(genTypes));

			// Inner loop: Inspect a single generic substitution.
			while(!typesToTraverse.isEmpty() && !targetIfaceFound.test((Class<?>)currParamRawType)) {
				Type currTraversedType = typesToTraverse.poll();
				paramTypeStack.add(currTraversedType);

				// Skip non-parameterized types (like other interfaces): Would be a wrong path.
				if(currTraversedType instanceof ParameterizedType) {
					currParamRawType = ((ParameterizedType)currTraversedType).getRawType();
					// Only follow the "path" of this generic parameter if it matches the previously
					// identified path to the given interface.
					if(inheritancePath.contains(currParamRawType)) {
						Collection<Type> matchedTypes =
								processParentParamerType(typesToTraverse, currParamRawType);
						typesToTraverse.addAll(0, matchedTypes);
					}
				} else {
					// Otherwise, remove the current parameter type from the "path" stack: It would
					// lead to the generic of the interface.
					paramTypeStack.pop();
				}
			}
			if(!targetIfaceFound.test((Class<?>)currParamRawType)) {
				// This generic substitution was not the right one --> clean the "path".
				paramTypeStack.clear();
			}
		} while(!hierTraversalList.isEmpty() && !targetIfaceFound.test((Class<?>)currParamRawType));

		// Return the class of the substituting parameter type.
		if(targetIfaceFound.test((Class<?>)currParamRawType)) {
			Type paramType = paramTypeStack.get(0);
			if(paramType instanceof ParameterizedType) {
				Type actParamType = ((ParameterizedType)paramType).getActualTypeArguments()[0];
				if(actParamType instanceof Class<?>) {
					return (Class<?>)actParamType;
				} else if(actParamType instanceof ParameterizedType) {
					return (Class<?>)((ParameterizedType)actParamType).getRawType();
				}
			}
		}

		return null;
	}

	/**
	 * Identifies the type parameter from the super classes/interface which matches the
	 * 
	 * @param typesToTraverse
	 * @param rawType
	 */
	private static Collection<Type> processParentParamerType(LinkedList<Type> typesToTraverse,
			Type rawType) {
		Collection<Type> matchingTypes = new ArrayList<>();
		Type[] typeParams = ((Class<?>)rawType).getGenericInterfaces();
		Collection<Type> xT = Arrays.asList(((Class<?>)rawType).getTypeParameters());
		for(Type parentType : typeParams) {
			if(parentType instanceof ParameterizedType) {
				Type[] typeArgs = ((ParameterizedType)parentType).getActualTypeArguments();
				if(typeArgs.length == 1 && xT.contains(typeArgs[0])) {
					matchingTypes.add(parentType);
				}
			}
		}
		return matchingTypes;
	}

	/**
	 * Determines the path in the class hierarchy from the given object's class to the interface. It
	 * returns a list view on the identified path.
	 * <p>
	 * <b>NOTE:</b> The given object's class must be always below the given interface's class in the
	 * class hierarchy. Although this property is enforced if the second class is an interface, it
	 * is specifically noted here since normal classes and interface cannot be differentiated here.
	 * 
	 * @param object
	 *            Object inheriting from {@code interfaceClass}.
	 * @param interfaceClass
	 *            Interface forming the top of the stack and defining the end of the path.
	 * @return Path between the given object's class and the given interface as a linked list.
	 */
	private static LinkedList<Class<?>> getClassPathToIface(Object object, Class<?> interfaceClass) {
		LinkedList<Class<?>> traversalList = new LinkedList<>();
		Stack<Class<?>> hierarchyBrachStack = new Stack<>();
		LinkedList<Class<?>> ifacesTraversalList = new LinkedList<>();

		ifacesTraversalList.addAll(Arrays.asList(object.getClass().getInterfaces()));
		while(!ifacesTraversalList.isEmpty()) {
			Class<?> currIface = ifacesTraversalList.poll();
			traversalList.add(currIface);
			if(currIface.equals(interfaceClass)) {
				break;
			}

			Class<?>[] parentIfaces = currIface.getInterfaces();
			if(parentIfaces.length > 0) {
				// Add candidate paths to the classes to traverse. If multiple candidate paths are
				// found, remember the branch point.
				// Note: always insert the parent interface at the top of the traversal list.
				ifacesTraversalList.addAll(0, Arrays.asList(parentIfaces));
				if(parentIfaces.length > 1) {
					hierarchyBrachStack.add(currIface);
				}
			} else {
				// No more parent classes: Remove invalid path fragment from the stack (does not
				// lead to the target interface).
				int lastBranchIdx = 1;
				if(!hierarchyBrachStack.isEmpty()) {
					Class<?> lastHierBranch = hierarchyBrachStack.lastElement();
					traversalList.indexOf(lastHierBranch);
				}
				List<Class<?>> rmFromTravStrack = new ArrayList<>();
				rmFromTravStrack.addAll(traversalList.subList(lastBranchIdx, traversalList.size()));
				traversalList.removeAll(rmFromTravStrack);
			}
		}
		return traversalList;
	}

	/**
	 * Returns the typed key from the map which is the best match for the given
	 * class. The best match is defined by the first match occurring in a breath
	 * first search of the inheritance tree, where the base class is always
	 * visited before the implemented interfaces. Interfaces are traversed in
	 * the order they are defined in the source file. The only exception is {@link Object}, which is
	 * considered only as the very last option.
	 * <p>
	 * As this lookup can be expensive (reflective iteration over the entire inheritance tree) the
	 * results should be cached if multiple lookups for the same class are expected.
	 * 
	 * @param clazz
	 *            the class being looked up.
	 * @param keySet
	 *            the set of keys to perform the lookup in.
	 * @return the best match found or <code>null</code> if no matching key
	 *         was found. Note that <code>null</code> will also be returned if
	 *         the entry for the best matching class was <code>null</code>.
	 * 
	 *         Adapted from {@link ReflectionUtils#performNearestClassLookup(Class, Map)}.
	 */
	@SuppressWarnings("unchecked")
	public static <K> Class<? extends K> performNearestClassLookup(Class<? extends K> clazz,
			Set<Class<? extends K>> keySet) {
		Queue<Class<? extends K>> q = new LinkedList<Class<? extends K>>();
		q.add(clazz);

		Class<? extends K> current = null;
		while(!q.isEmpty()) {
			current = q.poll();
			if(keySet.contains(current)) {
				return current;
			}

			Class<?> superClass = current.getSuperclass();
			if(superClass != null && superClass != Object.class) {
				q.add((Class<? extends K>)superClass);
			}

			for(Class<?> iface : current.getInterfaces()) {
				q.add((Class<? extends K>)iface);
			}
		}
		return current;
	}

	/**
	 * Returns the value from the {@link Multimap}, whose key is the best match for the given
	 * class. The best match is defined by the first match occurring in a breath
	 * first search of the inheritance tree, where the base class is always
	 * visited before the implemented interfaces. Interfaces are traversed in
	 * the order they are defined in the source file. The only exception is {@link Object}, which is
	 * considered only as the very last option.
	 * <p>
	 * As this lookup can be expensive (reflective iteration over the entire inheritance tree) the
	 * results should be cached if multiple lookups for the same class are expected.
	 * 
	 * 
	 * @param clazz
	 *            the class being looked up.
	 * @param classMap
	 *            the {@link Multimap} to perform the lookup in.
	 * @return the best match found or the empty set if no matching entry
	 *         was found. Note that the empty set will also be returned if
	 *         the entry for the best matching class was the empty set.
	 * 
	 *         Adapted from {@link ReflectionUtils#performNearestClassLookup(Class, Map)}.
	 */
	public static <K, T> Collection<T> performNearestClassLookup(Class<? extends K> clazz,
			Multimap<Class<? extends K>, T> classMap) {
		Class<? extends K> key = performNearestClassLookup(clazz, classMap.keySet());
		return classMap.get(key);
	}

	/**
	 * Creates an {@link IStatus} element that aggregates stack trace messages is a
	 * {@link MultiStatus} object, if possible.
	 * 
	 * @param e
	 *            Exception to be show.
	 * @param launcherPluginID
	 *            Corresponding ID of the caller's plugin.
	 * @param title
	 *            Title string to be shown.
	 * @return {@link MultiStatus}, if the given exception is non-{@code null}, otherwise a
	 *         {@link Status} object.
	 */
	public static IStatus createExceptionStatus(Exception e, String launcherPluginID, String title) {
		IStatus status =
				new Status(IStatus.ERROR, launcherPluginID, "Error executing " + title + ".", e);
		status = tryCreateMultiStatusException(status);

		if(ExplorationUtils.isDebugVerboseEnabled()) {
			e.printStackTrace();
		}
		return status;
	}

	/**
	 * Aggregates possibly nested stack traces of an {@link IStatus} element and returns a
	 * MultiStatus object if exceptions are found. Otherwise the given {@link IStatus} is returned.
	 * Inspired from
	 * http://www.vogella.com/tutorials/EclipseDialogs/article.html#dialogs_jface_errordialog.
	 */
	static IStatus tryCreateMultiStatusException(IStatus status) {
		Throwable t = status.getException();
		if(t != null) {
			StackTraceElement[] stackTraces = status.getException().getStackTrace();
			Stream<Status> childStatuses =
					Arrays.stream(stackTraces).map(
							tr -> new Status(IStatus.ERROR, status.getPlugin(), tr.toString()));

			return new MultiStatus(status.getPlugin(), status.getSeverity(),
					childStatuses.toArray(size -> new Status[size]), status.getMessage(), t);
		}
		return status;
	}
}
