/*--------------------------------------------------------------------------+
$Id: ExplorationDslModelElementFactory.java 3175 2016-07-04 15:25:48Z diewald $
|                                                                          |
| Copyright 2016 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.util;

import org.eclipse.emf.common.util.EList;
import org.fortiss.af3.exploration.dsl.model.expressions.ArithmeticExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.BooleanExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.ConstraintExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.ElementExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.Expression;
import org.fortiss.af3.exploration.dsl.model.expressions.ExpressionsFactory;
import org.fortiss.af3.exploration.dsl.model.expressions.ObjectiveExpression;
import org.fortiss.af3.exploration.dsl.model.expressions.PrefixExpression;
import org.fortiss.af3.exploration.dsl.model.operators.ArithmeticOp;
import org.fortiss.af3.exploration.dsl.model.operators.ArithmeticOperatorEnum;
import org.fortiss.af3.exploration.dsl.model.operators.ComparisonOperatorEnum;
import org.fortiss.af3.exploration.dsl.model.operators.LocationOp;
import org.fortiss.af3.exploration.dsl.model.operators.LocationOperatorEnum;
import org.fortiss.af3.exploration.dsl.model.operators.OperatorsFactory;
import org.fortiss.af3.exploration.dsl.model.operators.OptimizationDirection;
import org.fortiss.af3.exploration.dsl.model.operators.OptimizationDirectionEnum;
import org.fortiss.af3.exploration.dsl.model.operators.SummationOp;
import org.fortiss.af3.exploration.dsl.model.patterns.ElementProperty;
import org.fortiss.af3.exploration.dsl.model.patterns.LocationPattern;
import org.fortiss.af3.exploration.dsl.model.patterns.PatternsFactory;
import org.fortiss.af3.exploration.dsl.model.patterns.SinglePropValueComparison;
import org.fortiss.af3.exploration.dsl.model.patterns.TwoPropertyValueComparison;
import org.fortiss.af3.exploration.dsl.model.types.NumberLiteral;
import org.fortiss.af3.exploration.dsl.model.types.TypesFactory;
import org.fortiss.af3.exploration.model.location.LocationConstraint;
import org.fortiss.tooling.base.model.element.IModelElement;

/**
 * Model element factory for Expressions of the explorationDsl.
 * 
 * @author diewald
 * @author $Author: diewald $
 * @version $Rev: 3175 $
 * @ConQAT.Rating YELLOW Hash: 5C34767AEE1586D8B0F18DC96A8637D0
 */
public class ExplorationDslModelElementFactory {

	/** Creates an "empty" {@link SummationOp}. */
	public static <T> SummationOp<T> createSummationOp() {
		SummationOp<T> sumOp = OperatorsFactory.eINSTANCE.createSummationOp();
		sumOp.setOperator(ArithmeticOperatorEnum.ADD);
		return sumOp;
	}

	/**
	 * Creates an {@link ObjectiveExpression} optimizing towards the given
	 * {@link OptimizationDirection}.
	 */
	public static ObjectiveExpression
			createObjectiveExpression(OptimizationDirectionEnum direction) {
		ObjectiveExpression expr = ExpressionsFactory.eINSTANCE.createObjectiveExpression();
		expr.setOperator(direction);
		return expr;
	}

	/** Creates an "empty" {@link ConstraintExpression}. */
	public static <T extends BooleanExpression> ConstraintExpression<T> createConstraintExpression(
			EList<PrefixExpression> prefixes, EList<BooleanExpression> predicates, T expression) {
		ConstraintExpression<T> expr = ExpressionsFactory.eINSTANCE.createConstraintExpression();
		if(prefixes != null) {
			expr.getPrefix().addAll(prefixes);
		}
		if(predicates != null) {
			expr.getPredicate().addAll(predicates);
		}
		expr.setExpression(expression);
		return expr;
	}

	/** Creates a {@link LocationPattern} used in {@link LocationConstraint}s. */
	@SuppressWarnings("unchecked")
	public static <LHT extends IModelElement, RHT extends IModelElement> LocationPattern<LHT, RHT>
			createLocationPattern() {
		LocationPattern<IModelElement, IModelElement> locPattern =
				PatternsFactory.eINSTANCE.createLocationPattern();
		return (LocationPattern<LHT, RHT>)locPattern;
	}

	/** Creates a {@link LocationPattern} used in {@link LocationConstraint}s. */
	@SuppressWarnings("unchecked")
	public static <LHT extends IModelElement, RHT extends IModelElement> LocationPattern<LHT, RHT>
			createLocationPattern(LHT sourceObj, LocationOperatorEnum op, RHT targetObj) {
		LocationPattern<IModelElement, IModelElement> locPattern =
				PatternsFactory.eINSTANCE.createLocationPattern();
		// Create the left side, the operator, and the right hand side.
		LocationOp<IModelElement, IModelElement> lop =
				OperatorsFactory.eINSTANCE.createLocationOp();
		locPattern.setExpression(lop);
		ElementExpression<IModelElement> leftExpr = createElementExpression(sourceObj);
		locPattern.getExpression().setLeft(leftExpr);
		locPattern.getExpression().setOperator(op);
		ElementExpression<IModelElement> rightExpr = createElementExpression(targetObj);
		locPattern.getExpression().setRight(rightExpr);

		return (LocationPattern<LHT, RHT>)locPattern;
	}

	/**
	 * Creates an {@link ElementExpression} used to reference {@link IModelElement}s from
	 * {@link Expression}s.
	 */
	public static <ME extends IModelElement> ElementExpression<ME> createElementExpression(
			ME modelElement) {
		ElementExpression<ME> elemExpr = ExpressionsFactory.eINSTANCE.createElementExpression();
		elemExpr.setModelElement(modelElement);
		return elemExpr;
	}

	/** Creates an "empty" {@link ElementProperty} (no values/references are set). */
	public static <ME extends IModelElement, T> ElementProperty<ME, T> createElementProperty() {
		ElementProperty<ME, T> elemProp = PatternsFactory.eINSTANCE.createElementProperty();
		return elemProp;
	}

	/** Creates an "empty" {@link SinglePropValueComparison} (no values/references are set). */
	public static <T, ME extends IModelElement> SinglePropValueComparison<T, ME>
			createSinglePropValueCompExpression(ComparisonOperatorEnum compOpType) {
		ElementProperty<ME, T> modelElement = createElementProperty();
		// TODO: let the creator pass the literal? Differentiation by T? should the NumberLiteral
		// class be an interface?
		NumberLiteral<T> numLit = TypesFactory.eINSTANCE.createNumberLiteral();
		SinglePropValueComparison<T, ME> singleCompPropPattern =
				PatternsFactory.eINSTANCE.createSinglePropValueComparison();
		singleCompPropPattern.setLeft(modelElement);
		singleCompPropPattern.setOperator(compOpType);
		singleCompPropPattern.setRight(numLit);
		return singleCompPropPattern;
	}

	/** Creates an "empty" {@link TwoPropertyValueComparison} (no values/references are set). */
	public static <T, FME extends IModelElement, SME extends IModelElement>
			TwoPropertyValueComparison<T, FME, SME> createTwoProValCompExpression(
					ArithmeticOperatorEnum arithOpType, ComparisonOperatorEnum compOpType) {
		ElementProperty<FME, T> firstModelElement = createElementProperty();
		ElementProperty<SME, T> secondModelElement = createElementProperty();
		ArithmeticOp<T, ArithmeticExpression<T>, ArithmeticExpression<T>> arithOp =
				OperatorsFactory.eINSTANCE.createArithmeticOp();
		arithOp.setOperator(arithOpType);
		arithOp.setLeft(firstModelElement);
		arithOp.setRight(secondModelElement);
		// TODO: let the creator pass the literal? Differentiation by T? should the NumberLiteral
		// class be an interface?
		NumberLiteral<T> numLit = TypesFactory.eINSTANCE.createNumberLiteral();
		TwoPropertyValueComparison<T, FME, SME> twoCompPattern =
				PatternsFactory.eINSTANCE.createTwoPropertyValueComparison();
		twoCompPattern.setLeft(arithOp);
		twoCompPattern.setOperator(compOpType);
		twoCompPattern.setRight(numLit);
		return twoCompPattern;
	}
}
